/*****************************************************************************
/*
                                   FAO.c

Formatted write using sys$faol()-like directives, just a functional subset, it
also differs in some significant ways, including extended functionality.  The
functions can write into either specified storage or into the request's output
buffer (probably more efficient that sys$fao()ing into storage then copying
this to the network buffer).  With the advent of ODS-5, and extended file
specifications it becomes possible for file names to contain HTML and URL
forbidden characters.  A simple method for effectively escaping and encoding
these was required, hence the !HZ and !UZ (etc.) directives.  All parameters
are passed via a sys$faol()-like pointer-to-longword-vector.  Strings are
always assumed to be null-terminated.

A resultant string is always null-terminated (even after an error), hence the
'BufferSize' can always only store 'BufferSize'-1 characters.  'LengthPtr'
provides storage for the number of characters minus the null-termination.

It is no exageration to state that this functionality has made the sorts of
things undertaken with it ten times easier and more efficient!!


DIRECTIVES SUPPORTED
--------------------
!AC    counted ASCII string (first byte must be count, i.e. 0 to 255 chars)
!AZ    string
!HZ    string, add HTML-escaped
!UZ    string, add URL-encoded
!A5    ODS-5 file specification, (un-escapes ODS-5, i.e. "^_" into " ") 
!AF    file specification, (un-escapes ODS-5 characters, i.e. "^_" into " ") 
!AV    VMS-style specification, (un-escapes ODS-5 chars, i.e. "^_" into " ") 
!H5    ODS-5 file specification, add HTML-escaped
!HF    file specification, add HTML-escaped
!HV    VMS-style specification, add HTML-escaped
!IP    dotted-decimal from 32 bit network-order IP address
!SL    decimal, signed longword
!SQ    signed quadword
!U5    ODS-5 file specification, add URL-encoded 
!UF    file specification, add URL-encoded 
!UV    VMS-style specification, add URL-encoded
!UL    decimal, unsigned longword
!XL    hexadecimal, longword
!ZL    zero-filled decimal, unsigned longword
!%D    date/time
!%I    numeric to ASCII identifier
!%M    full status message string (returned from sys$getmsg())
!%m    text-only status message string (returned from sys$getmsg())
!%S    if the last converted numeric value was not 1 add "S"
!%s    if the last converted numeric value was not 1 add "s"
!%T    time
!%W    weekday/date/time (field-width applies to date/time component only)
!%?    simple conditional, "!%?string if *vector TRUE\rstring if FALSE\r"
!%%    nested format string address from vector (FILO stacks the current one)
!+     step-over the next parameter from the vector (anything BUT a "!%%")
!-     back-up to the previous parameter in the vector (anything BUT a "!%%")
!<     begin specified width output field (cannot nest, must have width!)
!>     end specified width output field
!!     exclamation mark

Each directive can have a positive, decimal integer between the '!' and the
directive providing a field width, e.g. "!10HZ", or have the field width taken
from the vector using "!#HZ".  These behave in the same manner as the sys$fao()
field width parameter. 

Any directive may use the indirect character (unlike sys$fao(), where only
numeric directives may).  This indicates the vector value is the 32 bit address
of the value, rather than the value itself, e.g. "!@AZ", "!10@UL", etc.

Any numeric directive may have a comma immediately following the '!'.  This
causes a number in excess of 3 digits to have commas inserted, e.g. "!,UL",
"!,10UL".  Caution - this can truncate to the right on field-width constrained
fields.

The 'xF' and 'xV' family of directives specially process file names.  The rules
are these.  If extended file specifications apply to the architecture and they
are enabled for processing (i.e. at least one is mapped), then the string
supplied is checked for ODS-2 compliance (i.e. "A-Z", "0-9", "_", "-", ".",
"$", "/", "~", ":", "[", "]") and if any character outside of these is present
it is considered an extended file name.

For an 'xF' directive an ODS-2 compliant string is always forced to lower-case,
extended-compliant strings are left as-is (note that all upper-case ODS-5 names
are also forced lower case using this algorithm).  For an 'xV' directive the
case is left unchanged.  Extended-compliant strings have '^' escape sequences
converted to the equivalent un-escaped character.  To render these as-are a
'!HZ' directive should be used.


VERSION HISTORY
---------------
17-MAR-2002  MGD  'SQ' kludge for VAX quadword values 
07-JUL-2001  MGD  add !IP to WriteFormatted()
17-MAY-2001  MGD  dignified with a module of it's own
02-FEB-2001  MGD  add !%I to WriteFormatted()
27-AUG-2000  MGD  add !SL to WriteFormatted()
04-MAR-2000  MGD  add WriteFormatted() and it's siblings
                  WriteFao(), WriteFaol(), WriteFaoNet()
*/
/*****************************************************************************/

/* standard C header files */
#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* VMS related header files */
#include <descrip.h>
#include <jpidef.h>
#include <libdef.h>
#include <libdtdef.h>
#include <lnmdef.h>
#include <opcdef.h>
#include <prvdef.h>
#include <psldef.h>
#include <rms.h>
#include <secdef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <syidef.h>

/* application related header files */
#include "wasd.h"

#define WASD_MODULE "FAO"

#define DEBUG_WRITEFAO 0

/******************/
/* global storage */
/******************/

int  OpcomMessages,
     OpcomTarget;

/********************/
/* external storage */
/********************/

#ifdef DBUG
extern BOOL Debug;
#else
#define Debug 0 
#endif

BOOL  OdsExtended;

extern int  WatchEnabled;

extern char  ErrorSanityCheck[],
             ProcessName[],
             SoftwareID[],
             Utility[];

extern char  *DayName[];

/*****************************************************************************/
/*
This is the primary FAO function described in the module prologue.  It will be
more than obvious it has been written to optimize efficiency (in the author's
limited appreciation anyway) rather than readability.  There is a #define in
NET.H making it available as NetWriteFaol(), and a #define in SUPPORT.H making
it available as WriteFaol().
*/ 

int WriteFormatted
(
REQUEST_STRUCT *rqptr,
char *BufferPtr,
int BufferSize,
unsigned short *LengthPtr,
char *FormatString,
unsigned long *VectorPtr
)
{
#define FORMAT_STACK_MAX 8

   static char  HexDigits [] = "0123456789abcdef";
   static char  EncBuffer [] = "%xx";
   static $DESCRIPTOR (DayDateTimeFaoDsc, "!AZ, !#%D\0");
   static $DESCRIPTOR (IPFaoDsc, "!UL.!UL.!UL.!UL\0");
   static $DESCRIPTOR (IPWidthFaoDsc, "!#<!UL.!UL.!UL.!UL!>\0");
   static $DESCRIPTOR (PDFaoDsc, "!%D\0");
   static $DESCRIPTOR (PIFaoDsc, "!%I\0");
   static $DESCRIPTOR (PTFaoDsc, "!%T\0");
   static $DESCRIPTOR (SLFaoDsc, "!SL\0");
   static $DESCRIPTOR (SLWidthFaoDsc, "!#SL\0");
   static $DESCRIPTOR (SQFaoDsc, "!@SQ\0");
   static $DESCRIPTOR (SQWidthFaoDsc, "!#@SQ\0");
   static $DESCRIPTOR (ULFaoDsc, "!UL\0");
   static $DESCRIPTOR (ULWidthFaoDsc, "!#UL\0");
   static $DESCRIPTOR (XLFaoDsc, "!XL\0");
   static $DESCRIPTOR (XLWidthFaoDsc, "!#XL\0");
   static $DESCRIPTOR (ZLFaoDsc, "!ZL\0");
   static $DESCRIPTOR (ZLWidthFaoDsc, "!#ZL\0");

   static char  String [256];
   static $DESCRIPTOR (StringDsc, String);

   BOOL  CommaNumber,
         HtmlEscaped,
         IndirectValue,
         PathOdsExtended,
         UrlEncoded,
         VmsFormat;
   int  ch, bcnt, brcnt, fw, ofw, idx,
        status,
        CommaCount,
        LastNumericValue, 
        PrevBcnt,
        SysGetMsgFlags;
   unsigned short  Length;
   unsigned long  DayOfWeek,
                  VectorValue;
   unsigned long  BinTime [2];
   unsigned long  *BinTimePtr;
   char  *bptr, *cptr, *fptr, *sptr,
         *LastFptr;
   char  *FormatStack [FORMAT_STACK_MAX];
#ifndef __ALPHA
   double  dValue;
#endif

   /*********/
   /* begin */
   /*********/

   if (Debug && DEBUG_WRITEFAO)
      fprintf (stdout, "WriteFormatted() |%s|\n", FormatString);

   status = SS$_NORMAL;
   LastNumericValue = 1; 
   /* start with output-field-width set to infinite */
   ofw = -1;

   if (LengthPtr != NULL) *LengthPtr = 0;

   if (BufferPtr == NULL)
   {
      /* first call, initialize a buffer */
      if (rqptr->rqOutput.BufferStructPtr == NULL)
         NetWriteBufferedInit (rqptr, true);

      bptr = rqptr->rqOutput.BufferCurrentPtr;
      bcnt = PrevBcnt = rqptr->rqOutput.BufferCount;
      brcnt = rqptr->rqOutput.BufferRemaining;
   }
   else
   {
      /* writing to specified storage */
      if (!(brcnt = BufferSize)) return (SS$_BADPARAM);
      bptr = BufferPtr;
      bcnt = PrevBcnt = 0;
   }

   /* put initial format string into index zero of the format string stack */
   idx = 0;
   FormatStack[idx] = FormatString;

   /*************************/
   /* nested format strings */
   /*************************/

   for (;;)
   {
      /*****************/
      /* format string */
      /*****************/

      fptr = FormatStack[idx];
      while (*fptr)
      {
         while (*fptr && *fptr != '!' && ofw)
         {
            /* a literal character */
            if (brcnt--)
            {
               *bptr++ = *fptr++;
               bcnt++; 
               ofw--;
               continue;
            }
            /* if writing to storage then it's overflowed */
            if (BufferPtr != NULL)
            {
               BufferPtr[BufferSize-1] = '\0';
               return (SS$_BUFFEROVF);
            }
            /* need another network buffer */
            NetWriteBufferedInit (rqptr, false);
            bptr = rqptr->rqOutput.BufferCurrentPtr;
            brcnt = rqptr->rqOutput.BufferRemaining;
         }
         /* if end of literal characters (i.e. not a '!') then break */
         if (!*fptr) break;

         /* if no character follows the '!' then just forget it */
         LastFptr = fptr;
         if (!*++fptr) break;

         /*********************/
         /* format directives */
         /*********************/

         /**
         if (Debug && DEBUG_WRITEFAO)
            fprintf (stdout, "fptr |%10.10s|\n", fptr);
         **/

         if (*fptr == ',')
         {
            fptr++;
            CommaNumber = true;
         }
         else
            CommaNumber = false;

         if (isdigit(*fptr))
         {
            /* field-width */
            fw = atoi(fptr);
            while (isdigit(*fptr)) fptr++;
            /* for writes into network buffers, let's keep this < 1024 */
            if (BufferPtr == NULL) fw &= 0xfff;
         }
         else
         if (*fptr == '#')
         {
            /* field-width from vector */
            fptr++;
            fw = (int)*VectorPtr++;
            /* for writes into network buffers, let's keep this < 1024 */
            if (BufferPtr == NULL) fw &= 0xfff;
         }
         else
            fw = -1;
         /**
         if (Debug && DEBUG_WRITEFAO)
            fprintf (stdout, "fw: %d\n", fw);
         **/

         if (*fptr == '@')
         {
            fptr++;
            IndirectValue = true;
         }
         else
            IndirectValue = false;

         /* default is no string to copy if it drops through */
         sptr = NULL;
         /* used in file name formatting directives */
         HtmlEscaped = PathOdsExtended = UrlEncoded = VmsFormat = false;

         /* directives in guessed order of most common usage */
         if (*(unsigned short*)fptr == 'AZ')
         {
            /**************************/
            /* null-terminated string */
            /**************************/

            fptr += 2;
            if (IndirectValue)
            {
               if (!*VectorPtr)
               {
                  status = SS$_ACCVIO;
                  break;
               }
               sptr = *(char*)*VectorPtr++;
            }
            else
               sptr = (char*)*VectorPtr++;
            if (!sptr)
            {
               status = SS$_ACCVIO;
               break;
            }
            if (Debug && DEBUG_WRITEFAO) fprintf (stdout, "!AZ |%s|\n", sptr);
            /* drop through to buffer the string pointed to by 'sptr' */
         }
         else
         if (*(unsigned short*)fptr == 'HZ')
         {
            /****************************************/
            /* null-terminated, HTML-escaped string */
            /****************************************/

            fptr += 2;
            if (IndirectValue)
            {
               if (!*VectorPtr)
               {
                  status = SS$_ACCVIO;
                  break;
               }
               cptr = *(char*)*VectorPtr++;
            }
            else
               cptr = (char*)*VectorPtr++;
            if (!cptr)
            {
               status = SS$_ACCVIO;
               break;
            }
            if (Debug && DEBUG_WRITEFAO) fprintf (stdout, "!HZ |%s|\n", cptr);

            while (*cptr && fw && ofw)
            {
               fw--;
               ofw--;
               switch (ch = *cptr++)
               {
                  case '<' :
                     sptr = "&lt;";
                     break;
                  case '>' :
                     sptr = "&gt;";
                     break;
                  case '&' :
                     sptr = "&amp;";
                     break;
                  case '\"' :
                     sptr = "&quot;";
                     break;
                  default :
                     /* insert this character as-is */
                     if (!brcnt)
                     {
                        /* if writing to storage then it's overflowed */
                        if (BufferPtr != NULL)
                        {
                           BufferPtr[BufferSize-1] = '\0';
                           return (SS$_BUFFEROVF);
                        }
                        /* need another network buffer */
                        NetWriteBufferedInit (rqptr, false);
                        bptr = rqptr->rqOutput.BufferCurrentPtr;
                        brcnt = rqptr->rqOutput.BufferRemaining;
                     }
                     *bptr++ = ch;
                     bcnt++; 
                     brcnt--;
                     continue;
               }

               /* add the escaped character */
               while (*sptr)
               {
                  if (brcnt--)
                  {
                     *bptr++ = *sptr++;
                     bcnt++; 
                     continue;
                  }
                  /* if writing to storage then it's overflowed */
                  if (BufferPtr != NULL)
                  {
                     BufferPtr[BufferSize-1] = '\0';
                     return (SS$_BUFFEROVF);
                  }
                  /* need another network buffer */
                  NetWriteBufferedInit (rqptr, false);
                  bptr = rqptr->rqOutput.BufferCurrentPtr;
                  brcnt = rqptr->rqOutput.BufferRemaining;
               }
            }
            /* only if required drop through to space-fill */
            if (fw <= 0) continue;
            /* no string to copy  */
            sptr = NULL;
         }
         else
         if (*(unsigned short*)fptr == 'UZ')
         {
            /***************************************/
            /* null-terminated, URL-encoded string */
            /***************************************/

            fptr += 2;
            if (IndirectValue)
            {
               if (!*VectorPtr)
               {
                  status = SS$_ACCVIO;
                  break;
               }
               cptr = *(char*)*VectorPtr++;
            }
            else
               cptr = (char*)*VectorPtr++;
            if (!cptr)
            {
               status = SS$_ACCVIO;
               break;
            }
            if (Debug && DEBUG_WRITEFAO) fprintf (stdout, "!UZ |%s|\n", cptr);

            while (*cptr && fw && ofw)
            {
               fw--;
               ofw--;
               ch = *cptr++;
               if (isalnum(ch) ||
                   ch == '/' ||
                   ch == '-' ||
                   ch == '_' ||
                   ch == '$' ||
                   ch == '*' ||
                   ch == ':' ||
                   ch == '[' ||
                   ch == ']' ||
                   ch == ';' ||
                   ch == '~' ||
                   ch == '.')
               {
                  /* insert this character as-is */
                  if (!brcnt)
                  {
                     /* if writing to storage then it's overflowed */
                     if (BufferPtr != NULL)
                     {
                        BufferPtr[BufferSize-1] = '\0';
                        return (SS$_BUFFEROVF);
                     }
                     /* need another network buffer */
                     NetWriteBufferedInit (rqptr, false);
                     bptr = rqptr->rqOutput.BufferCurrentPtr;
                     brcnt = rqptr->rqOutput.BufferRemaining;
                  }
                  *bptr++ = ch;
                  bcnt++;
                  brcnt--;
                  continue;
               }

               /* encode this character */
               EncBuffer[1] = HexDigits[(ch & 0xf0) >> 4];
               EncBuffer[2] = HexDigits[ch & 0x0f];
               sptr = EncBuffer;
               /* add the encoded character */
               while (*sptr)
               {
                  if (brcnt--)
                  {
                     *bptr++ = *sptr++;
                     bcnt++; 
                     continue;
                  }
                  /* if writing to storage then it's overflowed */
                  if (BufferPtr != NULL)
                  {
                     BufferPtr[BufferSize-1] = '\0';
                     return (SS$_BUFFEROVF);
                  }
                  /* need another network buffer */
                  NetWriteBufferedInit (rqptr, false);
                  bptr = rqptr->rqOutput.BufferCurrentPtr;
                  brcnt = rqptr->rqOutput.BufferRemaining;
               }
            }
            /* only if required drop through to space-fill */
            if (fw <= 0) continue;
            /* no string to copy  */
            sptr = NULL;
         }
         else
         if (*(unsigned short*)fptr == 'UL')
         {
            /*********************/
            /* unsigned longword */
            /*********************/
   
            fptr += 2;
            if (IndirectValue)
            {
               if (!*VectorPtr)
               {
                  status = SS$_ACCVIO;
                  break;
               }
               LastNumericValue = *(unsigned long*)VectorPtr++;
            }
            else
               LastNumericValue = *VectorPtr++;
            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "!UL %d\n", LastNumericValue);
            if (fw > 0)
               status = sys$fao (&ULWidthFaoDsc, 0, &StringDsc,
                                 fw, LastNumericValue);
            else
               status = sys$fao (&ULFaoDsc, 0, &StringDsc,
                                 LastNumericValue);
            if (VMSnok (status)) break;
            if (!CommaNumber) fw = -1;
            sptr = String;
            /* drop through to buffer the string pointed to by 'sptr' */
         }
         else
         if (*(unsigned short*)fptr == 'XL')
         {
            /************************/
            /* hexadecimal longword */
            /************************/

            fptr += 2;
            if (IndirectValue)
            {
               if (!*VectorPtr)
               {
                  status = SS$_ACCVIO;
                  break;
               }
               LastNumericValue = *(unsigned long*)VectorPtr++;
            }
            else
               LastNumericValue = *VectorPtr++;
            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "!UL %d\n", LastNumericValue);
            if (fw > 0)
               status = sys$fao (&XLWidthFaoDsc, 0, &StringDsc,
                                 fw, LastNumericValue);
            else
               status = sys$fao (&XLFaoDsc, 0, &StringDsc,
                                 LastNumericValue);
            if (VMSnok (status)) break;
            if (!CommaNumber) fw = -1;
            sptr = String;
            /* drop through to buffer the string pointed to by 'sptr' */
         }
         else
         if (*(unsigned short*)fptr == 'ZL')
         {
            /************************/
            /* zero-filled longword */
            /************************/

            fptr += 2;
            if (IndirectValue)
            {
               if (!*VectorPtr)
               {
                  status = SS$_ACCVIO;
                  break;
               }
               LastNumericValue = *(unsigned long*)VectorPtr++;
            }
            else
               LastNumericValue = *VectorPtr++;
            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "!UL %d\n", LastNumericValue);
            if (fw > 0)
               status = sys$fao (&ZLWidthFaoDsc, 0, &StringDsc,
                                 fw, LastNumericValue);
            else
               status = sys$fao (&ZLFaoDsc, 0, &StringDsc,
                                 LastNumericValue);
            if (VMSnok (status)) break;
            if (!CommaNumber) fw = -1;
            sptr = String;
            /* drop through to buffer the string pointed to by 'sptr' */
         }
         else
         if (*(unsigned short*)fptr == 'SQ')
         {
            /*******************/
            /* signed quadword */
            /*******************/
   
            /*
               Note that the "!SQ" directive is being used, for VMS 6.2
               the "!UQ" && "!UJ" loop infinitely for signed values.
               Also note the indirect character MUST be used (i.e. "!@SQ").
            */

            fptr += 2;
            if (IndirectValue)
            {
               if (!*VectorPtr)
               {
                  status = SS$_ACCVIO;
                  break;
               }
               LastNumericValue = *((unsigned long*)(*VectorPtr));
            }
            else
            {
               /* must be indirect */
               status = SS$_BADPARAM;
               break;
            }
            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "!SQ %d\n", LastNumericValue);
#ifdef __ALPHA
            if (fw > 0 && !CommaNumber)
               status = sys$fao (&SQWidthFaoDsc, 0, &StringDsc,
                                 fw, *VectorPtr++);
            else
               status = sys$fao (&SQFaoDsc, 0, &StringDsc,
                                 *VectorPtr++);
#else /* __ALPHA */
            /* kludge, apparently ok all the way up to 53 bits */
            dValue = (double)((unsigned long*)(*VectorPtr))[0];
            dValue += (double)((unsigned long*)(*VectorPtr))[1] * 4294967296.0;
            VectorPtr++;
            status = SS$_NORMAL;
            if (fw > 0 && !CommaNumber)
               sprintf (String, "%*.0f", fw, dValue);
            else
               sprintf (String, "%.0f", dValue);
#endif /* __ALPHA */
            if (VMSnok (status)) break;
            if (!CommaNumber) fw = -1;
            sptr = String;
            /* drop through to buffer the string pointed to by 'sptr' */
         }
         else
         if (*(unsigned short*)fptr == 'SL')
         {
            /*******************/
            /* signed longword */
            /*******************/
   
            fptr += 2;
            if (IndirectValue)
            {
               if (!*VectorPtr)
               {
                  status = SS$_ACCVIO;
                  break;
               }
               LastNumericValue = *(unsigned long*)VectorPtr++;
            }
            else
               LastNumericValue = *VectorPtr++;
            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "!SL %d\n", LastNumericValue);
            if (fw > 0)
               status = sys$fao (&SLWidthFaoDsc, 0, &StringDsc,
                                 fw, LastNumericValue);
            else
               status = sys$fao (&SLFaoDsc, 0, &StringDsc,
                                 LastNumericValue);
            if (VMSnok (status)) break;
            if (!CommaNumber) fw = -1;
            sptr = String;
            /* drop through to buffer the string pointed to by 'sptr' */
         }
         else
         if (*(unsigned short*)fptr == 'IP')
         {
            /*****************************/
            /* dotted-decimal IP address */
            /*****************************/
   
            fptr += 2;
            if (IndirectValue)
            {
               if (!*VectorPtr)
               {
                  status = SS$_ACCVIO;
                  break;
               }
               LastNumericValue = *(unsigned long*)VectorPtr++;
            }
            else
               LastNumericValue = *VectorPtr++;
            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "!SL %d\n", LastNumericValue);
            if (fw > 0)
               status = sys$fao (&IPWidthFaoDsc, 0, &StringDsc,
                                 fw, LastNumericValue & 0x000000ff,
                                 (LastNumericValue & 0x0000ff00) >> 8,
                                 (LastNumericValue & 0x00ff0000) >> 16,
                                 (LastNumericValue & 0xff000000) >> 24);
            else
               status = sys$fao (&IPFaoDsc, 0, &StringDsc,
                                 LastNumericValue & 0x000000ff,
                                 (LastNumericValue & 0x0000ff00) >> 8,
                                 (LastNumericValue & 0x00ff0000) >> 16,
                                 (LastNumericValue & 0xff000000) >> 24);
            if (VMSnok (status)) break;
            fw = -1;
            sptr = String;
            /* drop through to buffer the string pointed to by 'sptr' */
         }
         else
         if (*(unsigned short*)fptr == '%D')
         {
            /*************/
            /* date/time */
            /*************/

            fptr += 2;
            if (IndirectValue)
            {
               if (!*VectorPtr)
               {
                  status = SS$_ACCVIO;
                  break;
               }
               VectorValue = *(unsigned long*)VectorPtr++;
            }
            else
               VectorValue = *VectorPtr++;
            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "!%%D %d\n", VectorValue);

            status = sys$fao (&PDFaoDsc, 0, &StringDsc, VectorValue);
            if (VMSnok (status)) break;

            for (sptr = String; *sptr && *sptr != '-'; sptr++);
            /* absolute times have leading space changed to leading zero */
            if (*sptr && String[0] == ' ')
               (sptr = String)[0] = '0';
            else
            if (*sptr)
               sptr = String;
            else
               /* delta times have leading spaces absorbed */
               for (sptr = String; *sptr && *sptr == ' '; sptr++);
            /* drop through to buffer the string pointed to by 'sptr' */
         }
         else
         if (*(unsigned short*)fptr == '%I')
         {
            /***********************/
            /* numeric to ASCII ID */
            /***********************/

            fptr += 2;
            if (IndirectValue)
            {
               if (!*VectorPtr)
               {
                  status = SS$_ACCVIO;
                  break;
               }
               VectorValue = *(unsigned long*)VectorPtr++;
            }
            else
               VectorValue = *VectorPtr++;
            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "!%%D %d\n", VectorValue);

            status = sys$fao (&PIFaoDsc, 0, &StringDsc, VectorValue);
            if (VMSnok (status)) break;
            sptr = String;
            /* drop through to buffer the string pointed to by 'sptr' */
         }
         else
         if (*(unsigned short*)fptr == '%W')
         {
            /*********************/
            /* weekday/date/time */
            /*********************/

            fptr += 2;
            if (IndirectValue)
            {
               if (!*VectorPtr)
               {
                  status = SS$_ACCVIO;
                  break;
               }
               BinTimePtr = *(unsigned long*)VectorPtr++;
            }
            else
               BinTimePtr = *VectorPtr++;
            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "!%%W %d\n", BinTimePtr);

            if (BinTimePtr == NULL) sys$gettim (BinTimePtr = &BinTime);
            if (!BinTimePtr[0] && !BinTimePtr[1])
               sptr = "(none)";
            else
            if (VMSnok (lib$day_of_week (BinTimePtr, &DayOfWeek)))
               sptr = "*ERROR*1";
            else
            if (VMSnok (sys$fao (&DayDateTimeFaoDsc, 0, &StringDsc,
                DayName[DayOfWeek], fw > 0 ? fw : 23, BinTimePtr)))
               sptr = "*ERROR*2";
            else
               sptr = String;
            /* reset the field-width (only applies to date/time) */
            if (fw > 0) fw = -1;
            /* drop through to buffer the string pointed to by 'sptr' */
         }
         else
         if (*(unsigned short*)fptr == '%%')
         {
            /************************/
            /* nested format string */
            /************************/

            fptr += 2;
            if (idx >= FORMAT_STACK_MAX)
            {
               status = SS$_OVRMAXARG;
               break;
            }
            if (IndirectValue)
            {
               if (!*VectorPtr)
               {
                  status = SS$_ACCVIO;
                  break;
               }
               cptr = *(char*)*VectorPtr++;
            }
            else
               cptr = (char*)*VectorPtr++;
            if (!cptr)
            {
               status = SS$_ACCVIO;
               break;
            }
            /* if empty then just continue with the current format string */
            if (!cptr[0]) continue;
            /* save the current format string pointer */
            FormatStack[idx++] = fptr;
            /* set the new format string pointer */
            fptr = FormatStack[idx] = cptr;
            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "!%%%% idx: %d fptr |%s|\n", idx, fptr);
            continue;
         }
         else
         if (*(unsigned short*)fptr == 'AF' ||
             (VmsFormat = *(unsigned short*)fptr == 'AV') ||
             (PathOdsExtended = VmsFormat =
                 (*(unsigned short*)fptr == 'A5')) ||
             (HtmlEscaped = (*(unsigned short*)fptr == 'HF')) ||
             (HtmlEscaped = VmsFormat = (*(unsigned short*)fptr == 'HV')) ||
             (PathOdsExtended = HtmlEscaped = VmsFormat =
                 (*(unsigned short*)fptr == 'H5')) ||
             (UrlEncoded = (*(unsigned short*)fptr == 'UF')) ||
             (UrlEncoded = VmsFormat = (*(unsigned short*)fptr == 'UV')) ||
             (PathOdsExtended = UrlEncoded = VmsFormat =
                 (*(unsigned short*)fptr == 'U5')))
         {
            /*********************************************/
            /* null-terminated, encoded/escaped filename */
            /*********************************************/

            fptr += 2;
            if (IndirectValue)
            {
               if (!*VectorPtr)
               {
                  status = SS$_ACCVIO;
                  break;
               }
               cptr = *(char*)*VectorPtr++;
            }
            else
               cptr = (char*)*VectorPtr++;
            if (!cptr)
            {
               status = SS$_ACCVIO;
               break;
            }
            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "!%c%c |%s|\n", fptr[0], fptr[1], cptr);

#ifdef ODS_EXTENDED

            /* if it doesn't look ODS-2-like then treat as if extended */
            if (!PathOdsExtended && OdsExtended)
            {
               for (sptr = cptr; *sptr; sptr++)
                  if (!(isupper(*sptr) ||
                        isdigit(*sptr) ||
                        *sptr == '/' ||
                        *sptr == '~' ||
                        *sptr == '.' ||
                        *sptr == '_' ||
                        *sptr == '-' ||
                        *sptr == ':' ||
                        *sptr == '[' ||
                        *sptr == ']' ||
                        *sptr == '$'))
                     break;
               if (*sptr) PathOdsExtended = true;
            }

#else /* ODS_EXTENDED */

            PathOdsExtended = false;

#endif /* ODS_EXTENDED */

            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "PathOdsExtended: %d\n", PathOdsExtended);

            while (*cptr && fw && ofw)
            {
               fw--;
               ofw--;

#ifdef ODS_EXTENDED

               if (PathOdsExtended)
               {
                  if (*cptr == '^' && *(cptr+1))
                  {
                     /* ODS-5 extended specification escape character */
                     cptr++;
                     switch (*cptr)
                     {
                        case '!' :
                        case '#' :
                        case '&' :
                        case '\'' :
                        case '`' :
                        case '(' :
                        case ')' :
                        case '+' :
                        case '@' :
                        case '{' :
                        case '}' :
                        case '.' :
                        case ',' :
                        case ';' :
                        case '[' :
                        case ']' :
                        case '%' :
                        case '^' :
                        case '=' :
                        case ' ' :
                           ch = *cptr;
                           break;
                        case '_' :
                           ch = ' ';
                           break;
                        default :
                           if (isxdigit (ch))
                           {
                              ch = 0;
                              if (*cptr >= '0' && *cptr <= '9')
                                 ch = (*cptr - (int)'0') << 4;
                              else
                              if (tolower(*cptr) >= 'a' &&
                                  tolower(*cptr) <= 'f')
                                 ch = (tolower(*cptr) - (int)'a' + 10) << 4;
                              cptr++;   
                              if (*cptr >= '0' && *cptr <= '9')
                                 ch += (*cptr - (int)'0');
                              else
                              if (tolower(*cptr) >= 'a' &&
                                  tolower(*cptr) <= 'f')
                                 ch += (tolower(*cptr) - (int)'a' + 10);
                              if (*cptr) cptr;
                           }
                           else
                              ch = *cptr;
                     }
                     cptr++;
                  }
                  else
                     ch = *cptr++;
               }
               else

#endif /* ODS_EXTENDED */

               {
                  /* not an extended file specification */
                  if (VmsFormat)
                     ch = *cptr++;
                  else
                     ch = tolower(*cptr++);
               }
   
               if (HtmlEscaped)
               {
                  switch (ch)
                  {
                     case '<' :
                        sptr = "&lt;";
                        break;
                     case '>' :
                        sptr = "&gt;";
                        break;
                     case '&' :
                        sptr = "&amp;";
                        break;
                     case '\"' :
                        sptr = "&quot;";
                        break;
                     default :
                        /* insert this character as-is */
                        if (!brcnt)
                        {
                           /* if writing to storage then it's overflowed */
                           if (BufferPtr != NULL)
                           {
                              BufferPtr[BufferSize-1] = '\0';
                              return (SS$_BUFFEROVF);
                           }
                           /* need another network buffer */
                           NetWriteBufferedInit (rqptr, false);
                           bptr = rqptr->rqOutput.BufferCurrentPtr;
                           brcnt = rqptr->rqOutput.BufferRemaining;
                        }
                        *bptr++ = ch;
                        bcnt++;
                        brcnt--;
                        continue;
                  }
               }
               else
               if (UrlEncoded)
               {
                  if (isalnum(ch) ||
                      ch == '/' ||
                      ch == '-' ||
                      ch == '_' ||
                      ch == '$' ||
                      ch == '*' ||
                      ch == ':' ||
                      ch == '[' ||
                      ch == ']' ||
                      ch == ';' ||
                      ch == '~' ||
                      ch == '.')
                  {
                     /* insert this character as-is */
                     if (!brcnt)
                     {
                        /* if writing to storage then it's overflowed */
                        if (BufferPtr != NULL)
                        {
                           BufferPtr[BufferSize-1] = '\0';
                           return (SS$_BUFFEROVF);
                        }
                        /* need another network buffer */
                        NetWriteBufferedInit (rqptr, false);
                        bptr = rqptr->rqOutput.BufferCurrentPtr;
                        brcnt = rqptr->rqOutput.BufferRemaining;
                     }
                     *bptr++ = ch;
                     bcnt++;
                     brcnt--;
                     continue;
                  }
                  /* encode this character */
                  EncBuffer[1] = HexDigits[(ch & 0xf0) >> 4];
                  EncBuffer[2] = HexDigits[ch & 0x0f];
                  sptr = EncBuffer;
               }
               else
               {
                  /* insert the character as-is */
                  if (!brcnt)
                  {
                     /* if writing to storage then it's overflowed */
                     if (BufferPtr != NULL)
                     {
                        BufferPtr[BufferSize-1] = '\0';
                        return (SS$_BUFFEROVF);
                     }
                     /* need another network buffer */
                     NetWriteBufferedInit (rqptr, false);
                     bptr = rqptr->rqOutput.BufferCurrentPtr;
                     brcnt = rqptr->rqOutput.BufferRemaining;
                  }
                  *bptr++ = ch;
                  bcnt++;
                  brcnt--;
                  continue;
               }

               /* add the escaped/encoded character */
               while (*sptr)
               {
                  if (brcnt--)
                  {
                     *bptr++ = *sptr++;
                     bcnt++; 
                     continue;
                  }
                  /* if writing to storage then it's overflowed */
                  if (BufferPtr != NULL)
                  {
                     BufferPtr[BufferSize-1] = '\0';
                     return (SS$_BUFFEROVF);
                  }
                  /* need another network buffer */
                  NetWriteBufferedInit (rqptr, false);
                  bptr = rqptr->rqOutput.BufferCurrentPtr;
                  brcnt = rqptr->rqOutput.BufferRemaining;
               }
            }
            HtmlEscaped = UrlEncoded = VmsFormat = false;

            /* only if required drop through to space-fill */
            if (fw <= 0) continue;
            /* no string to copy  */
            sptr = NULL;
         }
         else
         if (*(unsigned short*)fptr == '%M' ||
             *(unsigned short*)fptr == '%m')
         {
            /******************/
            /* status message */
            /******************/

            if (*(unsigned short*)fptr == '%M')
               SysGetMsgFlags = 15;
            else
               SysGetMsgFlags = 1;
            fptr += 2;
            if (IndirectValue)
            {
               if (!*VectorPtr)
               {
                  status = SS$_ACCVIO;
                  break;
               }
               VectorValue = *(unsigned long*)VectorPtr++;
            }
            else
               VectorValue = *VectorPtr++;
            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "!%%M %d\n", VectorValue);

            status = sys$getmsg (VectorValue, &Length, &StringDsc,
                                 SysGetMsgFlags, 0);
            if (VMSnok (status))
            {
               ErrorNoticed (status, "sys$getmsg()", FI_LI);
               sptr = "*ERROR* sys$getmsg()";
            }
            else
            {
               (sptr = String)[Length] = '\0';
               if (SysGetMsgFlags == 15) sptr++;
            }
            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "sys$getmsg() |%s|\n", String);
            /* drop through to buffer the string pointed to by 'sptr' */
         }
         else
         if (*(unsigned short*)fptr == 'AC')
         {
            /************************/
            /* counted ASCII string */
            /************************/

            fptr += 2;
            if (IndirectValue)
            {
               if (!*VectorPtr)
               {
                  status = SS$_ACCVIO;
                  break;
               }
               cptr = *(char*)*VectorPtr++;
            }
            else
               cptr = (char*)*VectorPtr++;
            if (!cptr)
            {
               status = SS$_ACCVIO;
               break;
            }
            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "!AC %d |%s|\n", (unsigned int)*cptr, cptr+1);

            /* get the count byte */
            ch = (unsigned int)*cptr++;
            while (ch)
            {
               if (brcnt--)
               {
                  *bptr++ = *cptr++;
                  bcnt++; 
                  ch--;
                  continue;
               }
               /* if writing to storage then it's overflowed */
               if (BufferPtr != NULL)
               {
                  BufferPtr[BufferSize-1] = '\0';
                  return (SS$_BUFFEROVF);
               }
               /* need another network buffer */
               NetWriteBufferedInit (rqptr, false);
               bptr = rqptr->rqOutput.BufferCurrentPtr;
               brcnt = rqptr->rqOutput.BufferRemaining;
            }
            /* only if required drop through to space-fill */
            if (fw <= 0) continue;
            /* no string to copy  */
            sptr = NULL;
         }
         else
         if (*(unsigned short*)fptr == '%T')
         {
            /********/
            /* time */
            /********/

            fptr += 2;
            if (IndirectValue)
            {
               if (!*VectorPtr)
               {
                  status = SS$_ACCVIO;
                  break;
               }
               VectorValue = *(unsigned long*)VectorPtr++;
            }
            else
               VectorValue = *VectorPtr++;
            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "!%%T %d\n", VectorValue);

            status = sys$fao (&PTFaoDsc, 0, &StringDsc, VectorValue);
            if (VMSnok (status)) break;
            sptr = String;
            /* drop through to buffer the string pointed to by 'sptr' */
         }
         else
         if (*(unsigned short*)fptr == '%s' ||
             *(unsigned short*)fptr == '%S')
         {
            /******************/
            /* add 'S' or 's' */
            /******************/

            if (Debug && DEBUG_WRITEFAO) fprintf (stdout, "!%%%c\n", *(fptr+1));
            if (LastNumericValue == 1)
            {
               fptr += 2;
               continue;
            }
            fptr++;
            fw = -1;
            if (*fptr++ == 's')
               sptr = "s";
            else
               sptr = "S";
            /* drop through to buffer the string pointed to by 'sptr' */
         }
         else
         if (*(unsigned short*)fptr == '%?')
         {
            /**********************/
            /* simple conditional */
            /**********************/

            /* e.g. "!%?string if *vector TRUE\rstring if FALSE\r" */
            fptr += 2;
            if (IndirectValue)
            {
               if (!*VectorPtr)
               {
                  status = SS$_ACCVIO;
                  break;
               }
               VectorValue = *(unsigned long*)VectorPtr++;
            }
            else
               VectorValue = *VectorPtr++;
            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "!%%? %d\n", VectorValue);

            if (!VectorValue)
            {
               /* if false skip over the true string */
               while (*fptr && *fptr != '\r') fptr++;
               if (*fptr == '\r') fptr++;
               if (!*fptr) break;
            }
            while (*fptr && *fptr != '\r' && fw && ofw)
            {
               if (brcnt--)
               {
                  *bptr++ = *fptr++;
                  bcnt++; 
                  fw--;
                  ofw--;
                  continue;
               }
               /* if writing to storage then it's overflowed */
               if (BufferPtr != NULL)
               {
                  BufferPtr[BufferSize-1] = '\0';
                  return (SS$_BUFFEROVF);
               }
               /* need another network buffer */
               NetWriteBufferedInit (rqptr, false);
               bptr = rqptr->rqOutput.BufferCurrentPtr;
               brcnt = rqptr->rqOutput.BufferRemaining;
            }
            if (*fptr != '\r')
            {
               /* field-width has expired before the return */
               while (*fptr && *fptr != '\r') fptr++;
            }
            if (*fptr == '\r') fptr++;
            if (VectorValue)
            {
               /* was true, skip over the false string */
               while (*fptr && *fptr != '\r') fptr++;
               if (*fptr == '\r') fptr++;
            }
            /* only if required drop through to space-fill */
            if (fw <= 0) continue;
         }
         else
         if (*fptr == '+')
         {
            /**************************/
            /* discard next parameter */
            /**************************/

            fptr++;
            VectorValue = *VectorPtr++;
            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "!+ %d\n", VectorValue);
         }
         else
         if (*fptr == '-')
         {
            /****************************/
            /* reuse previous parameter */
            /****************************/

            fptr++;
            VectorValue = *--VectorPtr;
            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "!- %d\n", VectorValue);
         }
         else
         if (*fptr == '!')
         {
            /****************************/
            /* an escaped '!' character */
            /****************************/

            if (Debug && DEBUG_WRITEFAO) fprintf (stdout, "!!\n");
            fptr++;
            sptr = "!";
            /* drop through to buffer the string pointed to by 'sptr' */
         }
         else
         if (*fptr == '*')
         {
            /*****************************/
            /* multiple '*'+1 characters */
            /*****************************/

            if (Debug && DEBUG_WRITEFAO) fprintf (stdout, "!*\n");
            if (!*++fptr) fw = 0;
            while (fw > 0 && ofw)
            {
               if (brcnt--)
               {
                  *bptr++ = *fptr;
                  bcnt++; 
                  fw--;
                  ofw--;
                  continue;
               }
               /* if writing to storage then it's overflowed */
               if (BufferPtr != NULL)
               {
                  BufferPtr[BufferSize-1] = '\0';
                  return (SS$_BUFFEROVF);
               }
               /* need another network buffer */
               NetWriteBufferedInit (rqptr, false);
               bptr = rqptr->rqOutput.BufferCurrentPtr;
               brcnt = rqptr->rqOutput.BufferRemaining;
            }
            if (*fptr) fptr++;
            continue;
         }
         else
         if (*fptr == '<')
         {
            /****************************/
            /* begin output-field-width */
            /****************************/

            if (Debug && DEBUG_WRITEFAO) fprintf (stdout, "!<\n");
            fptr++;
            if (ofw >= 0)
            {
               /* can't nest these little blighters! */
               status = SS$_BADPARAM;
               break;
            }
            ofw = fw;
            continue;
         }
         else
         if (*fptr == '>')
         {
            /**************************/
            /* end output-field-width */
            /**************************/

            if (Debug && DEBUG_WRITEFAO) fprintf (stdout, "!>\n");
            fptr++;
            if (ofw < 0)
            {
               /* didn't notice a leading "!<" */
               status = SS$_BADPARAM;
               break;
            }
            /* get (any) current outstanding that needs space-filling */
            fw = ofw;
            /* reset output-field-width to infinite */
            ofw = -1;
            /* only if required drop through to space-fill */
            if (fw <= 0) continue;
         }
         else
         {
            if (Debug && DEBUG_WRITEFAO)
               fprintf (stdout, "!%c%c\n", fptr[0], fptr[1]);
            status = SS$_BADPARAM;
            break;
         }

         /*********************************/
         /* copy a string into the buffer */
         /*********************************/

         if (sptr != NULL && CommaNumber)
         {
            Length = strlen(sptr);
            if (((Length-1) / 3) >= 1)
            {
               if (!(CommaCount = Length % 3))
                  CommaCount = 3;

               while (*sptr && fw && ofw)
               {
                  if (brcnt--)
                  {
                     if (!CommaCount--)
                     {
                        *bptr++ = ',';
                        bcnt++; 
                        fw--;
                        ofw--;
                        CommaCount = 3;
                        continue;
                     }
                     *bptr++ = *sptr++;
                     bcnt++; 
                     fw--;
                     ofw--;
                     continue;
                  }
                  /* if writing to storage then it's overflowed */
                  if (BufferPtr != NULL)
                  {
                     BufferPtr[BufferSize-1] = '\0';
                     return (SS$_BUFFEROVF);
                  }
                  /* need another network buffer */
                  NetWriteBufferedInit (rqptr, false);
                  bptr = rqptr->rqOutput.BufferCurrentPtr;
                  brcnt = rqptr->rqOutput.BufferRemaining;
               }
               /* only if required drop through to space-fill */
               if (fw <= 0) continue;
               sptr = NULL;
            }
            /* drop thru to copy 3 or less digits into the buffer */
         }

         if (sptr != NULL)
         {
            while (*sptr && fw && ofw)
            {
               if (brcnt--)
               {
                  *bptr++ = *sptr++;
                  bcnt++; 
                  fw--;
                  ofw--;
                  continue;
               }
               /* if writing to storage then it's overflowed */
               if (BufferPtr != NULL)
               {
                  BufferPtr[BufferSize-1] = '\0';
                  return (SS$_BUFFEROVF);
               }
               /* need another network buffer */
               NetWriteBufferedInit (rqptr, false);
               bptr = rqptr->rqOutput.BufferCurrentPtr;
               brcnt = rqptr->rqOutput.BufferRemaining;
            }
         }

         /***************************************************/
         /* if positive field-width, right-fill with spaces */
         /***************************************************/

         if (fw <= 0) continue;

         while (fw > 0 && ofw)
         {
            if (brcnt--)
            {
               *bptr++ = ' ';
               bcnt++; 
               fw--;
               ofw--;
               continue;
            }
            /* if writing to storage then it's overflowed */
            if (BufferPtr != NULL)
            {
               BufferPtr[BufferSize-1] = '\0';
               return (SS$_BUFFEROVF);
            }
            /* need another network buffer */
            NetWriteBufferedInit (rqptr, false);
            bptr = rqptr->rqOutput.BufferCurrentPtr;
            brcnt = rqptr->rqOutput.BufferRemaining;
         }
      }

      /*********************/
      /* end format string */
      /*********************/

      if (VMSnok (status)) break;

      /* if there are still pointers on the format stack */
      if (!idx) break;
      idx--;
   }

   /*****************************/
   /* end nested format strings */
   /*****************************/

   if (VMSnok (status))
   {
      /* if an error was detected append from the offending directive */
      fptr = LastFptr;
      while (*fptr)
      {
         if (brcnt--)
         {
            *bptr++ = *fptr++;
            bcnt++; 
            fw--;
            ofw--;
            continue;
         }
         /* if writing to storage then it's overflowed */
         if (BufferPtr != NULL)
         {
            BufferPtr[BufferSize-1] = '\0';
            return (SS$_BUFFEROVF);
         }
         /* need another network buffer */
         NetWriteBufferedInit (rqptr, false);
         bptr = rqptr->rqOutput.BufferCurrentPtr;
         brcnt = rqptr->rqOutput.BufferRemaining;
      }
   }

   if (Debug && DEBUG_WRITEFAO)
      fprintf (stdout, "%d bytes\n", bcnt - PrevBcnt);

   if (BufferPtr == NULL)
   {
      rqptr->rqOutput.BufferCurrentPtr = bptr;
      rqptr->rqOutput.BufferRemaining = brcnt;
      rqptr->rqOutput.BufferCount = bcnt;
   }
   else
   {
      *bptr = '\0';
      if (Debug && DEBUG_WRITEFAO) fprintf (stdout, "|%s|\n", BufferPtr);
   }

   if (LengthPtr != NULL)
   {
      if (bcnt - PrevBcnt > 65535)
      {
         /* indicate that we can't represent what we've written to buffer */
         *LengthPtr = 65535;
         status = SS$_BUFFEROVF;
      }
      else
         *LengthPtr = bcnt - PrevBcnt;
   }

   return (status);
}

/*****************************************************************************/
/*
This is a variable-argument wrapper for WriteFormatted(), see that for greater
detail.  This function just gets all the arguments from the call stack and puts
them into a longword vector that WriteFormatted() can use.  'LengthPtr'
provides storage for returning the length of the generated string.  The
resultant string is always null-terminated (unless an error), hence the
'BufferSize' can always only store 'BufferSize'-1 characters.  'LengthPtr'
provides storage for the number of characters minus the null-termination.
*/

int WriteFao
(
char *BufferPtr,
int BufferSize,
unsigned short *LengthPtr,
char *FormatString,
...
)
{
   int  status,
        argcnt;
   unsigned long  *vecptr;
   unsigned long  FaoVector [128];
   va_list  argptr;

   /*********/
   /* begin */
   /*********/

   va_count (argcnt);

   if (Debug && DEBUG_WRITEFAO) fprintf (stdout, "WriteFao() %d\n", argcnt);

   if (argcnt > 128+4) return (SS$_OVRMAXARG);

   vecptr = FaoVector;
   va_start (argptr, FormatString);
   for (argcnt -= 4; argcnt; argcnt--)
      *vecptr++ = va_arg (argptr, unsigned long);
   va_end (argptr);

   status = WriteFormatted (NULL, BufferPtr, BufferSize, LengthPtr,
                            FormatString, &FaoVector);
   if (Debug && DEBUG_WRITEFAO)
      fprintf (stdout, "WriteFormatted() %%X%08.08X\n", status);
   return (status);
}

/*****************************************************************************/
/*
This is a variable-argument wrapper for WriteFormatted(), see that for greater
detail.  This function just gets all the arguments from the call stack and puts
them into a longword vector that WriteFormatted() can use.

This function has a #define in NET.H making it available as NetWriteFao().
*/

int WriteFaoNet
(
REQUEST_STRUCT *rqptr,
char *FormatString,
...
)
{
   int  status,
        argcnt;
   unsigned long  *vecptr;
   unsigned long  FaoVector [128];
   va_list  argptr;

   /*********/
   /* begin */
   /*********/

   va_count (argcnt);

   if (Debug && DEBUG_WRITEFAO) fprintf (stdout, "WriteFaoNet() %d\n", argcnt);

   if (argcnt > 128+2) return (SS$_OVRMAXARG);

   vecptr = FaoVector;
   va_start (argptr, FormatString);
   for (argcnt -= 2; argcnt; argcnt--)
      *vecptr++ = va_arg (argptr, unsigned long);
   va_end (argptr);

   status = WriteFormatted (rqptr, NULL, 0, NULL, FormatString, &FaoVector);
   if (Debug && DEBUG_WRITEFAO)
      fprintf (stdout, "WriteFormatted() %%X%08.08X\n", status);
   return (status);
}

/****************************************************************************/
/*
WriteFaol()-formatted print statement.  A fixed-size, internal buffer of 986
bytes maximum is used and the result output as an OPCOM message.
*/

int WriteFaoOpcom
(
char *FormatString,
...
)
{
   static $DESCRIPTOR (OpcomDsc, "");

   int  status,
        argcnt;
   unsigned short  Length;
   unsigned long  *vecptr;
   unsigned long  FaoVector [128+2];
   va_list  argptr;
   struct
   {
      unsigned long  TargetType;
      unsigned long  RequestId;
      char  MsgText [986+1];
   } OpcomMsg;

   /*********/
   /* begin */
   /*********/

   va_count (argcnt);

   if (Debug && DEBUG_WRITEFAO)
      fprintf (stdout, "WriteFaoOpcom() |%s| %d\n", FormatString, argcnt);

   if (argcnt > 128+1) return (SS$_OVRMAXARG);

   vecptr = FaoVector;
   *vecptr++ = ProcessName;
   *vecptr++ = FormatString;
   va_start (argptr, FormatString);
   for (argcnt -= 1; argcnt; argcnt--)
      *vecptr++ = va_arg (argptr, unsigned long);
   va_end (argptr);

   status = WriteFaol (OpcomMsg.MsgText, sizeof(OpcomMsg.MsgText), &Length,
                       "Process !AZ reports\r\n!%%", &FaoVector);
   if (VMSnok (status)) ErrorNoticed (status, "WriteFaol()", FI_LI);

   if (Debug && DEBUG_WRITEFAO)
      fprintf (stdout, "%d |%s|\n", Length, OpcomMsg.MsgText);

   OpcomMsg.TargetType = OPC$_RQ_RQST + ((OpcomTarget & 0xffffff) << 8);
   OpcomMsg.RequestId = 0;

   OpcomDsc.dsc$a_pointer = &OpcomMsg;
   OpcomDsc.dsc$w_length = Length + 8;

   status = sys$sndopr (&OpcomDsc, 0);
   if (Debug && DEBUG_WRITEFAO)
      fprintf (stdout, "sys$sndopr() %%X%08.08X\n", status);

   return (status);
}

/****************************************************************************/
/*
WriteFaol()-formatted print statement.  A fixed-size, internal buffer is used
and the result output to the <stdout> stream.
*/

int WriteFaoStdout
(
char *FormatString,
...
)
{
   int  status,
        argcnt;
   unsigned short  Length;
   unsigned long  *vecptr;
   unsigned long  FaoVector [128];
   char  Buffer [2048];
   va_list  argptr;

   /*********/
   /* begin */
   /*********/

   va_count (argcnt);

   if (Debug && DEBUG_WRITEFAO)
      fprintf (stdout, "WriteFaoStdout() |%s| %d\n", FormatString, argcnt);

   if (argcnt > 128+1) return (SS$_OVRMAXARG);

   vecptr = FaoVector;
   va_start (argptr, FormatString);
   for (argcnt -= 1; argcnt; argcnt--)
      *vecptr++ = va_arg (argptr, unsigned long);
   va_end (argptr);

   status = WriteFaol (Buffer, sizeof(Buffer), &Length,
                       FormatString, &FaoVector);
   if (VMSnok (status)) ErrorNoticed (status, "WriteFaol()", FI_LI);

   fputs (Buffer, stdout);

   return (status);
}

/****************************************************************************/

