/*****************************************************************************/
/*
                                Request.c

Get, process and execute the HTTP request from the client.

Some server directives are detected in this module.  Server directives take
the form of a special path or query string and are CASE SENSISTIVE.

  ?http=...                   any query string beginning "httpd=" will be
                              ignored by the default query script and the HTTPd
                              will pass it to the requested functionality

  /httpd/-/admin/             generates a server administration menu
  /httpd/-/admin/graphic/...  generates graphics
  /httpd/-/admin/report/...   generates various reports
  /httpd/-/admin/revise/...   allows form-based configuration
  /httpd/-/admin/control/...  allows server to be stopped, restarted, etc.

  /echo/                      echoes complete request (header and any body)
  /tree/                      directory tree
  /upd/                       activates the update module
  /where/                     reports mapped and parsed path
  /Xray/                      returns an "Xray" of the request (i.e. header
                              and body of response as a plain-text document)

A request for an HTTP/1.0 persistent connection is looked for in the request
header and if found a boolean set to indicate the request.

Error counting within this module attempts to track the reason that any 
connection accepted fails before being passsed to another module for 
processing (i.e. request format error, path forbidden by rule).


AUTHORIZATION MAY BE PERFORMED TWICE!
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
The first, to authorize the full path provided with the request (after it
has been mapped!)  The second, if the request path contained a script any
path specification following the script component is also checked.  In this
way access to controlled information is not inadvertantly allowed because
the script itself is not controlled.  In general, script specifications
should not be part of an authorized path if that script is used to access
other paths (i.e. does not return data generated by itself), apply
authorization to data paths.


VERSION HISTORY
---------------
31-MAR-2002  MGD  keep-alive decision logic to RequestFields()
14-OCT-2001  MGD  add an explicit test for, and message regarding, DECnet
                  use in mapped file names, reporting it as unsupported
11-JUL-2001  MGD  allow '?' on the end of a REDIRECT mapping template
                  to propagate the original request's query string
28-JUN-2001  MGD  extend local redirection syntax to reinstate "reverse proxy"
                  (e.g. "/http://the.host.name/path")
10-MAY-2001  MGD  calls to throttle module
27-FEB-2001  MGD  script path parse content-type check
08-JAN-2001  MGD  bugfix; RequestDiscardAst()
30-DEC-2000  MGD  rework for FILE.C getting file contents in-memory
01-OCT-2000  MGD  authorize either request *or* mapped path
                  (script path authorization is/has always been on mapped)
26-AUG-2000  MGD  WATCH processing, peek, or peek+processing
08-AUG-2000  MGD  bugfix; include Accept-Encoding when redirecting,
                  bugfix; (sort-of) ensure redirected BytesRawRx carried over
24-JUN-2000  MGD  persistant run-time environments,
                  bugfix; HEAD requests specifying content-length
                  bugfix; increase size of buffer in RequestRedirect()
07-MAY-2000  MGD  session track
04-MAR-2000  MGD  use NetWriteFaol(), et.al.,
                  add "http:///" and "https:///" to redirection syntax
08-FEB-2000  MGD  search script exclude specified file types
27-DEC-1999  MGD  support ODS-2 and ODS-5 using ODS module
11-NOV-1999  MGD  allow for "ETag:" (only for proxy propagation)
20-OCT-1999  MGD  redirect now substitutes the scheme and "Host:" or server
                  host:port into a mapping rule like "REDIRECT ///some/path"
                  or just the scheme into "REDIRECT //host.domain/path/"
28-AUG-1999  MGD  accomodation for asynchronous authorization
30-JUL-1999  MGD  bugfix; HttpdExit() requires a parameter!
12-JUN-1999  MGD  looks like a proxy request? send it to ProxyRequestBegin()
04-APR-1999  MGD  provide HTTP/0.9 functionality (finally!)
10-JAN-1999  MGD  proxy serving,
                  history report format refined,
                  added service information to history report
07-NOV-1998  MGD  WATCH facility
18-OCT-1998  MGD  error report redirection
19-SEP-1998  MGD  improve granularity of cache search,
                  RequestFileNoType() now redirects for directories,
                  automatic scripting suppressed with HTTPd query string
12-JUL-1998  MGD  bugfix; RequestEnd() no status returned from RequestBegin()!
14-MAY-1998  MGD  ?httpd=... generalized from index processing to all requests
02-APR-1998  MGD  no longer log internal redirects
28-MAR-1998  MGD  declare an AST for local redirection parsing
                  (in case a CGIplus script redirected and is exiting!)
28-JAN-1998  MGD  moved more appropriate functions from HTTPd.C into here,
                  header parsing now allows for hiatus in end-header blank line,
                  allow for directories specified as "/dir1/dir2"
07-JAN-1998  MGD  provide URL-encoded decode on path
05-OCT-1997  MGD  file cache,
                  added "Accept-Charset:", "Forwarded:" and "Host:"
18-SEP-1997  MGD  HTTP status code mapping
09-AUG-1997  MGD  message database
27-JUL-1997  MGD  modified "Accept:" header lines processing
08-JUN-1997  MGD  added "Pragma:" header field detection
27-MAR-1997  MGD  added "temporary" file detection (for UPD/PUT preview)
01-FEB-1997  MGD  HTTPd version 4
01-OCT-1996  MGD  added more reports
25-JUL-1996  MGD  use optional "length=" within "If-Modified-Since:" header
12-APR-1996  MGD  RMS parse structures moved to thread data;
                  persistent connections ("keep-alive");
                  changed internal directive from query string to path;
                  observed Multinet disconnection/zero-byte behaviour
                  (request now aborts if Multinet returns zero bytes)
01-DEC-1995  MGD  HTTPd version 3
27-SEP-1995  MGD  extensive rework of some functions;
                  added 'Referer:', 'User-Agent:', 'If-Modified-Since:'
01-APR-1995  MGD  initial development for addition to multi-threaded daemon
*/
/*****************************************************************************/

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

/* VMS related header files */
#include <descrip.h>
#include <iodef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application header files */
#include "wasd.h"

#define WASD_MODULE "REQUEST"

/******************/
/* global storage */
/******************/

LIST_HEAD  RequestList;
LIST_HEAD  RequestHistoryList;
int  RequestHistoryCount,
     RequestHistoryMax;

/********************/
/* external storage */
/********************/

#ifdef DBUG
extern BOOL Debug;
#else
#define Debug 0 
#endif

extern BOOL  CacheEnabled,
             ControlExitRequested,
             ControlExitNowRequested,
             ControlRestartRequested,
             ControlRestartNowRequested,
             HttpdTicking,
             LoggingEnabled,
             MonitorEnabled,
             OdsExtended;

extern int  WatchEnabled,
            ConnectCountCurrent,
            ConnectCountConcurrentMax,
            ExitStatus,
            HttpdGblSecPages,
            NetReadBufferSize,
            OpcomMessages,
            OutputBufferSize,
            ProxyServiceCount,
            ServiceCount,
            SsiSizeMax;

extern char  ConfigContentTypeIsMap[],
             ConfigContentTypeMenu[],
             ConfigContentTypeSsi[],
             ConfigContentTypeUrl[],
             ErrorSanityCheck[],
             HttpProtocol[],
             ServerHostPort[],
             SoftwareID[],
             Utility[];

extern ACCOUNTING_STRUCT  *AccountingPtr;
extern CONFIG_STRUCT  Config;
extern HTTPD_GBLSEC  *HttpdGblSecPtr;
extern MSG_STRUCT  Msgs;
extern MAP_RULE_LIST  MappingRulesList;
extern REQUEST_STRUCT  *WatchRequestPtr;
extern SUPERVISOR_LIST  SupervisorListArray[];

/*****************************************************************************/
/*
This function can be called from two sources.  First, NetAccept() where a new
connection has been established and memory has just been allocated for this
connection's thread structure.  Second, from RequestEnd() where a
persistent connection (HTTP/1.0) is involved.

Both then queue a read from the network connection which calls an AST
function when the read completes, to process the request.  The keep-alive
read usually has a far shorter timeout associated with it so that rapid,
sequential reads (as happens when a document is being populated with images
for instance) are serviced from the one connection, but these are not left
open consuming resources on the off-chance of another request in the
not-too-distance future.

Persistent connections are implemented by disposing of all memory allocated
on a request's heap, and all data in the connection thread structure by
zeroing, EXCEPT that above the field 'RetainAboveZeroBelow' (client
connection information), then treating the almost-amnesic structure like a
brand-new connection (only with a shorter timeout period).

Every hour recheck the GMT time offset to crudely detect daylight saving and
other timezone changes.
*/ 

RequestBegin
(
REQUEST_STRUCT *rqptr,
BOOL NewConnection
)
{
   static int  PrevHour = -1;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "RequestBegin()\n");

   if (NewConnection)
   {
      /******************/
      /* new connection */
      /******************/

      /* add entry to the top of the request list */
      ListAddHead (&RequestList, rqptr);

      /* timestamp the transaction */
      sys$gettim (&rqptr->rqTime.Vms64bit);
      sys$numtim (&rqptr->rqTime.VmsVector, &rqptr->rqTime.Vms64bit);
      HttpGmTimeString (rqptr->rqTime.GmDateTime, &rqptr->rqTime.Vms64bit);

      if (PrevHour != rqptr->rqTime.VmsVector[3])
      {
         /* every hour recheck the GMT time offset */
         if (VMSnok (status = TimeSetGMT ()))
            ErrorExitVmsStatus (status, "TimeSetGMT()", FI_LI);
         PrevHour = rqptr->rqTime.VmsVector[3];
      }

      /* if it's not already running kick-off the HTTPd ticker */
      if (!HttpdTicking) HttpdTick (true);

      /* initialize the timer for input */
      HttpdTimerSet (rqptr, TIMER_INPUT, 0);

      /* if available then set any initial report to be via the script */
      if (rqptr->ServicePtr->ErrorReportPath[0])
         rqptr->rqResponse.ErrorReportByRedirect = true;

      if (rqptr->ServicePtr->RequestScheme == SCHEME_HTTPS)
      {
         /* Secure Sockets Layer ("https://") transaction */
         SesolaBegin (rqptr);
         /* if NULL the SSL processing didn't get off the ground! */
         if (rqptr->SesolaPtr == NULL) RequestEnd (rqptr);
         return;
      }
   }
   else
   {
      /*************************/
      /* persistent connection */
      /*************************/

      if (rqptr->WatchItem &&
          (WatchEnabled & WATCH_CONNECT))
         WatchThis (rqptr, FI_LI, WATCH_CONNECT,
                    "KEEP-ALIVE !AZ,!UL",
                    rqptr->rqNet.ClientHostName, rqptr->rqNet.ClientPort); 

      /* zero the portion of the request structure that is not persistant */
      memset ((char*)&rqptr->ZeroedBegin, 0,
              (char*)&rqptr->ZeroedEnd - (char*)&rqptr->ZeroedBegin);

      /*
         Timestamp the keep-alive period for RequestReport()
         Transaction will again be time-stamped if/when request received!
      */
      sys$gettim (&rqptr->rqTime.Vms64bit);

      /* initialize the timer for keep-alive */
      HttpdTimerSet (rqptr, TIMER_KEEPALIVE, 0);

      if (rqptr->ServicePtr->ErrorReportPath[0])
      {
         /* set any initial report to be via a reporting script */
         rqptr->rqResponse.ErrorReportByRedirect = true;
         /* new request, reset (thread-permanent storage) */
         rqptr->RedirectErrorStatusCode = 0;
         rqptr->RedirectErrorAuthRealmDescrPtr = NULL;
      }
   }

   if (Debug)
      fprintf (stdout, "KeepAliveCount: %d\n", rqptr->rqNet.KeepAliveCount);

   rqptr->rqBody.BufferPtr = "";

   rqptr->rqNet.ReadBufferPtr =
      VmGetHeap (rqptr, rqptr->rqNet.ReadBufferSize = NetReadBufferSize);

   NetRead (rqptr, &RequestGet,
            rqptr->rqNet.ReadBufferPtr, rqptr->rqNet.ReadBufferSize);
}

/*****************************************************************************/
/*
This function may be AST-delivery executed ONE OR MORE TIMES, FROM ITSELF, 
before a connection is actually disposed of.  Basically, all ASTs hard-wired or
declared *MUST* deliver back to here!

This is the final request rundown function.  It checks for various tasks and/or
associated structures and calls their finalize function to progressively shut
the request processing down in a hopefully, orderly fashion.
*/

RequestEnd (REQUEST_STRUCT *rqptr)

{
   static long  Addx2 = 2,
                Subx2 = 2,
                EdivTen = 10;

   int  status,
        StatusCodeGroup;
   unsigned long  Remainder,
                  Seconds,
                  SubSeconds;
   unsigned long  BinaryTime [2],
                  QuadScratch [2],
                  ResultTime [2];
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "RequestEnd()\n");

   if (rqptr == WatchRequestPtr) WatchEnd (rqptr);

   if (rqptr->rqPathSet.ThrottleFrom) ThrottleEnd (rqptr);

   if (rqptr->rqNet.ClientChannel &&
       rqptr->BytesRx)
   {
      /* flush anything currently remaining in the thread's output buffer */
      if (rqptr->rqOutput.BufferCount)
      {
         NetWriteFullFlush (rqptr, &RequestEnd);
         return;
      }

      /* if an error has been reported send this to the client */
      if (ERROR_REPORTED (rqptr))
      {
         ErrorSendToClient (rqptr);
         return;
      }

      /* if the response header has not yet been sent then do it */
      if (!rqptr->rqResponse.HeaderSent)
      {
         NetWrite (rqptr, &RequestEnd, NULL, 0);
         return;
      }

      if (rqptr->rqResponse.LocationPtr != NULL &&
          rqptr->rqResponse.LocationPtr[0])
      {
         /***************/
         /* redirection */
         /***************/

         status = RequestRedirect (rqptr);

         if (status == SS$_NORMAL)
         {
            /* SS$_NORMAL is returned when local redirect */
            if (!RequestLineParse (rqptr))
            {
               RequestEnd (rqptr);
               return;
            }
            RequestParseAndExecute (rqptr);
            return;
         }

         /* SS$_NONLOCAL is returned when non-local redirect :^) */
         if (status == SS$_NONLOCAL) return;

         /* error? */
         RequestEnd (rqptr);
         return;
      }
   }

   /* if a proxied request */
   if (rqptr->ProxyTaskPtr != NULL)
   {
      ProxyEnd (rqptr->ProxyTaskPtr);
      return;
   }

   /* if Secure Sockets Layer request */
   if (rqptr->SesolaPtr != NULL)
   {
      SesolaEnd (rqptr);
      return;
   }

   /****************************/
   /* end of possible re-calls */
   /****************************/

   /* response elapsed time */
   if (!rqptr->KeepAliveTimeout)
   {
      sys$gettim (&BinaryTime);
      status = lib$subx (&BinaryTime, &rqptr->rqTime.Vms64bit, 
                         &ResultTime, &Subx2);
      /* convert to microseconds (longword becomes 71 minutes max) */
      if (VMSok (status))
         status = lib$ediv (&EdivTen, &ResultTime,
                            &rqptr->rqResponse.Duration, &Remainder);
      if (Debug) fprintf (stdout, "elapsed: %%X%08.08X\n", status);
      if (VMSnok (status)) rqptr->rqResponse.Duration = 0;

      /* some (double-precision) accounting */
      AccountingPtr->ResponseDurationCount++;
      QuadScratch[0] = rqptr->rqResponse.Duration;
      QuadScratch[1] = 0;
      lib$addx (&QuadScratch, &AccountingPtr->QuadResponseDuration,
                &AccountingPtr->QuadResponseDuration, &Addx2);

      /* only update non-zero minimums */
      if (rqptr->rqResponse.Duration &&
          rqptr->rqResponse.Duration < AccountingPtr->ResponseDurationMin)
         AccountingPtr->ResponseDurationMin = rqptr->rqResponse.Duration;
      if (rqptr->rqResponse.Duration > AccountingPtr->ResponseDurationMax)
         AccountingPtr->ResponseDurationMax = rqptr->rqResponse.Duration;

      /* a little more (double-precision) accounting */
      QuadScratch[0] = rqptr->BytesRawTx;
      QuadScratch[1] = 0;
      lib$addx (&QuadScratch, &AccountingPtr->QuadBytesRawTx,
                &AccountingPtr->QuadBytesRawTx, &Addx2);
      lib$addx (&QuadScratch, &rqptr->ServicePtr->QuadBytesRawTx,
                &rqptr->ServicePtr->QuadBytesRawTx, &Addx2);
      QuadScratch[0] = rqptr->BytesRawRx;
      QuadScratch[1] = 0;
      lib$addx (&QuadScratch, &AccountingPtr->QuadBytesRawRx,
                &AccountingPtr->QuadBytesRawRx, &Addx2);
      lib$addx (&QuadScratch, &rqptr->ServicePtr->QuadBytesRawRx,
                &rqptr->ServicePtr->QuadBytesRawRx, &Addx2);

      rqptr->ServicePtr->ReadErrorCount += rqptr->rqNet.ReadErrorCount;
      rqptr->ServicePtr->WriteErrorCount += rqptr->rqNet.WriteErrorCount;
      AccountingPtr->NetReadErrorCount += rqptr->rqNet.ReadErrorCount;
      AccountingPtr->NetWriteErrorCount += rqptr->rqNet.WriteErrorCount;

      StatusCodeGroup = rqptr->rqResponse.HttpStatus / 100;
      if (StatusCodeGroup < 0 || StatusCodeGroup > 5) StatusCodeGroup = 0;
      AccountingPtr->ResponseStatusCodeCountArray[StatusCodeGroup]++;
      if (rqptr->rqResponse.HttpStatus == 403)
         AccountingPtr->RequestForbiddenCount++;

      if (LoggingEnabled &&
          rqptr->BytesRawRx &&
          rqptr->BytesRawTx &&
          (!rqptr->RedirectErrorStatusCode ||
           (rqptr->RedirectErrorStatusCode &&
            !rqptr->rqResponse.ErrorReportByRedirect)))
         Logging (rqptr, LOGGING_ENTRY);

      /* if a request history is being kept then provide the entry */
      if (RequestHistoryMax) RequestHistory (rqptr);

      /* if keeping activity statistics */
      if (Config.cfMisc.ActivityNumberOfDays) GraphActivityIncrement (rqptr);

      if (rqptr->WatchItem && (WatchEnabled & WATCH_REQUEST))
      {
         Seconds = rqptr->rqResponse.Duration / USECS_IN_A_SECOND;
         SubSeconds = (rqptr->rqResponse.Duration % USECS_IN_A_SECOND) /
                      DURATION_UNITS_USECS;

         if (rqptr->rqNet.ReadErrorCount)
            WatchThis (rqptr, FI_LI, WATCH_REQUEST,
                       "RX errors !UL %X!08XL !-!%M",
                       rqptr->rqNet.ReadErrorCount,
                       rqptr->rqNet.ReadErrorStatus);

         if (rqptr->rqNet.WriteErrorCount)
            WatchThis (rqptr, FI_LI, WATCH_REQUEST,
                       "TX errors !UL %X!08XL !-!%M",
                       rqptr->rqNet.WriteErrorCount,
                       rqptr->rqNet.WriteErrorStatus);

         WatchThis (rqptr, FI_LI, WATCH_REQUEST, 
            "STATUS !3ZL rx:!UL tx:!UL bytes !UL.!#ZL seconds!AZ",
            rqptr->rqResponse.HttpStatus,
            rqptr->BytesRawRx, rqptr->BytesRawTx,
            Seconds, DURATION_DECIMAL_PLACES, SubSeconds,
            rqptr->rqTmr.TerminatedCount ? " TIMED-OUT" : "");
      }
   }

   AccountingPtr->ConnectCurrent = ConnectCountCurrent-1;
   rqptr->ServicePtr->ConnectCount++;

   /* update the global section used by HTTPDMON utility */
   if (MonitorEnabled) RequestGblSecUpdate (rqptr);

   HttpdSupervisorList (rqptr, -1);

   VmFreeHeap (rqptr, FI_LI);

   if (rqptr->rqNet.ClientChannel)
   {
      if (rqptr->KeepAliveRequest &&
          rqptr->KeepAliveResponse)
      {
         /*************************/
         /* PERSISTENT connection */
         /*************************/

         RequestBegin (rqptr, false);
         return;
      }

      NetCloseSocket (rqptr);
   }


   if (rqptr->WatchItem &&
       WatchEnabled == WATCH_ONE_SHOT)
   {
      /* one-shot WATCH request is finished, no need to continue WATCHing */
      NetCloseSocket (WatchRequestPtr);
      WatchEnd (WatchRequestPtr);
   }

   /* remove from the request list */
   ListRemove (&RequestList, rqptr);

   /* free memory allocated for the connection structure */
   VmFreeRequest (rqptr, FI_LI);

   /* one less to worry about! */
   if (ConnectCountCurrent) ConnectCountCurrent--;

   /* don't bother checking for exit or restart if connections still exist */
   if (ConnectCountCurrent) return;

   /* if the control module has requested server exit or restart */
   if (ControlExitRequested)
   {
      fprintf (stdout, "%%%s-I-CONTROL, delayed server exit\n", Utility);
      ExitStatus = SS$_NORMAL;
      HttpdExit (&ExitStatus);
      sys$delprc (0, 0);
   }
   else
   if (ControlRestartRequested)
   {
      fprintf (stdout, "%%%s-I-CONTROL, delayed server restart\n", Utility);
      exit (SS$_NORMAL);
   }

   HttpdCheckPriv ();
}

/****************************************************************************/
/*
The 'rqptr->rqResponse.LocationPtr' is non-NULL indicating redirection.  This 
can be local, indicated by a leading '/', or it can be non-local, indicated
by anything other than a leading '/', usually a full URL including scheme
(e.g. "http:").  If the URL comprises something like "///some/path/or/other"
then the request scheme ("http:", "https:") and "Host:" or service host:port is
substituted in the appropriate position in the URL.  If "//host.domain/path/" 
then just the request scheme is added.  If "http:///path/" or 
"https:"///path/" then the service host:port is added to the redirection URL
(essentially this provides for a change of request scheme).

If the 'rqptr->rqResponse.LocationPtr' has as it's last character a '?' and no
query string itself, and if the request contains a query string then that is
appended to the redirection URL when building the redirected request.

If authorization is enabled then all redirected requests should be returned
to the client in case authorization information needs to be applied to the
new request.  This may involve reformating a local redirection into a
non-local one.  If authentication is not enabled then check both local and
non-local redirections to see if it can be handled locally.

Return normal status to indicate local-redirection (and that the thread
should NOT be disposed of), or an error status to indicate client-handled,
non-local-redirection, or a legitimate error (and that the thread can go).
*/ 

int RequestRedirect (REQUEST_STRUCT *rqptr)

{
   BOOL  IncludeQueryString,
         LocalRedirection;
   int  status,
        ContentCount,
        PortNumber;
   unsigned  short  Length;
   unsigned  long  *vecptr;
   unsigned  long  BytesRawRx;
   unsigned  long  FaoVector [64];
   char  *cptr, *sptr,
         *ContentBufferPtr,
         *HttpMethodNamePtr,
         *NetReadBufferPtr;
   char  Buffer [4096],
         HttpMethodName [32],
         RedirectionUrl [1024];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "RequestRedirect() |%s|\n",
               rqptr->rqResponse.LocationPtr);

   if (rqptr->WatchItem && (WatchEnabled & WATCH_REQUEST))
      WatchThis (rqptr, FI_LI, WATCH_REQUEST,
                 "REDIRECT !AZ", rqptr->rqResponse.LocationPtr);

   if (rqptr->RedirectCount++ > REQUEST_REDIRECTION_MAX)
   {
      rqptr->rqResponse.LocationPtr = NULL;
      rqptr->rqResponse.HttpStatus = 500;
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_REQUEST_REDIRECTION), FI_LI);
      return (STS$K_ERROR);
   }

   rqptr->rqResponse.HttpStatus = 302;

   /*
      A leading space is normally not possible.
      This indicates a specific HTTP method must be used.
      Format is: "<space>METHOD<space><redirection-path>".
   */
   if (rqptr->rqResponse.LocationPtr[0] == ' ')
   {
      HttpMethodNamePtr = sptr = HttpMethodName;
      for (cptr = rqptr->rqResponse.LocationPtr + 1;
           *cptr && *cptr != ' ';
           *sptr++ = *cptr++);
      *sptr = '\0';
      rqptr->rqResponse.LocationPtr = cptr + 1;
   }
   else
      HttpMethodNamePtr = rqptr->rqHeader.MethodName;

   MapUrl_VirtualPath (rqptr->rqHeader.PathInfoPtr,
                       rqptr->rqResponse.LocationPtr,
                       RedirectionUrl, sizeof(RedirectionUrl));

   /* now ignore old value of location redirection pointer */
   rqptr->rqResponse.LocationPtr = NULL;

   for (cptr = RedirectionUrl; *cptr && *cptr != '?'; cptr++);
   if (IncludeQueryString = (*(unsigned short*)cptr == '?\0'))
      if (!rqptr->rqHeader.QueryStringPtr[0])
      {
         IncludeQueryString = false;
         *cptr = '\0';
      }

   if (!memcmp (RedirectionUrl, "///", 3) ||
       !memcmp (RedirectionUrl, "//", 2))
      LocalRedirection = false;
   else
   if (RedirectionUrl[0] == '/')
      LocalRedirection = true;
   else
      LocalRedirection = false;

   if (LocalRedirection)
   {
      /*********************/
      /* local redirection */
      /*********************/

      AccountingPtr->RedirectLocalCount++;

      vecptr = &FaoVector;

      *vecptr++ = HttpMethodNamePtr;
      if (strsame (RedirectionUrl, "/http://", 8) ||
          strsame (RedirectionUrl, "/ftp://", 7))
         *vecptr++ = RedirectionUrl + 1;
      else
         *vecptr++ = RedirectionUrl;
      if (IncludeQueryString)
         *vecptr++ = rqptr->rqHeader.QueryStringPtr;
      else
         *vecptr++ = "";
      *vecptr++ = HttpProtocol;

      if (rqptr->rqHeader.AcceptPtr != NULL)
      {
         *vecptr++ = "Accept: !AZ\r\n";
         *vecptr++ = rqptr->rqHeader.AcceptPtr;
      }
      else
         *vecptr++ = "";

      if (rqptr->rqHeader.AcceptLangPtr != NULL)
      {
         *vecptr++ = "Accept-Language: !AZ\r\n";
         *vecptr++ = rqptr->rqHeader.AcceptLangPtr;
      }
      else
         *vecptr++ = "";

      if (rqptr->rqHeader.AcceptCharsetPtr != NULL)
      {
         *vecptr++ = "Accept-Charset: !AZ\r\n";
         *vecptr++ = rqptr->rqHeader.AcceptCharsetPtr;
      }
      else
         *vecptr++ = "";

      if (rqptr->rqHeader.AcceptEncodingPtr != NULL)
      {
         *vecptr++ = "Accept-Encoding: !AZ\r\n";
         *vecptr++ = rqptr->rqHeader.AcceptEncodingPtr;
      }
      else
         *vecptr++ = "";

      if (rqptr->rqHeader.UserAgentPtr != NULL)
      {
         *vecptr++ = "User-Agent: !AZ\r\n";
         *vecptr++ = rqptr->rqHeader.UserAgentPtr;
      }
      else
         *vecptr++ = "";

      if (rqptr->rqHeader.RefererPtr != NULL)
      {
         *vecptr++ = "Referer: !AZ\r\n";
         *vecptr++ = rqptr->rqHeader.RefererPtr;
      }
      else
         *vecptr++ = "";

      if (rqptr->rqHeader.HostPtr != NULL)
      {
         *vecptr++ = "Host: !AZ\r\n";
         *vecptr++ = rqptr->rqHeader.HostPtr;
      }
      else
         *vecptr++ = "";

      if (rqptr->rqHeader.AuthorizationPtr != NULL)
      {
         *vecptr++ = "Authorization: !AZ\r\n";
         *vecptr++ = rqptr->rqHeader.AuthorizationPtr;
      }
      else
         *vecptr++ = "";

      if (rqptr->rqHeader.IfModifiedSincePtr != NULL)
      {
         *vecptr++ = "If-Modified-Since: !AZ\r\n";
         *vecptr++ = rqptr->rqHeader.IfModifiedSincePtr;
      }
      else
         *vecptr++ = "";

      if (rqptr->rqHeader.PragmaPtr != NULL)
      {
         *vecptr++ = "Pragma: !AZ\r\n";
         *vecptr++ = rqptr->rqHeader.PragmaPtr;
      }
      else
         *vecptr++ = "";

      if (rqptr->rqHeader.CookiePtr != NULL)
      {
         *vecptr++ = "Cookie: !AZ\r\n";
         *vecptr++ = rqptr->rqHeader.CookiePtr;
      }
      else
         *vecptr++ = "";

      if (rqptr->rqHeader.ForwardedPtr != NULL)
      {
         *vecptr++ = "Forwarded: !AZ\r\n";
         *vecptr++ = rqptr->rqHeader.ForwardedPtr;
      }
      else
         *vecptr++ = "";

      if (rqptr->rqHeader.ContentTypePtr != NULL)
      {
         *vecptr++ = "Content-Type: !AZ\r\n";
         *vecptr++ = rqptr->rqHeader.ContentTypePtr;
      }
      else
         *vecptr++ = "";

      if (rqptr->rqHeader.ProxyConnectionPtr)
      {
         *vecptr++ = "Proxy-Connection: !AZ\r\n";
         *vecptr++ = rqptr->rqHeader.ProxyConnectionPtr;
      }
      else
         *vecptr++ = "";

      if (rqptr->rqHeader.ETagPtr != NULL)
      {
         *vecptr++ = "ETag: !AZ\r\n";
         *vecptr++ = rqptr->rqHeader.ETagPtr;
      }
      else
         *vecptr++ = "";

      if (rqptr->rqBody.CurrentCount)
      {
         /* content count is content-length for POST/PUT requests */
         *vecptr++ = "Content-Length: !UL\r\n";
         *vecptr++ = rqptr->rqBody.CurrentCount;
      }
      else
         *vecptr++ = "";

      if (rqptr->KeepAliveRequest)
         *vecptr++ = "Connection: Keep-Alive\r\n";
      else
         *vecptr++ = "";

      if (rqptr->rqHeader.Method == HTTP_METHOD_POST ||
          rqptr->rqHeader.Method == HTTP_METHOD_PUT ||
          rqptr->rqHeader.Method == HTTP_METHOD_DELETE)
      {
         /* for POST/PUT create space for a new request header */
         rqptr->rqNet.ReadBufferPtr = VmGetHeap (rqptr, NetReadBufferSize);
         rqptr->rqNet.ReadBufferSize = NetReadBufferSize;
      }

      status = WriteFaol (rqptr->rqNet.ReadBufferPtr,
                          rqptr->rqNet.ReadBufferSize,
                          &Length,
"!AZ !AZ!AZ !AZ\r\n\
!%%\
!%%\
!%%\
!%%\
!%%\
!%%\
!%%\
!%%\
!%%\
!%%\
!%%\
!%%\
!%%\
!%%\
!%%\
!%%\
!AZ\
\r\n",
         &FaoVector);

      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         ErrorNoticed (status, "WriteFaol()", FI_LI);
         rqptr->rqResponse.LocationPtr = NULL;
         rqptr->rqResponse.ErrorTextPtr = "WriteFaol()";
         rqptr->rqResponse.HttpStatus = 500;
         ErrorVmsStatus (rqptr, status, FI_LI);
         return (STS$K_ERROR);
      }
      if (Debug)
         fprintf (stdout, "%d |%s|\n", Length, rqptr->rqNet.ReadBufferPtr);

      /* buffer anything that is necessary to retain after memset() */
      BytesRawRx = rqptr->BytesRawRx;
      ContentCount = rqptr->rqBody.CurrentCount;
      ContentBufferPtr = rqptr->rqBody.BufferPtr;
      NetReadBufferPtr = rqptr->rqNet.ReadBufferPtr;

      /* zero the portion of the request structure that is not persistant */
      memset ((char*)&rqptr->ZeroedBegin, 0,
              (char*)&rqptr->ZeroedEnd - (char*)&rqptr->ZeroedBegin);

      /* re-timestamp the transaction */
      sys$gettim (&rqptr->rqTime.Vms64bit);
      sys$numtim (&rqptr->rqTime.VmsVector, &rqptr->rqTime.Vms64bit);
      HttpGmTimeString (rqptr->rqTime.GmDateTime, &rqptr->rqTime.Vms64bit);

      /* initialize the timer for input */
      HttpdTimerSet (rqptr, TIMER_INPUT, 0);

      /* if available then set any initial report to be via the script */
      if (rqptr->ServicePtr->ErrorReportPath[0])
         rqptr->rqResponse.ErrorReportByRedirect = true;

      /* restore buffered stuff */
      rqptr->BytesRawRx = BytesRawRx;
      rqptr->rqBody.CurrentCount = ContentCount;
      rqptr->rqBody.BufferPtr = ContentBufferPtr;
      rqptr->rqNet.ReadBufferPtr = NetReadBufferPtr;
      /* (this one is "restored" from global storage!! */
      rqptr->rqNet.ReadBufferSize = NetReadBufferSize;

      /* point the request header at the newly-created contents */
      rqptr->rqHeader.RequestHeaderPtr = rqptr->rqNet.ReadBufferPtr;
      rqptr->rqHeader.RequestHeaderLength = Length;

      /* fudge the bytes received (for log and monitor purposes) */
      rqptr->BytesRx = rqptr->rqHeader.RequestHeaderLength +
                       rqptr->rqBody.CurrentCount;

      /* normal status indicates its a local redirection */
      return (SS$_NORMAL);
   }
   else
   {
      /*************************/
      /* non-local redirection */
      /*************************/

      AccountingPtr->RedirectRemoteCount++;

      vecptr = &FaoVector;

      *vecptr++ = HttpProtocol;
      *vecptr++ = cptr = HttpStatusCodeText(302);

      if (!memcmp (RedirectionUrl, "///", 3))
      {
         /* request scheme and host need to be supplied locally */
         *vecptr++ = "!AZ//!AZ!AZ";
         *vecptr++ = rqptr->ServicePtr->RequestSchemeNamePtr;
         if (rqptr->rqHeader.HostPtr != NULL && rqptr->rqHeader.HostPtr[0])
            *vecptr++ = rqptr->rqHeader.HostPtr;
         else
            *vecptr++ = rqptr->ServicePtr->ServerHostPort;
         *vecptr++ = RedirectionUrl + 2;
      }
      else
      if (!memcmp (RedirectionUrl, "//", 2))
      {
         /* request scheme needs to be supplied locally */
         *vecptr++ = "!AZ!AZ";
         *vecptr++ = rqptr->ServicePtr->RequestSchemeNamePtr;
         *vecptr++ = RedirectionUrl;
      }
      else
      if (!memcmp (RedirectionUrl, "http:///", 8))
      {
         /* host needs to be supplied locally */
         *vecptr++ = "http://!AZ!AZ";
         if (rqptr->rqHeader.HostPtr != NULL && rqptr->rqHeader.HostPtr[0])
            *vecptr++ = rqptr->rqHeader.HostPtr;
         else
            *vecptr++ = rqptr->ServicePtr->ServerHostPort;
         *vecptr++ = RedirectionUrl + 7;
      }
      else
      if (!memcmp (RedirectionUrl, "https:///", 9))
      {
         /* host needs to be supplied locally */
         *vecptr++ = "https://!AZ!AZ";
         if (rqptr->rqHeader.HostPtr != NULL && rqptr->rqHeader.HostPtr[0])
            *vecptr++ = rqptr->rqHeader.HostPtr;
         else
            *vecptr++ = rqptr->ServicePtr->ServerHostPort;
         *vecptr++ = RedirectionUrl + 8;
      }
      else
      {
         /* redirection contains full URL */
         *vecptr++ = "!AZ";
         *vecptr++ = RedirectionUrl;
      }

      if (IncludeQueryString)
         *vecptr++ = rqptr->rqHeader.QueryStringPtr;
      else
         *vecptr++ = "";

      *vecptr++ = SoftwareID;
      *vecptr++ = rqptr->rqTime.GmDateTime;

      status = WriteFaol (Buffer, sizeof(Buffer), &Length,
"!AZ 302 !AZ\r\n\
Location: !%%!AZ\r\n\
Server: !AZ\r\n\
Date: !AZ\r\n\
Content-Type: text/html\r\n\
\r\n",
         &FaoVector);

      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         ErrorNoticed (status, "WriteFaol()", FI_LI);
         rqptr->rqResponse.LocationPtr = NULL;
         rqptr->rqResponse.ErrorTextPtr = "WriteFaol()";
         rqptr->rqResponse.HttpStatus = 500;
         ErrorVmsStatus (rqptr, status, FI_LI);
         return (STS$K_ERROR);
      }

      rqptr->rqResponse.HeaderSent = false;
      rqptr->rqResponse.HeaderLength = Length;
      rqptr->rqResponse.HeaderPtr = VmGetHeap (rqptr, Length+1);
      memcpy (rqptr->rqResponse.HeaderPtr, Buffer, Length+1);

      vecptr = &FaoVector;
      *vecptr++ = cptr;

      if (!memcmp (RedirectionUrl, "///", 3))
      {
         *vecptr++ = "<A HREF=\"!AZ!AZ!AZ!AZ!AZ\">!AZ!AZ!AZ!AZ!AZ</A>";
         *vecptr++ = rqptr->ServicePtr->RequestSchemeNamePtr;
         *vecptr++ = "//";
         if (rqptr->rqHeader.HostPtr != NULL && rqptr->rqHeader.HostPtr[0])
            *vecptr++ = rqptr->rqHeader.HostPtr;
         else
            *vecptr++ = rqptr->ServicePtr->ServerHostPort;
         *vecptr++ = RedirectionUrl + 2;
         if (IncludeQueryString)
            *vecptr++ = rqptr->rqHeader.QueryStringPtr;
         else
            *vecptr++ = "";
         *vecptr++ = rqptr->ServicePtr->RequestSchemeNamePtr;
         *vecptr++ = "//";
         if (rqptr->rqHeader.HostPtr != NULL && rqptr->rqHeader.HostPtr[0])
            *vecptr++ = rqptr->rqHeader.HostPtr;
         else
            *vecptr++ = rqptr->ServicePtr->ServerHostPort;
         *vecptr++ = RedirectionUrl + 2;
         if (IncludeQueryString)
            *vecptr++ = rqptr->rqHeader.QueryStringPtr;
         else
            *vecptr++ = "";
      }
      else
      if (!memcmp (RedirectionUrl, "//", 2))
      {
         *vecptr++ = "<A HREF=\"!AZ!AZ!AZ\">!AZ!AZ!AZ</A>";
         *vecptr++ = rqptr->ServicePtr->RequestSchemeNamePtr;
         *vecptr++ = RedirectionUrl;
         if (IncludeQueryString)
            *vecptr++ = rqptr->rqHeader.QueryStringPtr;
         else
            *vecptr++ = "";
         *vecptr++ = rqptr->ServicePtr->RequestSchemeNamePtr;
         *vecptr++ = RedirectionUrl;
         if (IncludeQueryString)
            *vecptr++ = rqptr->rqHeader.QueryStringPtr;
         else
            *vecptr++ = "";
      }
      else
      if (!memcmp (RedirectionUrl, "http:///", 8))
      {
         *vecptr++ = "<A HREF=\"http://!AZ!AZ!AZ\">http://!AZ!AZ!AZ</A>";
         if (rqptr->rqHeader.HostPtr != NULL && rqptr->rqHeader.HostPtr[0])
            *vecptr++ = rqptr->rqHeader.HostPtr;
         else
            *vecptr++ = rqptr->ServicePtr->ServerHostPort;
         *vecptr++ = RedirectionUrl+7;
         if (IncludeQueryString)
            *vecptr++ = rqptr->rqHeader.QueryStringPtr;
         else
            *vecptr++ = "";
         if (rqptr->rqHeader.HostPtr != NULL && rqptr->rqHeader.HostPtr[0])
            *vecptr++ = rqptr->rqHeader.HostPtr;
         else
            *vecptr++ = rqptr->ServicePtr->ServerHostPort;
         *vecptr++ = RedirectionUrl+7;
         if (IncludeQueryString)
            *vecptr++ = rqptr->rqHeader.QueryStringPtr;
         else
            *vecptr++ = "";
      }
      else
      if (!memcmp (RedirectionUrl, "https:///", 9))
      {
         *vecptr++ = "<A HREF=\"https://!AZ!AZ!AZ\">https://!AZ!AZ!AZ</A>";
         if (rqptr->rqHeader.HostPtr != NULL && rqptr->rqHeader.HostPtr[0])
            *vecptr++ = rqptr->rqHeader.HostPtr;
         else
            *vecptr++ = rqptr->ServicePtr->ServerHostPort;
         *vecptr++ = RedirectionUrl+8;
         if (IncludeQueryString)
            *vecptr++ = rqptr->rqHeader.QueryStringPtr;
         else
            *vecptr++ = "";
         if (rqptr->rqHeader.HostPtr != NULL && rqptr->rqHeader.HostPtr[0])
            *vecptr++ = rqptr->rqHeader.HostPtr;
         else
            *vecptr++ = rqptr->ServicePtr->ServerHostPort;
         *vecptr++ = RedirectionUrl+8;
         if (IncludeQueryString)
            *vecptr++ = rqptr->rqHeader.QueryStringPtr;
         else
            *vecptr++ = "";
      }
      else
      if (IncludeQueryString)
      {
         *vecptr++ = "<A HREF=\"!AZ!AZ\">!AZ!AZ</A>";
         *vecptr++ = RedirectionUrl;
         *vecptr++ = rqptr->rqHeader.QueryStringPtr;
         *vecptr++ = RedirectionUrl;
         *vecptr++ = rqptr->rqHeader.QueryStringPtr;
      }
      else
      {
         *vecptr++ = "<A HREF=\"!AZ\">!AZ</A>";
         *vecptr++ = RedirectionUrl;
         *vecptr++ = RedirectionUrl;
      }

      status = NetWriteFaol (rqptr, "!AZ !%%\n", &FaoVector);
      if (VMSnok (status))
      {
         ErrorNoticed (status, "NetWriteFaol()", FI_LI);
         rqptr->rqResponse.LocationPtr = NULL;
         rqptr->rqResponse.ErrorTextPtr = "NetWriteFaol()";
         rqptr->rqResponse.HttpStatus = 500;
         ErrorVmsStatus (rqptr, status, FI_LI);
         return (STS$K_ERROR);
      }

      RequestEnd (rqptr);

      /* indicate it's non-local redirection */
      return (SS$_NONLOCAL);
   }
}

/*****************************************************************************/
/*
Process first packet read from the client.  In most cases this will contain 
the complete HTTP header and it can be processed without building up a 
buffered header.  If it does not contain the full header allocate heap memory
to contain the current packet and queue subsequent read(s), expanding the 
request header heap memory, until all is available (or it becomes completely 
rediculous!).
*/ 
 
RequestGet (REQUEST_STRUCT *rqptr)

{
   int  cnlcnt, cnt, status;
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout,
"RequestGet() Status: %%X%08.08X Count: %d BufferSize: %d HeaderLength: %d\n",
               rqptr->rqNet.ReadIOsb.Status,
               rqptr->rqNet.ReadIOsb.Count,
               rqptr->rqNet.ReadBufferSize,
               rqptr->rqHeader.RequestHeaderLength);

   if (VMSnok (rqptr->rqNet.ReadIOsb.Status))
   {
      /*
         Most common cause will be broken connection.  Most common reason
         for the break ... "keep-alive" timeout on persistent connection.
         Don't bother to report!
      */
      if (rqptr->rqNet.KeepAliveCount) rqptr->KeepAliveTimeout = true;
      NetCloseSocket (rqptr);
      RequestEnd (rqptr);
      return;
   }

   if (rqptr->rqNet.KeepAliveCount)
   {
      /* request is underway, reinitialize for input (longer period) */
      HttpdTimerSet (rqptr, TIMER_INPUT, 0);

      /* re-timestamp the kept-alive transaction */
      sys$gettim (&rqptr->rqTime.Vms64bit);
      sys$numtim (&rqptr->rqTime.VmsVector, &rqptr->rqTime.Vms64bit);
      HttpGmTimeString (rqptr->rqTime.GmDateTime, &rqptr->rqTime.Vms64bit);
   }

   rqptr->BytesRx += (cnt = rqptr->rqNet.ReadIOsb.Count);

   if (rqptr->rqNet.ReadBufferSize >= MAX_REQUEST_HEADER)
   {
      /* not likely to be a legitimate HTTP request! */
      RequestEnd (rqptr);
      return;
   }

   if (rqptr->rqHeader.RequestHeaderPtr == NULL)
   {
      rqptr->rqHeader.RequestHeaderPtr = rqptr->rqNet.ReadBufferPtr;
      rqptr->rqHeader.RequestHeaderLength = rqptr->rqHeader.RequestLineLength = 0;
   }

   rqptr->rqHeader.RequestHeaderPtr[cnt] = '\0';
   rqptr->rqHeader.RequestHeaderLength += cnt;
   if (Debug)
      fprintf (stdout, "%d |%s|\n",
               rqptr->rqHeader.RequestHeaderLength, rqptr->rqHeader.RequestHeaderPtr); 

   /* looks like an HTTP request? (not SSL on the wrong port for instance) */
   if (rqptr->rqHeader.RequestHeaderLength >= 7)
   {
      /* evaluated in order of probable occurance */
      if (memcmp (rqptr->rqNet.ReadBufferPtr, "GET ", 4) &&
          memcmp (rqptr->rqNet.ReadBufferPtr, "HEAD ", 5) &&
          memcmp (rqptr->rqNet.ReadBufferPtr, "POST ", 5) &&
          memcmp (rqptr->rqNet.ReadBufferPtr, "CONNECT ", 8) &&
          memcmp (rqptr->rqNet.ReadBufferPtr, "PUT ", 4) &&
          memcmp (rqptr->rqNet.ReadBufferPtr, "DELETE ", 7))
      {
         AccountingPtr->RequestErrorCount++;
         rqptr->rqResponse.HttpStatus = 501;
         ErrorGeneral (rqptr, MsgFor(rqptr,MSG_REQUEST_METHOD), FI_LI);
         RequestEnd (rqptr);
         return;
      }
   }

   if (!rqptr->rqHeader.HttpVersion)
   {
      /**************************/
      /* parse the request line */
      /**************************/

      cptr = rqptr->rqNet.ReadBufferPtr;
      while (*cptr && *cptr != '\n') cptr++;
      if (*cptr == '\n')
      {
         if (!RequestLineParse (rqptr))
         {
            RequestEnd (rqptr);
            return;
         }
      }
   }

   if (rqptr->rqHeader.HttpVersion == HTTP_VERSION_0_9)
   {
      /************/
      /* HTTP/0.9 */
      /************/

      rqptr->rqHeader.RequestHeaderPtr = rqptr->rqNet.ReadBufferPtr;
      rqptr->rqHeader.RequestHeaderLength = cptr - (char*)rqptr->rqNet.ReadBufferPtr;
      RequestParseAndExecute (rqptr);
      return;
   }
   else
   if (rqptr->rqHeader.HttpVersion)
   {
      /************/
      /* HTTP/1.n */
      /************/

      /* look for the end of header blank line */
      cnlcnt = rqptr->rqHeader.ConsecutiveNewLineCount;
      cptr = rqptr->rqHeader.RequestHeaderPtr;
      if (Debug) fprintf (stdout, "cnlcnt: %d *cptr: %d\n", cnlcnt, *cptr);
      while (*cptr && cnlcnt < 2)
      {
         if (*cptr == '\r') cptr++;
         if (!*cptr) break;
         if (*cptr != '\n')
         {
            cptr++;
            cnlcnt = 0;
            continue;
         }
         cptr++;
         cnlcnt++;
      }
      if (Debug) fprintf (stdout, "cnlcnt: %d\n", cnlcnt);

      if (cnlcnt >= 2)
      {
         rqptr->rqHeader.RequestHeaderPtr = rqptr->rqNet.ReadBufferPtr;
         rqptr->rqHeader.RequestHeaderLength = cptr - (char*)rqptr->rqNet.ReadBufferPtr;
         RequestParseAndExecute (rqptr);
         return;
      }

      rqptr->rqHeader.ConsecutiveNewLineCount = cnlcnt;
   }

   /*****************************/
   /* read more from the client */
   /*****************************/

   if (rqptr->rqHeader.RequestHeaderLength >= rqptr->rqNet.ReadBufferSize)
   {
      /* need more buffer space, allow one for termination of the buffer */
      rqptr->rqNet.ReadBufferSize += NetReadBufferSize;
      rqptr->rqNet.ReadBufferPtr =
         VmReallocHeap (rqptr, rqptr->rqNet.ReadBufferPtr,
                        rqptr->rqNet.ReadBufferSize+1, FI_LI);
   }

   /* because a realloc() may move the memory reset the pointers */
   rqptr->rqHeader.RequestHeaderPtr = rqptr->rqNet.ReadBufferPtr +
                             rqptr->rqHeader.RequestHeaderLength;

   /* asynchronous read to get more of the header from the client */
   NetRead (rqptr, &RequestGet, rqptr->rqHeader.RequestHeaderPtr,
            rqptr->rqNet.ReadBufferSize - rqptr->rqHeader.RequestHeaderLength);
}

/****************************************************************************/
/*
Get the method, path and query string, and some header field line values, then
execute the request.  This function can be called from two points.  First, from
the function RequestGet(), after it has received a complete HTTP header from
the client.  Second, from the function RequestRedirect(), where an HTTP header
has been reconstructed in the 'NetReadBufferPtr' in order to effect a local
redirection.
*/

RequestParseAndExecute (REQUEST_STRUCT *rqptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout, "RequestParseAndExecute() size: %d\n",
               rqptr->rqHeader.RequestHeaderLength);
/**
      fprintf (stdout, "|%.*s|\n",
               rqptr->rqHeader.RequestHeaderLength,
               rqptr->rqHeader.RequestHeaderPtr);
**/
   }

   if (rqptr->WatchItem && (WatchEnabled & WATCH_REQUEST_HEADER))
   {
      WatchThis (rqptr, FI_LI, WATCH_REQUEST_HEADER,
                 "HEADER !UL bytes", rqptr->rqHeader.RequestHeaderLength);
      WatchData (rqptr->rqHeader.RequestHeaderPtr,
                 rqptr->rqHeader.RequestHeaderLength);
   }

   AccountingPtr->RequestTotalCount++;
   AccountingPtr->RequestParseCount++;
   if (rqptr->rqNet.KeepAliveCount)
      AccountingPtr->RequestKeepAliveCount++;
   if (rqptr->rqNet.KeepAliveCount > AccountingPtr->RequestKeepAliveMax)
      AccountingPtr->RequestKeepAliveMax = rqptr->rqNet.KeepAliveCount;  

   if (!RequestFields (rqptr))
   {
      RequestEnd (rqptr);
      return;
   }

   if (rqptr->ServicePtr->TrackEnabled)
      if (!TrackGetCookie (rqptr))
         TrackSetCookie (rqptr);

   if (rqptr->WatchItem)
   {
      if (!rqptr->rqHeader.PathInfoPtr[0] ||
          rqptr->rqHeader.PathInfoPtr[0] == '/')
      {
         /* check if this non-proxy request should be filtered out on path */
         if (!WatchFilter (NULL, NULL, NULL, NULL,
                           rqptr->rqHeader.PathInfoPtr, rqptr->TrackId))
         {
            WatchThis (rqptr, FI_LI, WATCH_FILTER,
               "PATH/TRACK does not match filter, dropping from WATCH");
            rqptr->WatchItem = 0;
         }
      }
      else
      {
         /* check if this proxy request should be filtered out on path */
         if (!WatchFilter (NULL, NULL, NULL, NULL,
                           rqptr->rqHeader.RequestUriPtr, rqptr->TrackId))
         {
            WatchThis (rqptr, FI_LI, WATCH_FILTER,
               "PATH (proxy) does not match filter, dropping from WATCH");
            rqptr->WatchItem = 0;
         }
      }
   }

   /* evaluated in order of probable occurance */
   if (!strcmp (rqptr->rqHeader.MethodName, "GET"))
   {
      AccountingPtr->MethodGetCount++;
      rqptr->rqHeader.Method = HTTP_METHOD_GET;
      RequestExecute (rqptr);
      return;
   }

   /* simple requests (HTTP/0.9) are limited to GET requests */
   if (rqptr->rqHeader.HttpVersion == HTTP_VERSION_0_9)
   {
      AccountingPtr->RequestErrorCount++;
      rqptr->rqResponse.HttpStatus = 501;
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_REQUEST_METHOD), FI_LI);
      RequestEnd (rqptr);
      return;
   }

   if (!strcmp (rqptr->rqHeader.MethodName, "HEAD"))
   {
      AccountingPtr->MethodHeadCount++;
      rqptr->rqHeader.Method = HTTP_METHOD_HEAD;
      RequestExecute (rqptr);
      return;
   }
   if (!strcmp (rqptr->rqHeader.MethodName, "POST"))
   {
      AccountingPtr->MethodPostCount++;
      rqptr->rqHeader.Method = HTTP_METHOD_POST;
      rqptr->KeepAliveRequest = false;
      RequestBodyReadBegin (rqptr);    
      return;
   }
   if (!strcmp (rqptr->rqHeader.MethodName, "PUT"))
   {
      AccountingPtr->MethodPutCount++;
      rqptr->rqHeader.Method = HTTP_METHOD_PUT;
      rqptr->KeepAliveRequest = false;
      RequestBodyReadBegin (rqptr);
      return;
   }
   if (!strcmp (rqptr->rqHeader.MethodName, "CONNECT"))
   {
      AccountingPtr->MethodPutCount++;
      rqptr->rqHeader.Method = HTTP_METHOD_CONNECT;
      rqptr->KeepAliveRequest = false;
      RequestExecute (rqptr);
      return;
   }
   if (!strcmp (rqptr->rqHeader.MethodName, "DELETE"))
   {
      AccountingPtr->MethodDeleteCount++;
      rqptr->rqHeader.Method = HTTP_METHOD_DELETE;
      rqptr->KeepAliveRequest = false;
      RequestExecute (rqptr);
      return;
   }

   AccountingPtr->RequestErrorCount++;
   rqptr->rqResponse.HttpStatus = 501;
   ErrorGeneral (rqptr, MsgFor(rqptr,MSG_REQUEST_METHOD), FI_LI);
   RequestEnd (rqptr);
}

/*****************************************************************************/
/*
Read the body of a request (POST, PUT, etc., methods) into a buffer of heap
memory pointed to by 'rqptr->rqBody.BufferPtr'.
*/ 
 
RequestBodyReadBegin (REQUEST_STRUCT *rqptr)

{
   int  status,
        ReadSize;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "RequestBodyReadBegin() %d bytes\n",
               rqptr->rqHeader.ContentLength);

   if (rqptr->rqBody.CurrentCount)
   {
      /********************************/
      /* body exists from redirection */
      /********************************/

      RequestExecute (rqptr);
      return;
   }

   /****************************************/
   /* request body allowed to be this big? */
   /****************************************/

   if (Config.cfMisc.PutMaxKbytes &&
       rqptr->rqHeader.ContentLength / 1000 > Config.cfMisc.PutMaxKbytes)
   {
      /* local storage */
      char  String [128];
      $DESCRIPTOR (MaxLengthFaoDsc, "");
      $DESCRIPTOR (StringDsc, String);

      MaxLengthFaoDsc.dsc$a_pointer = MsgFor(rqptr,MSG_REQUEST_BODY_MAX);
      MaxLengthFaoDsc.dsc$w_length = strlen(MaxLengthFaoDsc.dsc$a_pointer);
      status = sys$fao (&MaxLengthFaoDsc, 0, &StringDsc,
                        rqptr->rqHeader.MethodName,
                        rqptr->rqHeader.ContentLength / 1000,
                        Config.cfMisc.PutMaxKbytes);
      if (VMSnok (status))
      {
         rqptr->rqResponse.ErrorTextPtr = "sys$fao()";
         rqptr->rqResponse.ErrorOtherTextPtr = "(possible msg problem)";
         ErrorVmsStatus (rqptr, RMS$_FNF, FI_LI);
      }
      else
      {
         rqptr->rqResponse.HttpStatus = 400;
         ErrorGeneral (rqptr, String, FI_LI);
      }

      /*
         Navigator 3 and MSIE 3 (all browsers?) spit if a POST is not read
         before returning an error, hence continue on if not an unreasonable
         number of bytes!  Disable this by defining NO_UPLOAD_GRACE.
         100 Mbytes doesn't seem an unreasonable limit!
      */

#ifndef NO_UPLOAD_GRACE
      if (rqptr->rqHeader.ContentLength < 100000000)
      {
         /* fake these for the discard AST routine checking */
         rqptr->rqNet.ReadIOsb.Status = SS$_NORMAL;
         rqptr->rqNet.ReadIOsb.Count = 1;
         RequestDiscardAst (rqptr);
         return;
      }
      else
#endif /* NO_UPLOAD_GRACE */
         RequestEnd (rqptr);

      return;
   }

   /****************************/
   /* allocate buffer for body */
   /****************************/

   /*
      We're going to stick an extra null char on the end
      so we can parse it as if it's a string if we want to!
   */
   rqptr->rqBody.BufferPtr = VmGetHeap (rqptr, rqptr->rqHeader.ContentLength+1);
   /* might as well do the above now! */
   rqptr->rqBody.BufferPtr[rqptr->rqHeader.ContentLength] = '\0';

   /*  allow for any content received along with the header */
   rqptr->rqBody.CurrentCount = rqptr->BytesRx -
                                rqptr->rqHeader.RequestHeaderLength;
   /* ensure this does not exceed the specified content length */
   if (rqptr->rqBody.CurrentCount > rqptr->rqHeader.ContentLength)
      rqptr->rqBody.CurrentCount = rqptr->rqHeader.ContentLength;

   if (Debug)
      fprintf (stdout,
"ContentLength: %d BytesRx: %d RequestHeaderLength: %d ContentCount: %d\n",
          rqptr->rqHeader.ContentLength, rqptr->BytesRx,
          rqptr->rqHeader.RequestHeaderLength, rqptr->rqBody.CurrentCount);

   if (rqptr->rqBody.CurrentCount)
   {
      /* copy into the buffer any content received along with the header */
      memcpy (rqptr->rqBody.BufferPtr,
              rqptr->rqNet.ReadBufferPtr + rqptr->rqHeader.RequestHeaderLength,
              rqptr->rqBody.CurrentCount);
   }

   if (rqptr->rqBody.CurrentCount >= rqptr->rqHeader.ContentLength)
   {
      /******************************************************/
      /* looks like it all arrived with the request header! */
      /******************************************************/

      RequestExecute (rqptr);
      return;
   }

   /*******************************/
   /* read (more) from the client */
   /*******************************/

   rqptr->rqBody.CurrentPtr = rqptr->rqBody.BufferPtr +
                              rqptr->rqBody.CurrentCount;
   if (rqptr->rqHeader.ContentLength - rqptr->rqBody.CurrentCount > 32767)
      ReadSize = 32767;
   else
      ReadSize = rqptr->rqHeader.ContentLength - rqptr->rqBody.CurrentCount;

   NetRead (rqptr, &RequestBodyReadAst, rqptr->rqBody.CurrentPtr, ReadSize);
}

/*****************************************************************************/
/*
A read from the client directly into the content buffer has completed.  Check
the status of that read.  If ok check whether all of the content body has been
read.  If not read more of the content body, otherwise declare an AST to
execute the next task.
*/ 

RequestBodyReadAst (REQUEST_STRUCT *rqptr)

{
   int  status,
        ReadSize;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout,
         "RequestBodyReadAst() NetReadIOsb.Status %%X%08.08X .Count %d\n",
         rqptr->rqNet.ReadIOsb.Status, rqptr->rqNet.ReadIOsb.Count);

   if (VMSnok (rqptr->rqNet.ReadIOsb.Status))
   {
      rqptr->rqResponse.ErrorTextPtr = MsgFor(rqptr,MSG_REQUEST_BODY_READ);
      ErrorVmsStatus (rqptr, rqptr->rqNet.ReadIOsb.Status, FI_LI);
      RequestEnd (rqptr);
      return;
   }

   rqptr->BytesRx += rqptr->rqNet.ReadIOsb.Count;
   rqptr->rqBody.CurrentCount += rqptr->rqNet.ReadIOsb.Count;

   if (Debug)
   {
      fprintf (stdout, "ContentLength: %d ContentCount: %d\n",
               rqptr->rqHeader.ContentLength, rqptr->rqBody.CurrentCount);
      /** if (Debug) fprintf (stdout, "|%s|\n", rqptr->rqBody.CurrentPtr); **/
   }

   if (rqptr->rqBody.CurrentCount >= rqptr->rqHeader.ContentLength)
   {
      /*******************************/
      /* looks like it's all arrived */
      /*******************************/

      /* ensure this does not exceed the specified content length */
      if (rqptr->rqBody.CurrentCount > rqptr->rqHeader.ContentLength)
         rqptr->rqBody.CurrentCount = rqptr->rqHeader.ContentLength;

      RequestExecute (rqptr);
      return;
   }

   /*****************************/
   /* read more from the client */
   /*****************************/

   rqptr->rqBody.CurrentPtr = rqptr->rqBody.BufferPtr +
                              rqptr->rqBody.CurrentCount;
   if (rqptr->rqHeader.ContentLength - rqptr->rqBody.CurrentCount > 32767)
      ReadSize = 32767;
   else
      ReadSize = rqptr->rqHeader.ContentLength - rqptr->rqBody.CurrentCount;

   NetRead (rqptr, &RequestBodyReadAst, rqptr->rqBody.CurrentPtr, ReadSize);
}

/*****************************************************************************/
/*
Throw-away anything read from the client.
*/ 

RequestDiscardAst (REQUEST_STRUCT *rqptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout,
"RequestDiscardAst() NetReadIOsb.Status %%X%08.08X .Count %d Length: %d\n",
         rqptr->rqNet.ReadIOsb.Status, rqptr->rqNet.ReadIOsb.Count,
         rqptr->rqHeader.ContentLength);

   if (VMSnok (rqptr->rqNet.ReadIOsb.Status))
   {
      RequestEnd (rqptr);
      return;
   }

   rqptr->rqHeader.ContentLength -= rqptr->rqNet.ReadIOsb.Count;
   if (rqptr->rqHeader.ContentLength <= 0)
   {
      RequestEnd (rqptr);
      return;
   }

   NetRead (rqptr, &RequestDiscardAst,
            rqptr->rqHeader.RequestHeaderPtr, rqptr->rqNet.ReadBufferSize);
}

/****************************************************************************/
/*
Parse the first <CR><LF> (strict HTTP) or newline (de facto HTTP) terminated
string from the client's request packet into HTTP method, path information and
query string.  Return true to continue processing the request, false to stop
immediately.  If false this function or any it called must have reported the
problem to the client and/or disposed of the thread.
*/ 
 
RequestLineParse (REQUEST_STRUCT *rqptr)

{
   unsigned char  c;
   char  *cptr, *rptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout, "RequestLineParse()\n");
      /** fprintf (stdout, "|%s|\n", rqptr->rqNet.ReadBufferPtr); **/
      fprintf (stdout, "|%s|\n", rqptr->rqNet.ReadBufferPtr);
   }

   cptr = rqptr->rqNet.ReadBufferPtr;

   zptr = (sptr = rqptr->rqHeader.MethodName) + sizeof(rqptr->rqHeader.MethodName);
   while (!ISLWS(*cptr) && NOTEOL(*cptr) && sptr < zptr) *sptr++ = *cptr++;
   if (sptr >= zptr) sptr = rqptr->rqHeader.MethodName;
   *sptr = '\0';
   if (Debug) fprintf (stdout, "method |%s|\n", rqptr->rqHeader.MethodName);

   /*error, no method on request line */
   if (!rqptr->rqHeader.MethodName[0]) goto RequestLineParseFormatError;

   /* skip across white-space between method and path */
   while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;

   /* get the path information, noting the start of the requested resource */
   rptr = zptr = cptr;
   while (!ISLWS(*zptr) && *zptr != '?' && NOTEOL(*zptr)) zptr++;

   /*error, no path on request line */
   if (cptr == zptr) goto RequestLineParseFormatError;

   rqptr->rqHeader.PathInfoPtr = sptr =
      VmGetHeap (rqptr, (rqptr->rqHeader.PathInfoLength = zptr-cptr)+1);
   memcpy (sptr, cptr, zptr-cptr);
   sptr[zptr-cptr] = '\0';
   if (Debug) fprintf (stdout, "path |%s|\n", rqptr->rqHeader.PathInfoPtr);
   /* unescape any hex-encoded characters */
   for (cptr = sptr; *cptr && *cptr != '%'; cptr++);
   if (*cptr)
   {
      sptr = cptr;
      while (*cptr)
      {
         while (*cptr && *cptr != '%') *sptr++ = *cptr++;
         if (!*cptr) break;
         /* an escaped character ("%xx" where xx is a hex number) */
         cptr++;
         c = 0;

         if (*cptr >= '0' && *cptr <= '9')
            { c = (*cptr - (int)'0') << 4; cptr++; }
         else
         if (tolower(*cptr) >= 'a' && tolower(*cptr) <= 'f')
            { c = (tolower(*cptr) - (int)'a' + 10) << 4; cptr++; }
         else
            goto RequestLineParseUrlEncodeError;

         if (*cptr >= '0' && *cptr <= '9')
            { c += (*cptr - (int)'0'); cptr++; }
         else
         if (tolower(*cptr) >= 'a' && tolower(*cptr) <= 'f')
            { c += (tolower(*cptr) - (int)'a' + 10); cptr++; }
         else
            goto RequestLineParseUrlEncodeError;

         *sptr++ = c;
      }
      *sptr = '\0';
      rqptr->rqHeader.PathInfoLength =
         sptr - (char*)rqptr->rqHeader.PathInfoPtr;
   }
   /* point back into the request */ 
   cptr = zptr;

   /* get any query string, or create an empty one */
   if (*cptr == '?') cptr++;
   zptr = cptr;
   while (!ISLWS(*zptr) && NOTEOL(*zptr)) zptr++;
   rqptr->rqHeader.QueryStringPtr = sptr = VmGetHeap (rqptr, zptr-cptr+1);
   if (rqptr->rqHeader.QueryStringLength = zptr-cptr)
      memcpy (sptr, cptr, zptr-cptr);
   sptr[zptr-cptr] = '\0';
   if (Debug) fprintf (stdout, "query |%s|\n", rqptr->rqHeader.QueryStringPtr);

   /* store the requested resource */
   rqptr->rqHeader.RequestUriPtr = sptr = VmGetHeap (rqptr, zptr-rptr+1);
   memcpy (sptr, rptr, rqptr->rqHeader.RequestUriLength = zptr-rptr);
   sptr[zptr-rptr] = '\0';

   /* check the HTTP version  */
   if (Debug) fprintf (stdout, "zptr |%s|\n", zptr);
   while (ISLWS(*zptr) && NOTEOL(*zptr)) zptr++;
   cptr = zptr;
   while (*zptr && *zptr != '\n') zptr++;
   if (*zptr) zptr++;
   rqptr->rqHeader.RequestLineLength = zptr - (char*)rqptr->rqNet.ReadBufferPtr;

   if (!memcmp (cptr, "HTTP/", 5))
   {
      if (!memcmp (cptr+5, "1.0", 3))
      {
         rqptr->rqHeader.HttpVersion = HTTP_VERSION_1_0;
         return (true);
      }
      else
      if (!memcmp (cptr+5, "1.1", 3))
      {
         rqptr->rqHeader.HttpVersion = HTTP_VERSION_1_1;
         return (true);
      }
      rqptr->rqHeader.HttpVersion = HTTP_VERSION_UNKNOWN;
      return (true);
   }
   else
   {
      /* no HTTP version followed the path */
      rqptr->rqHeader.HttpVersion = HTTP_VERSION_0_9;
      return (true);
   }

   /* report a request line format problem and return false */
   RequestLineParseFormatError:
   {
      AccountingPtr->RequestErrorCount++;
      rqptr->rqResponse.HttpStatus = 400;
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_REQUEST_FORMAT), FI_LI);
      return (false);
   }
   RequestLineParseUrlEncodeError:
   {
      AccountingPtr->RequestErrorCount++;
      rqptr->rqResponse.HttpStatus = 400;
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_REQUEST_URL_ENC), FI_LI);
      return (false);
   }
}

/*****************************************************************************/
/*
Scan the header buffer looking for relevant fields. Return true to continue
processing the request, false to stop immediately.  If false this function or
any it called must have reported the problem to the client and/or disposed of
the thread.
*/ 
 
BOOL RequestFields (REQUEST_STRUCT *rqptr)

{
   int  idx, Length;
   char  *cptr, *hptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "RequestFields()\n");

   /* ensure any old field values are ignored (in the case of redirection) */

   rqptr->rqHeader.AcceptPtr =
      rqptr->rqHeader.AcceptLangPtr =
      rqptr->rqHeader.AcceptCharsetPtr =
      rqptr->rqHeader.AcceptEncodingPtr =
      rqptr->rqHeader.AuthorizationPtr =
      rqptr->rqHeader.ContentTypePtr =
      rqptr->rqHeader.CookiePtr =
      rqptr->rqHeader.ForwardedPtr =
      rqptr->rqHeader.HostPtr =
      rqptr->rqHeader.IfModifiedSincePtr =
      rqptr->rqHeader.RefererPtr =
      rqptr->rqHeader.UserAgentPtr = NULL;

   for (idx = 0; idx < RESPONSE_COOKIE_MAX; idx++)
      rqptr->rqResponse.CookiePtr[idx] = NULL;

   rqptr->rqHeader.ContentLength =
      rqptr->rqHeader.AcceptLength =
      rqptr->rqHeader.AcceptCharsetLength =
      rqptr->rqHeader.AcceptLangLength =
      rqptr->rqHeader.ForwardedLength =
      rqptr->rqTime.IfModifiedSinceVMS64bit[0] =
      rqptr->rqTime.IfModifiedSinceVMS64bit[1] = 0;

   rqptr->KeepAliveRequest = false;

   if (rqptr->WatchItem &&
       (WatchEnabled & WATCH_REQUEST_HEADER && WatchEnabled & WATCH_DETAIL))
   {
      if (Length = rqptr->rqHeader.RequestLineLength) Length--;
      while (rqptr->rqHeader.RequestHeaderPtr[Length] == '\r' ||
             rqptr->rqHeader.RequestHeaderPtr[Length] == '\n') Length--;
      WatchThis (rqptr, FI_LI, WATCH_REQUEST_HEADER+WATCH_DETAIL,
                 "!#HZ", Length, rqptr->rqHeader.RequestHeaderPtr);
   }

   /* step to the end of the request line (e.g. "GET /path HTTP/1.0\n") */
   hptr = rqptr->rqHeader.RequestHeaderPtr + rqptr->rqHeader.RequestLineLength;
   /* look for the request-terminating blank line */
   while (*hptr)
   {
      /** if (Debug) fprintf (stdout, "hptr |%s|\n", hptr); **/

      cptr = hptr;
      while (NOTEOL(*hptr)) hptr++;
      if (hptr[0] == '\n' || (hptr[0] == '\r' && hptr[1] == '\n'))
      {
         /* break if blank line (i.e. end-of-header) */
         if (cptr == hptr) break;
      }

      if (rqptr->WatchItem &&
          (WatchEnabled & WATCH_REQUEST_HEADER && WatchEnabled & WATCH_DETAIL))
         WatchThis (rqptr, FI_LI, WATCH_REQUEST_HEADER+WATCH_DETAIL,
                    "!#HZ", hptr-cptr, cptr);

      /* step over the carriage control at the end of the line */
      if (*hptr == '\r') hptr++;
      if (*hptr == '\n') hptr++;
      if (Debug)
         fprintf (stdout, "field |%*.*s|\n", hptr-cptr, hptr-cptr, cptr);

      /***********/
      /* Accept: */
      /***********/

      if (toupper(cptr[0]) == 'A' && cptr[6] == ':' &&
          strsame (cptr, "Accept:", 7))
      {
         cptr += 7;
         while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
         sptr = cptr;
         while (NOTEOL(*sptr)) sptr++;
         Length = sptr - cptr;
         if (rqptr->rqHeader.AcceptPtr == NULL)
            rqptr->rqHeader.AcceptPtr = sptr = VmGetHeap (rqptr, Length+1);
         else
         {
            rqptr->rqHeader.AcceptPtr =
               VmReallocHeap (rqptr, rqptr->rqHeader.AcceptPtr, 
                              rqptr->rqHeader.AcceptLength+Length+2, FI_LI);
            sptr = rqptr->rqHeader.AcceptPtr + rqptr->rqHeader.AcceptLength;
            *sptr++ = ',';
         }
         while (NOTEOL(*cptr))
         {
            if (ISLWS(*cptr))
               while (ISLWS(*cptr)) cptr++;
            else
               *sptr++ = *cptr++;
         }
         *sptr = '\0';
         rqptr->rqHeader.AcceptLength = sptr - (char*)rqptr->rqHeader.AcceptPtr;
         if (Debug)
            fprintf (stdout, "HttpAcceptPtr %d |%s|\n",
                     rqptr->rqHeader.AcceptLength, rqptr->rqHeader.AcceptPtr);
         continue;
      }

      /*******************/
      /* Accept-Charset: */
      /*******************/

      if (toupper(cptr[0]) == 'A' && tolower(cptr[7]) == 'c' &&
          strsame (cptr, "Accept-charset:", 15))
      {
         cptr += 15;
         while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
         sptr = cptr;
         while (NOTEOL(*sptr)) sptr++;
         Length = sptr - cptr;
         if (rqptr->rqHeader.AcceptCharsetPtr == NULL)
            rqptr->rqHeader.AcceptCharsetPtr = sptr = VmGetHeap (rqptr, Length+1);
         else
         {
            rqptr->rqHeader.AcceptCharsetPtr =
               VmReallocHeap (rqptr, rqptr->rqHeader.AcceptCharsetPtr, 
                              rqptr->rqHeader.AcceptCharsetLength+Length+2, FI_LI);
            sptr = rqptr->rqHeader.AcceptCharsetPtr + rqptr->rqHeader.AcceptCharsetLength;
            *sptr++ = ',';
         }
         while (NOTEOL(*cptr))
         {
            if (ISLWS(*cptr))
               while (ISLWS(*cptr)) cptr++;
            else
               *sptr++ = *cptr++;
         }
         *sptr = '\0';
         rqptr->rqHeader.AcceptCharsetLength =
            sptr - (char*)rqptr->rqHeader.AcceptCharsetPtr;
         if (Debug)
            fprintf (stdout, "HttpAcceptCharsetPtr %d |%s|\n",
                     rqptr->rqHeader.AcceptCharsetLength,
                     rqptr->rqHeader.AcceptCharsetPtr);
         continue;
      }

      /********************/
      /* Accept-Encoding: */
      /********************/

      if (toupper(cptr[0]) == 'A' && tolower(cptr[7]) == 'e' &&
          strsame (cptr, "Accept-encoding:", 16))
      {
         cptr += 16;
         while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
         sptr = cptr;
         while (NOTEOL(*sptr)) sptr++;
         Length = sptr - cptr;
         if (rqptr->rqHeader.AcceptEncodingPtr == NULL)
            rqptr->rqHeader.AcceptEncodingPtr = sptr = VmGetHeap (rqptr, Length+1);
         else
         {
            rqptr->rqHeader.AcceptEncodingPtr =
               VmReallocHeap (rqptr, rqptr->rqHeader.AcceptEncodingPtr, 
                              rqptr->rqHeader.AcceptEncodingLength+Length+2, FI_LI);
            sptr = rqptr->rqHeader.AcceptEncodingPtr +
                   rqptr->rqHeader.AcceptEncodingLength;
            *sptr++ = ',';
         }
         while (NOTEOL(*cptr))
         {
            if (ISLWS(*cptr))
               while (ISLWS(*cptr)) cptr++;
            else
               *sptr++ = *cptr++;
         }
         *sptr = '\0';
         rqptr->rqHeader.AcceptEncodingLength =
            sptr - (char*)rqptr->rqHeader.AcceptEncodingPtr;
         if (Debug)
            fprintf (stdout, "HttpAcceptEncodingPtr %d |%s|\n",
                     rqptr->rqHeader.AcceptEncodingLength,
                     rqptr->rqHeader.AcceptEncodingPtr);
         continue;
      }

      /********************/
      /* Accept-Language: */
      /********************/

      if (toupper(cptr[0]) == 'A' && tolower(cptr[7]) == 'l' &&
          strsame (cptr, "Accept-language:", 16))
      {
         cptr += 16;
         while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
         sptr = cptr;
         while (NOTEOL(*sptr)) sptr++;
         Length = sptr - cptr;
         if (rqptr->rqHeader.AcceptLangPtr == NULL)
            rqptr->rqHeader.AcceptLangPtr = sptr = VmGetHeap (rqptr, Length+1);
         else
         {
            rqptr->rqHeader.AcceptLangPtr =
               VmReallocHeap (rqptr, rqptr->rqHeader.AcceptLangPtr, 
                              rqptr->rqHeader.AcceptLangLength+Length+2, FI_LI);
            sptr = rqptr->rqHeader.AcceptLangPtr + rqptr->rqHeader.AcceptLangLength;
            *sptr++ = ',';
         }
         while (NOTEOL(*cptr))
         {
            if (ISLWS(*cptr))
               while (ISLWS(*cptr)) cptr++;
            else
               *sptr++ = *cptr++;
         }
         *sptr = '\0';
         rqptr->rqHeader.AcceptLangLength = sptr - (char*)rqptr->rqHeader.AcceptLangPtr;
         if (Debug)
            fprintf (stdout, "HttpAcceptLangPtr %d |%s|\n",
                     rqptr->rqHeader.AcceptLangLength, rqptr->rqHeader.AcceptLangPtr);
         continue;
      }

      /******************/
      /* Authorization: */
      /******************/

      if (toupper(cptr[0]) == 'A' && tolower(cptr[1]) == 'u' &&
          strsame (cptr, "Authorization:", 14))
      {
         cptr += 14;
         while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
         sptr = cptr;
         while (NOTEOL(*sptr)) sptr++;
         Length = sptr - cptr;
         rqptr->rqHeader.AuthorizationPtr = sptr = VmGetHeap (rqptr, Length+1);
         memcpy (sptr, cptr, rqptr->rqHeader.AuthorizationLength = Length);
         sptr[Length] = '\0';
         if (Debug) fprintf (stdout, "Authorization |%s|\n", sptr);

         continue;
      }

      /***************/
      /* Connection: */
      /***************/

      /*
         Only check for persistent connection request if it's configured.
         Persistent connection only allowed for object retrievals.
      */
      if (Config.cfTimeout.KeepAlive &&
          rqptr->rqNet.KeepAliveCount < KEEPALIVE_MAX &&
          rqptr->ServicePtr->RequestScheme != SCHEME_HTTPS)
      {
         if (toupper(cptr[0]) == 'C' && tolower(cptr[8]) == 'o' &&
             strsame (cptr, "Connection:", 11))
         {
            cptr += 11;
            while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
            if (toupper(cptr[0]) == 'K' && strsame (cptr, "Keep-Alive", 10))
            {
               rqptr->KeepAliveRequest = true;
               if (Debug) fprintf (stdout, "KEEP-ALIVE!\n");
            }
            continue;
         }
      }

      /*******************/
      /* Content-Length: */
      /*******************/

      if (toupper(cptr[0]) == 'C' && tolower(cptr[8]) == 'l' &&
          strsame (cptr, "Content-length:", 15))
      {
         cptr += 15;
         while (!isdigit(*cptr) && NOTEOL(*cptr)) cptr++;
         if (!isdigit(*cptr))
         {
            AccountingPtr->RequestErrorCount++;
            rqptr->rqResponse.HttpStatus = 400;
            ErrorGeneral (rqptr, MsgFor(rqptr,MSG_REQUEST_FORMAT), FI_LI);
            return (false);
         }
         rqptr->rqHeader.ContentLength = atoi (cptr);
         if (Debug)
            fprintf (stdout, "ContentLength |%d|\n", rqptr->rqHeader.ContentLength);
         continue;
      }

      /*****************/
      /* Content-Type: */
      /*****************/

      if (toupper(cptr[0]) == 'C' && tolower(cptr[8]) == 't' &&
          strsame (cptr, "Content-type:", 13))
      {
         cptr += 13;
         while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
         sptr = cptr;
         while (NOTEOL(*sptr)) sptr++;
         Length = sptr - cptr;
         rqptr->rqHeader.ContentTypePtr = sptr = VmGetHeap (rqptr, Length+1);
         memcpy (sptr, cptr, rqptr->rqHeader.ContentTypeLength = Length);
         sptr[Length] = '\0';
         if (Debug) fprintf (stdout, "HttpContentType |%s|\n", sptr);
         continue;
      }

      /***********/
      /* Cookie: */
      /***********/

      if (toupper(cptr[0]) == 'C' && cptr[6] == ':' &&
          strsame (cptr, "Cookie:", 7))
      {
         cptr += 7;
         while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
         sptr = cptr;
         while (NOTEOL(*sptr)) sptr++;
         Length = sptr - cptr;
         rqptr->rqHeader.CookiePtr = sptr = VmGetHeap (rqptr, Length+1);
         memcpy (sptr, cptr, Length);
         sptr[Length] = '\0';
         if (Debug) fprintf (stdout, "HttpCookie |%s|\n", sptr);
         continue;
      }

      /*********/
      /* ETag: */
      /*********/

      if (toupper(cptr[0]) == 'E' && cptr[4] == ':' &&
          strsame (cptr, "ETag:", 5))
      {
         cptr += 5;
         while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
         sptr = cptr;
         while (NOTEOL(*sptr)) sptr++;
         Length = sptr - cptr;
         rqptr->rqHeader.ETagPtr = sptr = VmGetHeap (rqptr, Length+1);
         memcpy (sptr, cptr, Length);
         sptr[Length] = '\0';
         if (Debug) fprintf (stdout, "HttpETag |%s|\n", sptr);
         continue;
      }

      /**************/
      /* Forwarded: */
      /**************/

      if (toupper(cptr[0]) == 'F' && tolower(cptr[1]) == 'o' &&
          strsame (cptr, "Forwarded:", 10))
      {
         cptr += 10;
         while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
         sptr = cptr;
         while (NOTEOL(*sptr)) sptr++;
         Length = sptr - cptr;
         if (rqptr->rqHeader.ForwardedPtr == NULL)
            rqptr->rqHeader.ForwardedPtr = sptr = VmGetHeap (rqptr, Length+1);
         else
         {
            rqptr->rqHeader.ForwardedPtr =
               VmReallocHeap (rqptr, rqptr->rqHeader.ForwardedPtr,
                              rqptr->rqHeader.ForwardedLength+Length+2, FI_LI);
            sptr = rqptr->rqHeader.ForwardedPtr + rqptr->rqHeader.ForwardedLength;
            *sptr++ = ',';
         }
         memcpy (sptr, cptr, Length);
         sptr[Length] = '\0';
         rqptr->rqHeader.ForwardedLength =
            sptr - (char*)rqptr->rqHeader.ForwardedPtr + Length;
         if (Debug)
            fprintf (stdout, "HttpForwardedPtr %d |%s|\n",
                     rqptr->rqHeader.ForwardedLength, rqptr->rqHeader.ForwardedPtr);
         continue;
      }

      /**********************/
      /* If-Modified-Since: */
      /**********************/

      if (toupper(cptr[0]) == 'I' && tolower(cptr[1]) == 'f' &&
          strsame (cptr, "If-Modified-Since:", 18))
      {
         cptr += 18;
         while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
         sptr = cptr;
         while (NOTEOL(*sptr)) sptr++;
         Length = sptr - cptr;
         rqptr->rqHeader.IfModifiedSincePtr = sptr = VmGetHeap (rqptr, Length+1);
         memcpy (sptr, cptr, Length);
         sptr[Length] = '\0';
         if (Debug) fprintf (stdout, "HttpIfModifiedSince |%s|\n", sptr);

         if (VMSnok (HttpGmTime (sptr, &rqptr->rqTime.IfModifiedSinceVMS64bit)))
         {
            if (Debug) fprintf (stdout, "If-Modified-Since: NBG!\n");
            rqptr->rqHeader.IfModifiedSincePtr = NULL;
            rqptr->rqTime.IfModifiedSinceVMS64bit[0] =
            rqptr->rqTime.IfModifiedSinceVMS64bit[1] = 0;

            continue;
         }

         /* indicate that no "length=" was found (if none is found :^) */
         rqptr->rqHeader.IfModifiedSinceLength = -1;

         /* look for a "length=" parameter to this header field */
         while (*sptr)
         {
            while (*sptr && *sptr != ';') sptr++; 
            if (!*sptr) break;
            sptr++;
            while (*sptr && ISLWS(*sptr)) sptr++; 
            if (!*sptr) break;
            if (!strsame (sptr, "length", 6)) continue;
            sptr += 6;
            while (*sptr && *sptr != '=') sptr++; 
            if (!*sptr) break;
            sptr++;
            while (*sptr && ISLWS(*sptr)) sptr++; 
            if (!isdigit (*sptr)) continue;
            rqptr->rqHeader.IfModifiedSinceLength = atol(sptr);
            if (Debug)
               fprintf (stdout, "length=%d\n",
                        rqptr->rqHeader.IfModifiedSinceLength);
         }

         continue;
      }

      /*********/
      /* Host: */
      /*********/

      if (toupper(cptr[0]) == 'H' && tolower(cptr[1]) == 'o' &&
          strsame (cptr, "Host:", 5))
      {
         cptr += 5;
         while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
         sptr = cptr;
         while (NOTEOL(*sptr)) sptr++;
         Length = sptr - cptr;
         rqptr->rqHeader.HostPtr = sptr = VmGetHeap (rqptr, Length+1);
         memcpy (sptr, cptr, Length);
         sptr[Length] = '\0';
         if (Debug) fprintf (stdout, "HttpHost |%s|\n", sptr);
         continue;
      }

      /***********/
      /* Pragma: */
      /***********/

      if (toupper(cptr[0]) == 'P' && tolower(cptr[1]) == 'r' &&
          strsame (cptr, "Pragma:", 7))
      {
         cptr += 7;
         while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
         sptr = cptr;
         while (NOTEOL(*sptr)) sptr++;
         Length = sptr - cptr;
         rqptr->rqHeader.PragmaPtr = sptr = VmGetHeap (rqptr, Length+1);
         memcpy (sptr, cptr, Length);
         sptr[Length] = '\0';
         if (strsame (sptr, "no-cache", -1))
         {
            rqptr->PragmaNoCache = true;
            AccountingPtr->RequestPragmaNoCacheCount++;
         }
         if (Debug)
            fprintf (stdout, "HttpPragma |%s| no-cache: %d\n",
                     sptr, rqptr->PragmaNoCache);
         continue;
      }

      /************************/
      /* Proxy-Authorization: */
      /************************/

      if (toupper(cptr[0]) == 'P' && tolower(cptr[1]) == 'r' &&
          strsame (cptr, "Proxy-Authorization:", 20))
      {
         cptr += 20;
         while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
         sptr = cptr;
         while (NOTEOL(*sptr)) sptr++;
         Length = sptr - cptr;
         rqptr->rqHeader.ProxyAuthorizationPtr = sptr = VmGetHeap (rqptr, Length+1);
         memcpy (sptr, cptr, rqptr->rqHeader.ProxyAuthorizationLength = Length);
         sptr[Length] = '\0';
         if (Debug) fprintf (stdout, "Proxy-Authorization |%s|\n", sptr);

         continue;
      }

      /*********************/
      /* Proxy-Connection: */
      /*********************/

      if (toupper(cptr[0]) == 'P' && tolower(cptr[1]) == 'r' &&
          strsame (cptr, "Proxy-Connection:", 17))
      {
         cptr += 17;
         while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
         sptr = cptr;
         while (NOTEOL(*sptr)) sptr++;
         Length = sptr - cptr;
         rqptr->rqHeader.ProxyConnectionPtr = sptr = VmGetHeap (rqptr, Length+1);
         memcpy (sptr, cptr, Length);
         sptr[Length] = '\0';
         if (Debug) fprintf (stdout, "ProxyConnection |%s|\n", sptr);
         continue;
      }

      /************/
      /* Referer: */
      /************/

      if (toupper(cptr[0]) == 'R' && tolower(cptr[1]) == 'e' &&
          strsame (cptr, "Referer:", 8))
      {
         cptr += 8;
         while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
         sptr = cptr;
         while (NOTEOL(*sptr)) sptr++;
         Length = sptr - cptr;
         rqptr->rqHeader.RefererPtr = sptr = VmGetHeap (rqptr, Length+1);
         memcpy (sptr, cptr, rqptr->rqHeader.RefererLength = Length);
         sptr[Length] = '\0';
         if (Debug) fprintf (stdout, "Referer |%s|\n", sptr);
         continue;
      }

      /***************/
      /* User-Agent: */
      /***************/

      if (toupper(cptr[0]) == 'U' && tolower(cptr[1]) == 's' &&
          strsame (cptr, "User-Agent:", 11))
      {
         cptr += 11;
         while (ISLWS(*cptr) && NOTEOL(*cptr)) cptr++;
         sptr = cptr;
         while (NOTEOL(*sptr)) sptr++;
         Length = sptr - cptr;
         rqptr->rqHeader.UserAgentPtr = sptr = VmGetHeap (rqptr, Length+1);
         memcpy (sptr, cptr, rqptr->rqHeader.UserAgentLength = Length);
         sptr[Length] = '\0';
         if (Debug) fprintf (stdout, "UserAgent |%s|\n", sptr);
         continue;
      }
   }

   return (true);
}

/*****************************************************************************/
/*
If a file specification does not contain a file name the function looks for a
home page in the directory.  If no home page found it generates a directory
listing ("Index of" documents).

If the file specification contains a wildcard and there is no query string a
directory listing (index) is generated.  If there is a query string it must
begin with the literal "?httpd=index" for a directory listing (index) to be
generated (this allows directives in the query string to be passed to the
directory listing functions).  If it does not begin with this literal the
default query (search) script is invoked.

Detects file types that have an associated script.  If a script is returned by
ConfigContentType(), and the file specification does not include a specific
version number (even such as ";0"), an automatic redirection is generated so
that the specified document is processed by the script, and the output from
that returned to the client.  If it does contain a version number the file is
always directly returned to the client.
*/ 

RequestExecute (REQUEST_STRUCT *rqptr)

{
   int  status,
        Length;
   char  *cptr, *mpptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "RequestExecute()\n");

   if (!(rqptr->ServicePtr->ProxyService ||
         rqptr->ServicePtr->ConnectService))
   {
      if (!NetVirtualService (rqptr))
      {
         /* virtual service was not found */
         if (Config.cfServer.ServiceNotFoundUrl[0])
         {
            /* configuration specifies a redirection URL */
            rqptr->rqResponse.LocationPtr = Config.cfServer.ServiceNotFoundUrl;
            RequestEnd (rqptr);
            return;
         }
      }
   }

   if (rqptr->WatchItem)
   {
      if ((WatchEnabled & WATCH_REQUEST_BODY) &&
          rqptr->rqHeader.ContentLength)
      {
         WatchThis (rqptr, FI_LI, WATCH_REQUEST_BODY,
                    "BODY !UL/!UL bytes",
                    rqptr->rqBody.CurrentCount,
                    rqptr->rqHeader.ContentLength);
         WatchDataDump (rqptr->rqBody.BufferPtr, rqptr->rqBody.CurrentCount);
      }

      if (WatchEnabled & WATCH_REQUEST)
         WatchThis (rqptr, FI_LI, WATCH_REQUEST,
                    "!AZ !AZ", rqptr->rqHeader.MethodName,
                               rqptr->rqHeader.RequestUriPtr);
   }

   /* ensure any old values are ignored, in case its a redirection */
   rqptr->rqResponse.LocationPtr = NULL;

   /************************/
   /* map the request path */
   /************************/

   rqptr->RequestMappedFile[0] =
      rqptr->ScriptName[0] =
      rqptr->RequestMappedScript[0] = '\0';

   /*
      Map the path, converting it to a VMS file path (or script file name)
      and returning a pointer to a static buffer containing the mapped path.
   */
   mpptr = MapUrl_Map (rqptr->rqHeader.PathInfoPtr,
                       sizeof(rqptr->rqHeader.PathInfoPtr),
                       rqptr->RequestMappedFile,
                       sizeof(rqptr->RequestMappedFile),
                       rqptr->ScriptName,
                       sizeof(rqptr->ScriptName),
                       rqptr->RequestMappedScript,
                       sizeof(rqptr->RequestMappedScript),
                       rqptr->RequestMappedRunTime,
                       sizeof(rqptr->RequestMappedRunTime),
                       &rqptr->PathOds, rqptr);

   /* buffer the mapped path */
   rqptr->MappedPathLength = Length = strlen(mpptr);
   rqptr->MappedPathPtr = VmGetHeap (rqptr, Length+1);
   memcpy (rqptr->MappedPathPtr, mpptr, Length+1);

   if (Debug)
      fprintf (stdout, "|%s|%s|%s|%s|%s|\n",
               mpptr, rqptr->rqHeader.PathInfoPtr, rqptr->RequestMappedFile,
               rqptr->ScriptName, rqptr->RequestMappedScript);

   for (cptr = rqptr->RequestMappedFile; *cptr && *cptr != ':'; cptr++);
   if (*(unsigned short*)cptr == '::')
   {
      /* if a node name is part of the VMS file name then it's not supported */
      rqptr->rqResponse.HttpStatus = 501;
      rqptr->rqResponse.ErrorTextPtr = rqptr->rqHeader.PathInfoPtr;
      ErrorVmsStatus (rqptr, RMS$_SUPPORT, FI_LI);
      RequestEnd (rqptr);
      return;
   }

   /* request is underway, initialize the output (response) timeout */
   HttpdTimerSet (rqptr, TIMER_OUTPUT, 0);

   if (!mpptr[0] && mpptr[1])
   {
      /* MapUrl_Map() returns errors with a leading null character */
      mpptr++;
      if (isdigit(*mpptr))
      {
         if (!strcmp (mpptr, MAPURL_USER_RULE_FORBIDDEN_MSG+1))
         {
            /* convert USER mapping rule forbidden into directory not found */
            rqptr->rqResponse.HttpStatus = 404;
            rqptr->rqResponse.ErrorTextPtr = rqptr->rqHeader.PathInfoPtr;
            ErrorVmsStatus (rqptr, RMS$_DNF, FI_LI);
            RequestEnd (rqptr);
            return;
         }

         /* HTTP status code mapping, with following message */
         rqptr->rqResponse.HttpStatus = atol(mpptr);
         while (*mpptr && isdigit(*mpptr)) mpptr++;
         while (*mpptr && ISLWS(*mpptr)) mpptr++;

         if (rqptr->rqResponse.HttpStatus >= 300 &&
             rqptr->rqResponse.HttpStatus <= 399)
         {
            /* redirection, message "text" should be the location */
            rqptr->rqResponse.LocationPtr = mpptr;
            RequestEnd (rqptr);
            return;
         }
         else
         if (rqptr->rqResponse.HttpStatus >= 400 &&
             rqptr->rqResponse.HttpStatus <= 599)
         {
            ErrorGeneral (rqptr, mpptr, FI_LI);
            RequestEnd (rqptr);
            return;
         }
         else
         {
            /* no point to codes other 3nn/4nn/5nn, use to abort connection */
            RequestEnd (rqptr);
            return;
         }
      }
      else
      {
         /* just a rule-mapping message */
         rqptr->rqResponse.HttpStatus = 403;
         ErrorGeneral (rqptr, mpptr, FI_LI);
         RequestEnd (rqptr);
         return;
      }
   }

#ifdef ODS_EXTENDED
   if (OdsExtended)
      rqptr->PathOdsExtended = (rqptr->PathOds == 5);
   else
#endif /* ODS_EXTENDED */
      rqptr->PathOds = rqptr->PathOdsExtended = 0;

   /* MapUrl_Map() returns CGIplus mappings indicated by a leading '+' */
   if (rqptr->ScriptName[0] == '+')
   {
      rqptr->IsCgiPlusScript = true;
      rqptr->ScriptName[0] = '/';
   }
   else
      rqptr->IsCgiPlusScript = false;

   if (Debug)
      fprintf (stdout, "|%s|%s|%s|%d|%s|%s|%s|\n",
               rqptr->rqHeader.PathInfoPtr, rqptr->RequestMappedFile,
               rqptr->ScriptName, rqptr->IsCgiPlusScript,
               rqptr->RequestMappedScript, rqptr->RequestMappedRunTime, mpptr);

   /********************************************/
   /* check for redirection (from the mapping) */
   /********************************************/

   if (mpptr[0] == '<' || mpptr[0] == '>')
   {
      /* local will begin with a '<', non-local with ">", etc. */
      rqptr->rqResponse.LocationPtr = VmGetHeap (rqptr, strlen(mpptr+1));
      strcpy (rqptr->rqResponse.LocationPtr, mpptr+1);
      RequestEnd (rqptr);
      return;
   }

   /******************/
   /* proxy request? */
   /******************/

   if (mpptr[0] != '/' && !rqptr->ScriptName[0])
   {
      rqptr->ProxyRequest = true;
      if (rqptr->ServicePtr->LocalAuthRequired)
         rqptr->rqAuth.RequestAuthorizationPtr = rqptr->rqHeader.AuthorizationPtr;
      else
      if (rqptr->ServicePtr->ProxyAuthRequired)
         rqptr->rqAuth.RequestAuthorizationPtr = rqptr->rqHeader.ProxyAuthorizationPtr;
      else
      {
         /* look for a URL scheme delimiter */
         for (cptr = mpptr; *cptr && *cptr != '/' && *cptr != ':'; cptr++);
         if (*cptr == ':')
         {
            ProxyRequestBegin (rqptr);
            return;
         }
      }
   }
   else
      rqptr->rqAuth.RequestAuthorizationPtr = rqptr->rqHeader.AuthorizationPtr;

   /*********************************/
   /* authorization of request path */
   /*********************************/

   if (!rqptr->rqPathSet.AuthorizeMapped)
      Authorize (rqptr, rqptr->rqHeader.PathInfoPtr,
                 rqptr->rqHeader.PathInfoLength, &RequestExecutePostAuth1);
   else
   if (rqptr->ScriptName[0])
      Authorize (rqptr, rqptr->ScriptName,
                 strlen(rqptr->ScriptName), &RequestExecutePostAuth1);
   else
      Authorize (rqptr, rqptr->MappedPathPtr,
                 rqptr->MappedPathLength, &RequestExecutePostAuth1);
   if (Debug)
      fprintf (stdout, "rqAuth.FinalStatus: %%X%08.08X\n",
               rqptr->rqAuth.FinalStatus);

   if (VMSnok (rqptr->rqAuth.FinalStatus))
   {
      /* if asynchronous authentication is underway then just wait for it */
      if (rqptr->rqAuth.FinalStatus == AUTH_PENDING) return;

      RequestEnd (rqptr);
      return;
   }

   /* authorization completed, continue to process */ 
   RequestExecutePostAuth1 (rqptr);
}

/*****************************************************************************/
/*
This function continues to initiate the request processing, either called
directly from RequestExecute() or as an AST after asynchronous authentication.
*/

RequestExecutePostAuth1 (REQUEST_STRUCT *rqptr)

{
   int  status;
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "RequestExecutePostAuth1() %%X%08.08X\n",
               rqptr->rqAuth.FinalStatus);

   if (VMSnok (rqptr->rqAuth.FinalStatus))
   {
      if (Debug)
         fprintf (stdout, "rqAuth.FinalStatus: %%X%08.08X\n",
                  rqptr->rqAuth.FinalStatus);
      RequestEnd (rqptr);
      return;
   }

   /*****************************/
   /* authorized proxy request? */
   /*****************************/

   if (rqptr->ProxyRequest)
   {
      if ((rqptr->ServicePtr->LocalAuthRequired ||
           rqptr->ServicePtr->ProxyAuthRequired) &&
          !rqptr->RemoteUser[0])
      {
         /* authorization never occured! */
         rqptr->rqResponse.HttpStatus = 403;
         ErrorGeneral (rqptr, MsgFor(rqptr,MSG_GENERAL_DISABLED), FI_LI);
         RequestEnd (rqptr);
         return;
      }
      /* look for a URL scheme delimiter */
      for (cptr = rqptr->MappedPathPtr;
           *cptr && *cptr != '/' && *cptr != ':';
           cptr++);
      if (*cptr == ':')
      {
         ProxyRequestBegin (rqptr);
         return;
      }
   }

   /************************************/
   /* handled internally by the server */
   /************************************/

   if (strsame (rqptr->rqHeader.RequestUriPtr, HTTPD_ADMIN,
                sizeof(HTTPD_ADMIN)-1))
   {
      AdminBegin (rqptr, &RequestEnd);
      return;
   }

   if (strsame (rqptr->rqHeader.RequestUriPtr, INTERNAL_PASSWORD_CHANGE,
                sizeof(INTERNAL_PASSWORD_CHANGE)-1))
   {
      HTAdminBegin (rqptr, &RequestEnd);
      return;
   }

   if (strsame (rqptr->rqHeader.RequestUriPtr, INTERNAL_TRACK_CANCEL,
                sizeof(INTERNAL_TRACK_CANCEL)-1))
   {
      TrackCancelCookie (rqptr, &RequestEnd);
      return;
   }

   /******************************/
   /* check for script execution */
   /******************************/

   if (rqptr->ScriptName[0])
   {
      /* get the path derived from the script specification */
      cptr = rqptr->rqHeader.PathInfoPtr;
      sptr = rqptr->ScriptName;
      while (*sptr)
      {
         if (tolower(*sptr) != tolower(*cptr)) break;
         sptr++;
         cptr++;
      }
      if (!*sptr)
      {
         for (sptr = rqptr->rqHeader.PathInfoPtr; *cptr; *sptr++ = *cptr++);
         *sptr++ = '\0';
         rqptr->rqHeader.PathInfoLength = sptr - rqptr->rqHeader.PathInfoPtr;
      }

      if (!rqptr->rqPathSet.AuthorizeOnce)
      {
         /******************************************/
         /* authorization of script path parameter */
         /******************************************/

         if (!rqptr->rqPathSet.AuthorizeMapped)
            Authorize (rqptr, rqptr->rqHeader.PathInfoPtr,
                       rqptr->rqHeader.PathInfoLength,
                       &RequestExecutePostAuth2);
         else
            Authorize (rqptr, rqptr->MappedPathPtr,
                       rqptr->MappedPathLength,
                       &RequestExecutePostAuth2);

         if (Debug)
            fprintf (stdout, "rqAuth.FinalStatus: %%X%08.08X\n",
                     rqptr->rqAuth.FinalStatus);

         if (VMSnok (rqptr->rqAuth.FinalStatus))
         {
            /* if authentication by an agent underway then just wait for it */
            if (rqptr->rqAuth.FinalStatus == AUTH_PENDING) return;

            RequestEnd (rqptr);
            return;
         }
      }
   }

   /***********************/
   /* continue non-script */ 
   /***********************/

   RequestExecutePostAuth2 (rqptr);
}

/*****************************************************************************/
/*
This function continues to initiate the request processing, either called
directly from RequestExecutePostAuth() or as an AST after asynchronous authentication.
*/

RequestExecutePostAuth2 (REQUEST_STRUCT *rqptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "RequestExecutePostAuth2() %%X%08.08X\n",
               rqptr->rqAuth.FinalStatus);

   if (VMSnok (rqptr->rqAuth.FinalStatus))
   {
      if (Debug)
         fprintf (stdout, "rqAuth.FinalStatus: %%X%08.08X\n",
                  rqptr->rqAuth.FinalStatus);
      RequestEnd (rqptr);
      return;
   }

   /* if this path has any request throttling limits */
   if (rqptr->rqPathSet.ThrottleFrom)
   {
      status = ThrottleBegin (rqptr, &RequestExecutePostThrottle);
      if (VMSnok (status))
      {
         /* either on the waiting list or waiting list has been exceeded */
         if (status == SS$_ABORT) RequestEnd (rqptr);
         return;
      }
      /* continue to process the request (is on the active list) */
   }

   RequestExecutePostThrottle (rqptr);
}

/*****************************************************************************/
/*
*/

RequestExecutePostThrottle (REQUEST_STRUCT *rqptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "RequestExecutePostThrottle() %%X%08.08X\n",
               rqptr->rqAuth.FinalStatus);

   if (rqptr->ScriptName[0])
   {
      /**********/
      /* script */
      /**********/

      RequestScript (rqptr,
                     rqptr->RequestMappedScript,
                     rqptr->RequestMappedFile,
                     rqptr->RequestMappedRunTime);
      return;
   }

   /*******************************************/
   /* check the path could actually be mapped */
   /*******************************************/

   if (!rqptr->RequestMappedFile[0])
   {
      rqptr->rqResponse.ErrorTextPtr = rqptr->rqHeader.PathInfoPtr;
      ErrorVmsStatus (rqptr, RMS$_SYN, FI_LI);
      RequestEnd (rqptr);
      return;
   }

   /********************/
   /* check file cache */
   /********************/

   /* cache use never occurs for requests with a VMS authentication profile */
   if (CacheEnabled &&
       !rqptr->rqAuth.VmsUserProfileLength &&
       (rqptr->rqHeader.Method == HTTP_METHOD_GET ||
        rqptr->rqHeader.Method == HTTP_METHOD_HEAD))
   {
      if (VMSok (CacheSearch (rqptr))) return;
      /* if not OK then drop through to continue processing */
   }

   RequestExecuteFinal (rqptr);
}

/*****************************************************************************/
/*
This function continues to initiate the request processing, either called
directly from RequestExecutePostAuth2() or from CacheBegin() or CacheAcpInfoAst() if
the cache contents are not valid and the request required continued processing
here.
*/ 

RequestExecuteFinal (REQUEST_STRUCT *rqptr)

{
   int  status,
        Count;
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "RequestExecuteFinal()\n");

   if (VMSnok (rqptr->rqAuth.FinalStatus))
   {
      if (Debug)
         fprintf (stdout, "rqAuth.FinalStatus: %%X%08.08X\n",
                  rqptr->rqAuth.FinalStatus);
      RequestEnd (rqptr);
      return;
   }

   /********************************/
   /* parse the file specification */
   /********************************/

   OdsParse (&rqptr->ParseOds,
             rqptr->RequestMappedFile, 0, NULL, 0,
             NAM$M_SYNCHK, NULL, rqptr);

   if (VMSnok (status = rqptr->ParseOds.Fab.fab$l_sts))
   {
      rqptr->rqResponse.ErrorOtherTextPtr = rqptr->RequestMappedFile;
      ErrorVmsStatus (rqptr, status, FI_LI);
      RequestEnd (rqptr);
      return;
   }

   if (VMSnok (status = OdsParseTerminate (&rqptr->ParseOds)))
   {
      ErrorNoticed (status, "OdsParseTerminate()", FI_LI);
      RequestEnd (rqptr);
      return;
   }

   /*
      If the file name component comprises a leading hyphen followed by
      16 digits and a trailing hyphen  ("-yyyymmddhhmmsshh-") then consider
      it a temporary file and delete it on close.
   */
   rqptr->DeleteOnClose = false;
   if (rqptr->ParseOds.NamNameLength == 18 &&
       rqptr->ParseOds.NamNamePtr[0] == '-' &&
       rqptr->ParseOds.NamNamePtr[17] == '-')
   {
      for (cptr = rqptr->ParseOds.NamNamePtr+1;
           *cptr && isdigit(*cptr);
           cptr++);
      if (cptr == rqptr->ParseOds.NamNamePtr+17)
      {
         rqptr->DeleteOnClose = true;
         rqptr->rqResponse.PreExpired = PRE_EXPIRE_DELETE_ON_CLOSE;
      }
   }

   /* all these "file write" methods are handled by the one set of functions */
   if ((rqptr->rqHeader.Method == HTTP_METHOD_PUT ||
        rqptr->rqHeader.Method == HTTP_METHOD_POST ||
        rqptr->rqHeader.Method == HTTP_METHOD_DELETE) &&
       !rqptr->RedirectErrorStatusCode)
   {
      PutBegin (rqptr, &RequestEnd);
      return;
   }

   if ((tolower(rqptr->rqHeader.QueryStringPtr[0]) == 'h' &&
        strsame (rqptr->rqHeader.QueryStringPtr, "httpd=index", 11)) ||
       (((rqptr->ParseOds.Nam_fnb & NAM$M_WILDCARD) ||
         (rqptr->ParseOds.Nam_fnb & NAM$M_EXP_VER &&
          rqptr->ParseOds.NamNamePtr == rqptr->ParseOds.NamTypePtr)) &&
        !rqptr->rqHeader.QueryStringPtr[0]))
   {
      /******************************/
      /* generate directory listing */
      /******************************/

      if (Config.cfDir.WildcardEnabled)
      {
         if (rqptr->rqPathSet.IndexPtr != NULL)
            cptr = rqptr->rqPathSet.IndexPtr;
         else
            cptr = rqptr->rqHeader.QueryStringPtr;

         DirBegin (rqptr, &RequestEnd, rqptr->ParseOds.ExpFileName,
                   cptr, "", false);

         return;
      }
      else
      {
         rqptr->rqResponse.HttpStatus = 403;
         ErrorGeneral (rqptr, MsgFor(rqptr,MSG_GENERAL_DISABLED), FI_LI);
         RequestEnd (rqptr);
         return;
      }
   }

   if (!rqptr->ParseOds.NamNameLength &&
       !rqptr->ParseOds.NamTypeLength &&
       !rqptr->rqHeader.QueryStringPtr[0])
   {
      /*********************************************/
      /* no file name supplied, look for home page */
      /*********************************************/

      RequestHomePage (rqptr);
      return;
   }

   /********************************/
   /* get the content-type details */
   /********************************/

   ConfigContentType (&rqptr->rqContentInfo, rqptr->ParseOds.NamTypePtr);

   if (!isdigit(rqptr->ParseOds.NamVersionPtr[1]) &&
       !(tolower(rqptr->rqHeader.QueryStringPtr[0]) == 'h' &&
         strsame (rqptr->rqHeader.QueryStringPtr, "httpd=", 6)))
   {
      /* 
         If a file specification does not include a specific version
         number (which indicates send-me-this-file-regardless), and
         the request does not contain an HTTPd query string, and
         ConfigContentType() has returned a script name, indicating
         an automatic script handling for this file type, then redirect.
      */
      if (rqptr->rqContentInfo.AutoScriptNamePtr[0] == '/')
      {
         /********************************/
         /* automatic script redirection */
         /********************************/

         AccountingPtr->DoAutoScriptCount++;

         Count = strlen(rqptr->rqContentInfo.AutoScriptNamePtr);
         Count += rqptr->rqHeader.PathInfoLength;
         Count += rqptr->rqHeader.QueryStringLength;
         Count += 2;

         rqptr->rqResponse.LocationPtr = sptr = VmGetHeap (rqptr, Count);
         for (cptr = rqptr->rqContentInfo.AutoScriptNamePtr;
              *cptr;
              *sptr++ = *cptr++);
         for (cptr = rqptr->rqHeader.PathInfoPtr; *cptr; *sptr++ = *cptr++);
         if (rqptr->rqHeader.QueryStringLength)
         {
            *sptr++ = '?';
            for (cptr = rqptr->rqHeader.QueryStringPtr; *cptr; *sptr++ = *cptr++);
         }
         *sptr = '\0';
         if (Debug)
            fprintf (stdout, "LocationPtr |%s|\n",
                     rqptr->rqResponse.LocationPtr);

         /* dispose of thread function does the actual redirection */
         RequestEnd (rqptr);
         return;
      }
   }

   /*************************************/
   /* internal types with query strings *
   /*************************************/

   if (ConfigSameContentType (rqptr->rqContentInfo.ContentTypePtr,
                              ConfigContentTypeIsMap, -1))
   {
      /* once completely read, give the file over to the ISMAP engine */
      FileSetAuthorizePath (rqptr, false);
      FileSetContentHandler (rqptr, &IsMapBegin, FILE_TYPE_ISMAP_SIZE_MAX);
      FileBegin (rqptr, &RequestEnd, NULL,
                 rqptr->ParseOds.ExpFileName,
                 rqptr->rqContentInfo.ContentTypePtr);
      return;
   }

   /**************************/
   /* implied keyword search */
   /**************************/

   if (rqptr->rqHeader.QueryStringPtr[0] &&
       Config.cfScript.DefaultSearchLength &&
       !(tolower(rqptr->rqHeader.QueryStringPtr[0]) == 'h' &&
         strsame (rqptr->rqHeader.QueryStringPtr, "httpd=", 6)))
   {
      if (Config.cfScript.DefaultSearchExcludePtr != NULL)
      {
         /********************************/
         /* exclude specified file types */
         /********************************/

         if (Debug)
            fprintf (stdout, "Config.cfScript.DefaultSearchExcludePtr |%s|\n",
                     Config.cfScript.DefaultSearchExcludePtr);
         cptr = "";
         sptr = Config.cfScript.DefaultSearchExcludePtr;
         while (*sptr)
         {
            if (Debug) fprintf (stdout, "sptr |%s|\n", sptr);
            cptr = rqptr->ParseOds.NamTypePtr;
            while (*cptr)
            {
               if (*sptr == ',' || *cptr == ';') break;
               if (toupper(*cptr) != toupper(*sptr)) break;
               if (*cptr) cptr++;
               if (*sptr) sptr++;
            }
            if ((!*cptr || *cptr == ';') && (!*sptr || *sptr == ','))
            {
               /* the comma is used to indicate a hit! */
               sptr = ",";
               break;
            }
            while (*sptr && *sptr != ',') sptr++;
            if (*sptr) sptr++;
         }

         if ((!*cptr || *cptr == ';') && *sptr == ',')
            if (WatchEnabled & WATCH_REQUEST)
               WatchThis (rqptr, FI_LI, WATCH_REQUEST,
                          "KEYWORD search \"!HZ\" excluded in \"!HZ\"",
                          rqptr->ParseOds.NamTypePtr,
                          Config.cfScript.DefaultSearchExcludePtr);
      }

      if (Config.cfScript.DefaultSearchExcludePtr == NULL ||
          (*cptr && *cptr != ';') || *sptr != ',')
      {
         /******************/
         /* search file(s) */
         /******************/

         if (WatchEnabled & WATCH_REQUEST)
            WatchThis (rqptr, FI_LI, WATCH_REQUEST,
                       "KEYWORD search \"!HZ\"",
                       rqptr->rqHeader.QueryStringPtr);

         /* local redirection, so keyword search script gets remapped */
         Count = Config.cfScript.DefaultSearchLength;
         Count += rqptr->rqHeader.PathInfoLength;
         Count += rqptr->rqHeader.QueryStringLength;
         Count += 2;

         rqptr->rqResponse.LocationPtr = sptr = VmGetHeap (rqptr, Count);

         for (cptr = Config.cfScript.DefaultSearch; *cptr; *sptr++ = *cptr++);
         for (cptr = rqptr->rqHeader.PathInfoPtr; *cptr; *sptr++ = *cptr++);
         if (rqptr->rqHeader.QueryStringLength)
         {
            *sptr++ = '?';
            for (cptr = rqptr->rqHeader.QueryStringPtr;
                *cptr;
                *sptr++ = *cptr++);
         }
         *sptr = '\0';
         RequestEnd (rqptr);
         return;
      }
   }

   /****************************************/ 
   /* otherwise ... send the file contents */
   /****************************************/ 

   /* allow for directories specified as "/dir1/dir2" (i.e. no trailing '/') */
   if (rqptr->ParseOds.NamTypeLength <= 1)
      RequestFile (rqptr, &RequestEnd, &RequestFileNoType);
   else
      RequestFile (rqptr, &RequestEnd, NULL);
}

/*****************************************************************************/
/*
This function can be called one or more time per request.  It steps through all
possible home page files until either one is found or a default directory
listing is generated.  'rqptr->HomePageStatus' attempts to short-circuit
searches for conditions other than file-not-found.  For example, if the status
is directory-not-found then there is little point in continuing to look for
home pages in that location!
*/ 

RequestHomePage (REQUEST_STRUCT *rqptr)

{
   char  *cptr,
         *HomePageNamePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "RequestHomePage() %d %d\n",
               rqptr->RequestHomePageIndex, rqptr->rqResponse.HeaderPtr);

   if (rqptr->rqResponse.HeaderPtr != NULL ||
       ERROR_REPORTED (rqptr))
   {
      RequestEnd (rqptr);
      return;
   }

   if (rqptr->RequestHomePageIndex &&
       VMSnok (rqptr->HomePageStatus) &&
       rqptr->HomePageStatus != RMS$_FNF)
   {
      rqptr->rqResponse.ErrorTextPtr = rqptr->rqHeader.PathInfoPtr;
      rqptr->rqResponse.ErrorOtherTextPtr = rqptr->ParseOds.NamDevicePtr;
      ErrorVmsStatus (rqptr, rqptr->HomePageStatus, FI_LI);
      RequestEnd (rqptr);
      return;
   }

   if (!*(HomePageNamePtr = ConfigHomePage(rqptr->RequestHomePageIndex++)))
   {
      /********************************************/
      /* no home page, generate directory listing */
      /********************************************/

      if (!Config.cfDir.Access)
      {
         /* directory listing is disabled */
         rqptr->rqResponse.ErrorTextPtr = rqptr->rqHeader.PathInfoPtr;
         rqptr->rqResponse.ErrorOtherTextPtr = rqptr->ParseOds.NamDevicePtr;
         ErrorVmsStatus (rqptr, RMS$_FNF, FI_LI);
         RequestEnd (rqptr);
         return;
      }

      rqptr->ParseOds.NamNamePtr[0] = '\0';

      if (rqptr->rqPathSet.IndexPtr != NULL)
         cptr = rqptr->rqPathSet.IndexPtr;
      else
         cptr = rqptr->rqHeader.QueryStringPtr;

      DirBegin (rqptr, &RequestEnd, rqptr->ParseOds.NamDevicePtr,
                cptr, "", false);

      return;
   }

   /********************************/
   /* generate home page file name */
   /********************************/

   /* overwrite any existing file name in the NAM block */
   strcpy (rqptr->ParseOds.NamNamePtr, HomePageNamePtr);
   /* locate the start of the file type */
   for (cptr = rqptr->ParseOds.NamNamePtr; *cptr && *cptr != '.'; cptr++);
   /* set the mime content type for this file type */
   ConfigContentType (&rqptr->rqContentInfo, cptr);

   rqptr->HomePageStatus = SS$_NORMAL;
   RequestFile (rqptr, &RequestEnd, &RequestHomePage);
}

/*****************************************************************************/
/*
Send a file to the client.  Special cases are menu and pre-processed HTML.  
*/ 

RequestFile
(
REQUEST_STRUCT *rqptr,
REQUEST_AST NextTaskFunction,
REQUEST_AST NoSuchFileFunction
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "RequestFile() |%s|%s|\n",
               rqptr->ParseOds.ExpFileName,
               rqptr->rqContentInfo.ContentTypePtr);

   if (ConfigSameContentType (rqptr->rqContentInfo.ContentTypePtr,
                              ConfigContentTypeMenu, -1))
   {
      /* once completely read, give the file over to the MENU engine */
      FileSetAuthorizePath (rqptr, false);
      FileSetContentHandler (rqptr, &MenuBegin, FILE_TYPE_MENU_SIZE_MAX);
      FileBegin (rqptr, NextTaskFunction, NoSuchFileFunction,
                 rqptr->ParseOds.ExpFileName,
                 rqptr->rqContentInfo.ContentTypePtr);
      return;
   }

   if (ConfigSameContentType (rqptr->rqContentInfo.ContentTypePtr,
                              ConfigContentTypeUrl, -1))
   {
      /* once completely read, give the file over to the URL engine */
      FileSetAuthorizePath (rqptr, false);
      FileSetContentHandler (rqptr, &FileDotUrlBegin, FILE_TYPE_URL_SIZE_MAX);
      FileBegin (rqptr, NextTaskFunction, NoSuchFileFunction,
                 rqptr->ParseOds.ExpFileName,
                 rqptr->rqContentInfo.ContentTypePtr);
      return;
   }

   if (ConfigSameContentType (rqptr->rqContentInfo.ContentTypePtr,
                              ConfigContentTypeSsi, -1) &&
       (!rqptr->rqHeader.QueryStringPtr[0] ||
        (rqptr->rqHeader.QueryStringPtr[0] &&
         tolower(rqptr->rqHeader.QueryStringPtr[0]) == 'h' &&
         (strsame (rqptr->rqHeader.QueryStringPtr, "httpd=ssi", 9) ||
          strsame (rqptr->rqHeader.QueryStringPtr, "httpd=error", 11)))))
   {
      /* once completely read, give the file over to the SSI engine */
      FileSetAuthorizePath (rqptr, false);
      FileSetContentHandler (rqptr, &SsiBegin, SsiSizeMax);
      FileBegin (rqptr, NextTaskFunction, NoSuchFileFunction,
                 rqptr->ParseOds.ExpFileName,
                 rqptr->rqContentInfo.ContentTypePtr);
      return;
   }

   /* JAF */
   FileSetAuthorizePath (rqptr, false);
   FileSetCacheAllowed (rqptr, true);
   FileBegin (rqptr, NextTaskFunction, NoSuchFileFunction,
              rqptr->ParseOds.ExpFileName,
              rqptr->rqContentInfo.ContentTypePtr);
}

/*****************************************************************************/
/*
The file requested had no type (extension) and was not found.  Check if there
is a directory of that name.  This allows for directories to be specified as
"/dir1/dir2" (i.e. no trailing '/') which is not strictly kosher in WWW syntax
but is allowed by some servers and so does occur in some documents.  Generate
a redirection to the same URL but with a trailing slash on the directory name.
*/ 

RequestFileNoType (REQUEST_STRUCT *rqptr)

{
   int  Count;
   char  DirName [ODS_MAX_FILE_NAME_LENGTH+1];
   char  c;
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "RequestFileNoType()\n");

   zptr = (sptr = DirName) + sizeof(DirName)-1;
   cptr = rqptr->ParseOds.NamNamePtr;
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   cptr = ".DIR;";
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';

   c = rqptr->ParseOds.NamNamePtr[0];
   rqptr->ParseOds.NamNamePtr[0] = '\0';

   if (VMSok (OdsFileExists (rqptr->ParseOds.NamDevicePtr, DirName)))
   {
      Count = strlen(rqptr->ServicePtr->RequestSchemeNamePtr);
      if (rqptr->ServicePtr->ServerPort == 80)
         Count += strlen(rqptr->ServicePtr->ServerHostName);
      else
         Count += strlen(rqptr->ServicePtr->ServerHostPort);
      Count += rqptr->rqHeader.PathInfoLength;
      Count += rqptr->rqHeader.QueryStringLength;
      Count += 5;

      rqptr->rqResponse.LocationPtr = sptr = VmGetHeap (rqptr, Count);

      for (cptr = rqptr->ServicePtr->RequestSchemeNamePtr;
           *cptr;
           *sptr++ = *cptr++);
      *sptr++ = '/';
      *sptr++ = '/';
      if (rqptr->ServicePtr->ServerPort == 80)
         cptr = rqptr->ServicePtr->ServerHostName;
      else
         cptr = rqptr->ServicePtr->ServerHostPort;
      while (*cptr) *sptr++ = *cptr++;
      for (cptr = rqptr->rqHeader.PathInfoPtr; *cptr; *sptr++ = *cptr++);
      /* add the all-important trailing slash */
      *sptr++ = '/';
      if (rqptr->rqHeader.QueryStringLength)
      {
         *sptr++ = '?';
         for (cptr = rqptr->rqHeader.QueryStringPtr; *cptr; *sptr++ = *cptr++);
      }
      *sptr = '\0';
      RequestEnd (rqptr);

      return;
   }
   else
   {
      rqptr->ParseOds.NamNamePtr[0] = c;
      if (!rqptr->AccountingDone)
         rqptr->AccountingDone = ++AccountingPtr->DoFileCount;
      rqptr->rqResponse.ErrorTextPtr =
         MapVmsPath (rqptr->ParseOds.NamDevicePtr, rqptr);
      rqptr->rqResponse.ErrorOtherTextPtr = rqptr->ParseOds.NamDevicePtr;
      ErrorVmsStatus (rqptr, RMS$_FNF, FI_LI);
      RequestEnd (rqptr);

      return;
   }
}

/*****************************************************************************/
/*
Process request by executing a script.  The 'MappedFile' was the 
non-script-name  portion of the path, returned by the original MapUrl(), attempt
to parse this as a VMS file specification, but if that fails (and it well might
not be a file specification) then do not report it.  Provide to the script a
"best-guess" as to the file system ODS.
*/ 

RequestScript
(
REQUEST_STRUCT *rqptr,
char *MappedScript,
char *MappedFile,
char *MappedRunTime
)
{
   int  status;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "RequestScript() |%s|%s|%s|%s|%s|\n",
               rqptr->ScriptName, rqptr->rqHeader.PathInfoPtr,
               MappedScript, MappedFile, MappedRunTime);

   /***************************************/
   /* check for internal server "scripts" */
   /***************************************/

   if (tolower(rqptr->ScriptName[0]) == tolower(ADMIN_SCRIPT_ECHO[0]) &&
       strsame (rqptr->ScriptName, ADMIN_SCRIPT_ECHO,
                sizeof(ADMIN_SCRIPT_ECHO)-1))
   {
      RequestEcho (rqptr);
      return;
   }
   else
   if (tolower(rqptr->ScriptName[0]) == tolower(ADMIN_SCRIPT_WHERE[0]) &&
       strsame (rqptr->ScriptName, ADMIN_SCRIPT_WHERE,
                sizeof(ADMIN_SCRIPT_WHERE)-1))
   {
      RequestWhere (rqptr, MappedFile);
      return;
   }

   /********************************/
   /* parse any file specification */
   /********************************/

   if (MappedFile[0])
   {
      OdsParse (&rqptr->ParseOds,
                MappedFile, 0, NULL, 0,
                NAM$M_SYNCHK, NULL, rqptr);

      if (VMSok (status = rqptr->ParseOds.Fab.fab$l_sts))
      {
         if (VMSnok (status = OdsParseTerminate (&rqptr->ParseOds)))
         {
            ErrorNoticed (status, "OdsParseTerminate()", FI_LI);
            RequestEnd (rqptr);
            return;
         }
         if (!(rqptr->ParseOds.Nam_fnb & NAM$M_WILDCARD))
            ConfigContentType (&rqptr->rqContentInfo,
                               rqptr->ParseOds.NamTypePtr);
      }
      else
      {
         /* problem parsing (probably not intended as a file specification) */
         char  *cptr, *sptr, *zptr;
         zptr = (sptr = rqptr->ParseOds.ExpFileName) +
                sizeof(rqptr->ParseOds.ExpFileName);
         for (cptr = MappedFile; *cptr && sptr < zptr; *sptr++ = *cptr++);
         if (sptr >= zptr) sptr = rqptr->ParseOds.ExpFileName;
         *sptr = '\0';
         while (sptr > rqptr->ParseOds.ExpFileName && *sptr != '.') sptr--;
         if (*sptr == '.') ConfigContentType (&rqptr->rqContentInfo, sptr);
      }
   }
   else
      rqptr->ParseOds.ExpFileName[0] = '\0';

   if (Debug)
      fprintf (stdout, "rqptr->ParseOds.ExpFileName |%s|\n",
               rqptr->ParseOds.ExpFileName);

   /*********************************************/
   /* again check for internal server "scripts" */
   /*********************************************/

   if ((tolower(rqptr->ScriptName[0]) == tolower(ADMIN_SCRIPT_UPD[0]) &&
        strsame (rqptr->ScriptName, ADMIN_SCRIPT_UPD,
                 sizeof(ADMIN_SCRIPT_UPD)-1)))
   {
      UpdBegin (rqptr, &RequestEnd);
      return;
   }
   else
   if ((tolower(rqptr->ScriptName[0]) == tolower(ADMIN_SCRIPT_TREE[0]) &&
        strsame (rqptr->ScriptName, ADMIN_SCRIPT_TREE,
                 sizeof(ADMIN_SCRIPT_TREE)-1)))
   {
      UpdBegin (rqptr, &RequestEnd);
      return;
   }
   else
   if (tolower(rqptr->ScriptName[0]) == tolower(ADMIN_SCRIPT_XRAY[0]) &&
       strsame (rqptr->ScriptName, ADMIN_SCRIPT_XRAY,
                sizeof(ADMIN_SCRIPT_XRAY)-1))
   {
      /* indicate that it's an "Xray" and then cause a LOCAL redirection */
      if (Debug) fprintf (stdout, "Xray!\n");
      rqptr->XrayRequest = true;
      /* the redirection location is the request minus "/xray" */
      rqptr->rqResponse.LocationPtr = rqptr->rqHeader.RequestUriPtr +
                                      sizeof(ADMIN_SCRIPT_XRAY)-1;
      /* this can never be a legitimate keep-alive */
      rqptr->KeepAliveRequest = false;
      RequestEnd (rqptr);
      return;
   }

   /***************************/
   /* "real", external script */
   /***************************/

   for (cptr = MappedScript; *cptr && *cptr != ':'; cptr++);
   if (*(unsigned short*)cptr == '::')
      DECnetBegin (rqptr, &RequestEnd, MappedScript);
   else
   if (rqptr->IsCgiPlusScript)
      DclBegin (rqptr, &RequestEnd, NULL,
                rqptr->ScriptName, NULL, MappedScript, MappedRunTime, NULL);
   else
      DclBegin (rqptr, &RequestEnd, NULL,
                rqptr->ScriptName, MappedScript, NULL, MappedRunTime, NULL);
}

/*****************************************************************************/
/*
This simple little function is called directly and as an AST routine to echo
back to the client as a plain-text document the entire request header and any
body content. The call is set up in RequestScript() immediately above.
*/ 

RequestEcho (REQUEST_STRUCT *rqptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "RequestEcho() %d %d\n",
               rqptr->rqBody.CurrentPtr, rqptr->rqBody.CurrentCount);

   if (!rqptr->rqResponse.PreExpired)
   {
      /* detected that this is the setup call from RequestScript() */
      rqptr->rqResponse.PreExpired = true;
      /* pre-expired set also means this if() won't be executed again! */
      HTTP_HEADER_200_PLAIN (rqptr);
      rqptr->rqBody.CurrentPtr = rqptr->rqBody.BufferPtr;
   }

   if (rqptr->rqHeader.RequestHeaderPtr != NULL)
   {
      /* AST from writing echo header, write response header */
      NetWrite (rqptr, &RequestEcho,
                rqptr->rqHeader.RequestHeaderPtr,
                rqptr->rqHeader.RequestHeaderLength);
      rqptr->rqHeader.RequestHeaderPtr = NULL;
      return;
   }

   /* write response body */
   if (rqptr->rqBody.CurrentCount <= OutputBufferSize)
   {
      /* queue a network write to the client, AST to end request processing */
      NetWrite (rqptr, &RequestEnd, rqptr->rqBody.CurrentPtr,
                                    rqptr->rqBody.CurrentCount);
      return;
   }
   else
   {   
      /* break it into smaller chunks */
      char  *cptr;
      cptr = rqptr->rqBody.CurrentPtr;
      rqptr->rqBody.CurrentPtr += OutputBufferSize;
      rqptr->rqBody.CurrentCount -= OutputBufferSize;
      /* queue a network write to the client, AST to echo more/rest */
      NetWrite (rqptr, &RequestEcho, cptr, OutputBufferSize);
   }
}

/*****************************************************************************/
/*
After the mapping the path, etc., parse it again (in case of error,
RequestScript() ignores errors) and return the VMS file name as a plain text
document.  This function merely translates any logicals, etc., and reports the
resulting name, it does not demonstrate the directory or file actually exists. 
The call is set up in RequestScript() immediately above.
*/ 

RequestWhere
(
REQUEST_STRUCT *rqptr,
char *MappedFile
)
{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "RequestWhere()\n");

   OdsParse (&rqptr->ParseOds,
             MappedFile, 0, NULL, 0,
             NAM$M_SYNCHK, NULL, rqptr);

   if (VMSok (status = rqptr->ParseOds.Fab.fab$l_sts))
   {
      if (VMSnok (status = OdsParseTerminate (&rqptr->ParseOds)))
      {
         ErrorNoticed (status, "OdsParseTerminate()", FI_LI);
         RequestEnd (rqptr);
         return;
      }
      rqptr->rqResponse.PreExpired = true;
      HTTP_HEADER_200_PLAIN (rqptr);
      /* queue a network write to the client, AST to end processing */
      NetWrite (rqptr, &RequestEnd,
                rqptr->ParseOds.ExpFileName, rqptr->ParseOds.ExpFileNameLength);
      return;
   }
   else
   {
      rqptr->rqResponse.ErrorTextPtr = rqptr->rqHeader.PathInfoPtr;
      ErrorVmsStatus (rqptr, status, FI_LI);
      RequestEnd (rqptr);
      return;
   }
}

/*****************************************************************************/
/*
This function updates the global section with a formatted representation of
the data of the latest request.  The acounting data is of course permanently
located in the area and so does not require any explicit placement into the
section.
*/ 

RequestGblSecUpdate (REQUEST_STRUCT *rqptr)

{
   int  status;
   unsigned long  FaoVector [16];
   unsigned long  *vecptr;
   unsigned long  Seconds,
                  SubSeconds;
   char  *cptr, *sptr, *zptr;
   HTTPD_GBLSEC  *gsptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "RequestGblSecUpdate()\n");

   gsptr = HttpdGblSecPtr;

   if (rqptr == NULL)
   {
      /* reset request data */
      memset (&gsptr->Request, 0, sizeof(gsptr->Request));
      return;
   }

   /* if a keep-alive timeout then just return */
   if (rqptr->KeepAliveTimeout) return;

   /***********************************/
   /* update the monitor request data */
   /***********************************/

   gsptr->Request.HttpStatus = rqptr->rqResponse.HttpStatus;
   gsptr->Request.BytesRawRx = rqptr->BytesRawRx;
   gsptr->Request.BytesRawTx = rqptr->BytesRawTx;

   vecptr = FaoVector;
   *vecptr++ = rqptr->rqTime.VmsVector[2];
   *vecptr++ = rqptr->rqTime.VmsVector[3];
   *vecptr++ = rqptr->rqTime.VmsVector[4];
   *vecptr++ = rqptr->rqTime.VmsVector[5];
   WriteFaol (gsptr->Request.Time, sizeof(gsptr->Request.Time), NULL,
              "!2ZL !2ZL:!2ZL:!2ZL", &FaoVector);

   Seconds = rqptr->rqResponse.Duration / USECS_IN_A_SECOND;
   SubSeconds = (rqptr->rqResponse.Duration % USECS_IN_A_SECOND) /
                DURATION_UNITS_USECS;
   vecptr = FaoVector;
   *vecptr++ = Seconds;
   *vecptr++ = DURATION_DECIMAL_PLACES;
   *vecptr++ = SubSeconds;
   WriteFaol (gsptr->Request.Duration, sizeof(gsptr->Request.Duration), NULL,
              "!UL.!#ZL", &FaoVector);

   /* e.g. http://the.server.host.name:port */
   zptr = (sptr = gsptr->Request.Service) + sizeof(gsptr->Request.Service)-1;
   for (cptr = rqptr->ServicePtr->RequestSchemeNamePtr;
        *cptr && sptr < zptr;
        *sptr++ = *cptr++); 
   if (sptr < zptr) *sptr++ = '/';
   if (sptr < zptr) *sptr++ = '/';
   for (cptr = rqptr->ServicePtr->ServerHostName;
        *cptr && sptr < zptr;
        *sptr++ = *cptr++); 
   if (sptr < zptr) *sptr++ = ':';
   for (cptr = rqptr->ServicePtr->ServerPortString;
        *cptr && sptr < zptr;
        *sptr++ = *cptr++); 
   *sptr = '\0';

   zptr = (sptr = gsptr->Request.ClientHostName) +
          sizeof(gsptr->Request.ClientHostName)-1;
   for (cptr = rqptr->rqNet.ClientHostName;
        *cptr && sptr < zptr;
        *sptr++ = *cptr++); 
   *sptr = '\0';

   zptr = (sptr = gsptr->Request.MethodName) +
          sizeof(gsptr->Request.MethodName)-1;
   for (cptr = rqptr->rqHeader.MethodName;
        *cptr && sptr < zptr;
        *sptr++ = *cptr++); 
   *sptr = '\0';

   if (rqptr->rqHeader.RequestUriPtr == NULL)
      gsptr->Request.Uri[0] = '\0';
   else
   {
      zptr = (sptr = gsptr->Request.Uri) + sizeof(gsptr->Request.Uri)-1;
      for (cptr = rqptr->rqHeader.RequestUriPtr;
           *cptr && sptr < zptr;
           *sptr++ = *cptr++); 
      *sptr = '\0';
   }

   if (rqptr->rqNet.ReadErrorCount)
      status = WriteFao (gsptr->Request.ReadError,
                         sizeof(gsptr->Request.ReadError),
                         NULL, "%X!8XL !-!%m",
                         rqptr->rqNet.ReadErrorStatus);
   else
      gsptr->Request.ReadError[0] = '\0';

   if (rqptr->rqNet.WriteErrorCount)
      status = WriteFao (gsptr->Request.WriteError,
                         sizeof(gsptr->Request.WriteError),
                         NULL, "%X!8XL !-!%m",
                         rqptr->rqNet.WriteErrorStatus);
   else
      gsptr->Request.WriteError[0] = '\0';
}

/*****************************************************************************/
/*
Keeps a linked list history from most to least recent that can be reported on
by RequestReport().
*/

RequestHistory (REQUEST_STRUCT *rqptr)

{
   int  status;
   char  *cptr, *sptr, *zptr;
   struct RequestHistoryStruct  *hlptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "RequestHistory()\n");

   if (RequestHistoryCount < RequestHistoryMax)
   {
      /* allocate memory for a new entry */
      hlptr = VmGet (sizeof (struct RequestHistoryStruct));
      RequestHistoryCount++;
   }
   else
   {
      /* reuse the tail entry (least recent) */
      hlptr = RequestHistoryList.TailPtr;
      ListRemove (&RequestHistoryList, hlptr);
   }

   /* add entry to the head of the history list (most recent) */
   ListAddHead (&RequestHistoryList, hlptr);

   memcpy (&hlptr->BinaryTime, &rqptr->rqTime.Vms64bit, 8);
   hlptr->ConnectNumber = rqptr->ConnectNumber;
   hlptr->ServicePtr = rqptr->ServicePtr;
   hlptr->BytesRawRx = rqptr->BytesRawRx;
   hlptr->BytesRawTx = rqptr->BytesRawTx;
   hlptr->ResponseDuration = rqptr->rqResponse.Duration;
   hlptr->ResponseStatusCode = rqptr->rqResponse.HttpStatus;

   zptr = (sptr = hlptr->ClientAndRequest) + sizeof(hlptr->ClientAndRequest);

   if (rqptr->RemoteUser[0])
   {
      cptr = rqptr->RemoteUser;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      if (rqptr->rqAuth.RealmPtr != NULL)
      {
         if (sptr < zptr) *sptr++ = '.';
         cptr = rqptr->rqAuth.RealmPtr;
         while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      }
      if (sptr < zptr) *sptr++ = '@';
   }
   cptr = rqptr->rqNet.ClientHostName;
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr < zptr) *sptr++ = '\0';
   hlptr->RequestPtr = sptr;

   if (rqptr->rqHeader.PathInfoPtr != NULL && rqptr->rqHeader.QueryStringPtr != NULL)
   {
      cptr = rqptr->rqHeader.MethodName;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      if (sptr < zptr) *sptr++ = ' ';
      cptr = rqptr->ScriptName;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      cptr = rqptr->rqHeader.PathInfoPtr;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      if (rqptr->rqHeader.QueryStringPtr[0] && sptr < zptr) *sptr++ = '?';
      cptr = rqptr->rqHeader.QueryStringPtr;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   }

   if (sptr >= zptr)
   {
      hlptr->Truncated = true;
      sptr--;
   }
   else
      hlptr->Truncated = false;
   *sptr = '\0';
}

/*****************************************************************************/
/*
Return a report on current and request history, listed most to least recent.
This function blocks while executing.
*/ 

RequestReport
(
REQUEST_STRUCT *rqptr,
REQUEST_AST NextTaskFunction,
BOOL IncludeHistory
)
{
   static long  Subx2 = 2,
                EdivTen = 10;

   static char  ResponseFao [] =
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ ... Request Report</TITLE>\n\
</HEAD>\n\
!AZ\n\
<H2><NOBR>HTTPd !AZ</NOBR></H1>\n\
<H3>Request Report</H2>\n\
!20%W\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH COLSPAN=9><FONT SIZE=+1>Current</FONT></TH></TR>\n\
<TR><TH COLSPAN=2>Client</TH><TH>Time</TH>\
<TH>Rx</TH><TH>Tx</TH><TH>Connect</TH>\
<TH>Duration</TH><TH>!AZ</TH><TH>WATCH</TH></TR>\
<TR><TH COLSPAN=9>Request</TH></TR>\n\
<TR></TR>\n";

   /* this is the same as below but with links for the connect number */
   static char  CurrentRequestFao [] =
"<TR><TH ROWSPAN=2 ALIGN=right>&nbsp;!UL&nbsp;</TH>\
<TD ALIGN=left>!%%</TD><TD><FONT SIZE=-1>!23%W</FONT></TD>\
<TD ALIGN=right>!UL</TD>\
<TD ALIGN=right>!UL</TD>\
<TD ALIGN=right>!UL</TD>\
<TD ALIGN=right>!UL.!#ZL!AZ</TD>\
<TD ALIGN=right>!%%</TD>\
<TD ALIGN=right>!%%</TD></TR>\n\
<TR><TD ALIGN=left COLSPAN=8>!%%</TD></TR>\n";

   static char  HistoryFao [] =
"</TABLE>\n\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH COLSPAN=8><FONT SIZE=+1>History &nbsp;&nbsp;</FONT>\
<FONT SIZE=-1>!%%</FONT></TH></TR>\n\
<TR><TH COLSPAN=2>Client</TH><TH>Time</TH>\
<TH>Status</HT><TH>Rx</TH><TH>Tx</TH><TH>Duration</TH>\
<TH>Number</TH></TR>\n\
<TR><TH COLSPAN=8>Request</TH></TR>\n\
<TR></TR>\n";

   static char  HistoryRequestFao [] =
"<TR><TH ROWSPAN=2 ALIGN=right>&nbsp;!UL&nbsp;</TH>\
<TD ALIGN=left>!AZ</TD><TD><FONT SIZE=-1>!23%W</FONT></TD>\
<TD ALIGN=right>!UL</TD>\
<TD ALIGN=right>!UL</TD><TD ALIGN=right>!UL</TD>\
<TD ALIGN=right>!UL.!#ZL</TD>\
<TD ALIGN=right>!UL</TD></TR>\n\
<TR><TD ALIGN=left COLSPAN=7><NOBR><TT>!AZ</TT></NOBR>!AZ</TD></TR>\n";

   static char  HistoryButtonFao [] = 
"</TABLE>\n\
<P><FONT SIZE=+1>[<A HREF=\"!AZ\">History</A>]</FONT>\n\
</BODY>\n\
</HTML>\n";

   static char  EndPageFao [] =
"</TABLE>\n\
</BODY>\n\
</HTML>\n";

   int  status,
        Count;
   unsigned long  ResponseDuration,
                  Remainder,
                  Seconds,
                  SubSeconds;
   unsigned long  *vecptr;
   unsigned long  BinaryTime [2],
                  FaoVector [32],
                  ResultTime [2];
   char  *cptr, *sptr, *zptr;
   char  Request [512],
         RequestBuffer [512];
   REQUEST_STRUCT  *rqeptr;
   LIST_ENTRY  *leptr;
   struct RequestHistoryStruct  *rqhptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "RequestReport()\n");

   rqptr->rqResponse.PreExpired = PRE_EXPIRE_ADMIN;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;
   *vecptr++ = HtmlMetaInfo (rqptr, NULL);
   *vecptr++ = ServerHostPort;
   *vecptr++ = Config.cfServer.AdminBodyTag;
   *vecptr++ = ServerHostPort;
   *vecptr++ = &rqptr->rqTime.Vms64bit;
   if (MappingRulesList.ThrottleTotal)
      *vecptr++ = "Throttle";
   else
      *vecptr++ = "<I>Throttle</I>";

   status = NetWriteFaol (rqptr, ResponseFao, &FaoVector);
   if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);

   /********************/
   /* current requests */
   /********************/

   Count = 0;

   sys$gettim (&BinaryTime);

   if (Debug)
      fprintf (stdout, "Head %d\nTail %d\n",
               RequestList.HeadPtr, RequestList.TailPtr);

   /* process the request list from least to most recent */
   for (leptr = RequestList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
   {
      rqeptr = (REQUEST_STRUCT*)leptr;

      if (Debug)
         fprintf (stdout, "%d <- %d -> %d (%d)\n",
                  leptr->PrevPtr, leptr, leptr->NextPtr, rqeptr);

      vecptr = FaoVector;

      *vecptr++ = ++Count;
      if (rqeptr->RemoteUser[0])
      {
         if (rqeptr->rqAuth.RealmPtr == NULL)
            *vecptr++ = "!HZ@!AZ";
         else
            *vecptr++ = "!HZ.!HZ@!AZ";
         *vecptr++ = rqeptr->RemoteUser;
         if (rqeptr->rqAuth.RealmPtr != NULL)
            *vecptr++ = rqeptr->rqAuth.RealmPtr;
      }
      else
         *vecptr++ = "!AZ";
      *vecptr++ = rqeptr->rqNet.ClientHostName;

      *vecptr++ = &rqeptr->rqTime.Vms64bit;
      *vecptr++ = rqeptr->BytesRawRx;
      *vecptr++ = rqeptr->BytesRawTx;

      *vecptr++ = rqeptr->ConnectNumber;

      status = lib$subx (&BinaryTime, &rqeptr->rqTime.Vms64bit, 
                         &ResultTime, &Subx2);
      /* convert to microseconds (longword becomes 71 minutes max) */
      if (VMSok (status))
         status = lib$ediv (&EdivTen, &ResultTime,
                            &ResponseDuration, &Remainder);
      if (Debug) fprintf (stdout, "elapsed: %%X%08.08X\n", status);
      if (VMSnok (status)) ResponseDuration = 0;

      Seconds = ResponseDuration / USECS_IN_A_SECOND;
      SubSeconds = (ResponseDuration % USECS_IN_A_SECOND) /
                   DURATION_UNITS_USECS;
      *vecptr++ = Seconds;
      *vecptr++ = DURATION_DECIMAL_PLACES;
      *vecptr++ = SubSeconds;

      if (rqeptr->rqTmr.TerminatedCount)
         *vecptr++ = " (timed-out)";
      else
         *vecptr++ = "";

      if (rqeptr->rqPathSet.ThrottleFrom)
      {
         if (rqeptr->ThrottleNextFunction == NULL)
            *vecptr++ = "<FONT SIZE=-1>processing</FONT>";
         else
         {
            *vecptr++ =
"<FONT SIZE=-1>\
[<A HREF=\"!AZ?this=!UL\">release</A>]<BR>\
[<A HREF=\"!AZ?this=!UL\">terminate</A>]\
</FONT>";
            *vecptr++ = ADMIN_CONTROL_THROTTLE_RELEASE;
            *vecptr++ = rqeptr->ConnectNumber;
            *vecptr++ = ADMIN_CONTROL_THROTTLE_TERMINATE;
            *vecptr++ = rqeptr->ConnectNumber;
         }
      }
      else
         *vecptr++ = "";

      if (rqeptr->ConnectNumber != rqptr->ConnectNumber)
      {
         *vecptr++ =
"<FONT SIZE=-1>\
[<A HREF=\"!AZ?at=!UL\">peek</A>]<BR>\
[<A HREF=\"!AZ?at=!UL&this=!UL\">peek+</A>]<BR>\
[<A HREF=\"!AZ?this=!UL\">watch</A>]\
</FONT>";
         *vecptr++ = ADMIN_REPORT_WATCH;
         *vecptr++ = rqeptr->ConnectNumber;
         *vecptr++ = ADMIN_REPORT_WATCH;
         *vecptr++ = rqeptr->ConnectNumber;
         *vecptr++ = rqeptr->ConnectNumber;
         *vecptr++ = ADMIN_REPORT_WATCH;
         *vecptr++ = rqeptr->ConnectNumber;
      }
      else
         *vecptr++ = "<FONT SIZE=-1>current</FONT>";

      if (rqeptr->rqHeader.PathInfoPtr == NULL ||
          rqeptr->rqHeader.QueryStringPtr == NULL)
      {
         if (rqeptr->rqNet.KeepAliveCount)
         {
            *vecptr++ = "&quot;Keep-Alive:&quot !UL";
            *vecptr++ = rqeptr->rqNet.KeepAliveCount;
         }
         else
            *vecptr++ = "";
      }
      else
      {
         *vecptr++ = "<NOBR><TT>!%%!AZ !UZ!UZ!%??\r\r!AZ</TT></NOBR>";
         if (ServiceCount > 1)
         {
            *vecptr++ = "!AZ//!AZ ";
            *vecptr++ = rqeptr->ServicePtr->RequestSchemeNamePtr;
            *vecptr++ = rqeptr->ServicePtr->ServerHostPort;
         }
         else
            *vecptr++ = "";
         *vecptr++ = rqeptr->rqHeader.MethodName;
         *vecptr++ = rqeptr->ScriptName;
         *vecptr++ = rqeptr->rqHeader.PathInfoPtr;
         *vecptr++ = rqeptr->rqHeader.QueryStringPtr[0];
         *vecptr++ = rqeptr->rqHeader.QueryStringPtr;
      }

      status = NetWriteFaol (rqptr, CurrentRequestFao, &FaoVector);
      if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);
   }

   if (!IncludeHistory)
   {
      status = NetWriteFao (rqptr, HistoryButtonFao,
                            ADMIN_REPORT_REQUEST_HISTORY);
      if (VMSnok (status)) ErrorNoticed (status, "NetWriteFao()", FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   /*******************/
   /* request history */
   /*******************/

   vecptr = FaoVector;
   if (RequestHistoryMax)
   {
      *vecptr++ = "(maximum !UL)";
      *vecptr++ = RequestHistoryMax;
   }
   else
      *vecptr++ = "<I>(disabled)</I>";

   status = NetWriteFaol (rqptr, HistoryFao, &FaoVector);
   if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);

   Count = 0;

   if (Debug)
      fprintf (stdout, "Head %d\nTail %d\n",
               RequestHistoryList.HeadPtr, RequestHistoryList.TailPtr);

   /* process the request list from least to most recent */
   for (leptr = RequestHistoryList.HeadPtr;
        leptr != NULL;
        leptr = leptr->NextPtr)
   {
      rqhptr = (struct RequestHistoryStruct*)leptr;

      if (Debug)
         fprintf (stdout, "%d <- %d -> %d (%d)\n",
                  leptr->PrevPtr, leptr, leptr->NextPtr, rqhptr);

      vecptr = FaoVector;
      *vecptr++ = ++Count;
      *vecptr++ = rqhptr->ClientAndRequest;
      *vecptr++ = &rqhptr->BinaryTime;
      *vecptr++ = rqhptr->ResponseStatusCode;
      *vecptr++ = rqhptr->BytesRawRx;
      *vecptr++ = rqhptr->BytesRawTx;

      Seconds = rqhptr->ResponseDuration / USECS_IN_A_SECOND;
      SubSeconds = (rqhptr->ResponseDuration % USECS_IN_A_SECOND) /
                   DURATION_UNITS_USECS;
      *vecptr++ = Seconds;
      *vecptr++ = DURATION_DECIMAL_PLACES;
      *vecptr++ = SubSeconds;

      *vecptr++ = rqhptr->ConnectNumber;

      zptr = (sptr = RequestBuffer) + sizeof(RequestBuffer);
      if (ServiceCount > 1)
      {
         cptr = rqhptr->ServicePtr->RequestSchemeNamePtr;
         while (*cptr && sptr < zptr) *sptr++ = *cptr++;
         if (sptr < zptr) *sptr++ = '/';
         if (sptr < zptr) *sptr++ = '/';
         cptr = rqhptr->ServicePtr->ServerHostPort;
         while (*cptr && sptr < zptr) *sptr++ = *cptr++;
         if (sptr < zptr) *sptr++ = ' ';
      }
      cptr = rqhptr->RequestPtr;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      if (sptr >= zptr)
         ErrorExitVmsStatus (SS$_BUFFEROVF, "RequestReport()", FI_LI);
      *sptr = '\0';

      if (RequestBreakLine (RequestBuffer, Request, sizeof(Request)) == -1)
         ErrorExitVmsStatus (SS$_BUFFEROVF, "RequestReport()", FI_LI);

      *vecptr++ = Request;

      if (rqhptr->Truncated)
         *vecptr++ = "[truncated!]";
      else
         *vecptr++ = "";

      status = NetWriteFaol (rqptr, HistoryRequestFao, &FaoVector);
      if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);
   }

   status = NetWriteFaol (rqptr, EndPageFao, NULL);
   if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);

   SysDclAst (NextTaskFunction, rqptr);
}

/*****************************************************************************/
/*
Used for request history report.  HTML-break the supplied source string every
eighty characters.  Escape HTML-forbidden characters.  Returns the number of
characters in the destination string, or -1 if an error.
*/

int RequestBreakLine
(
char *Source,
char *Destination,
int SizeOfDestination
)
{
   int  cnt;
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "RequestBreakLine() |%s|\n", Source);

   /* provide a line-break every 80 characters */
   cnt = -1;
   zptr = (sptr = Destination) + SizeOfDestination;
   cptr = Source;
   while (*cptr && sptr < zptr)
   {
      cnt++;
      if (!(cnt % 80) && cnt)
         if (sptr+4 < zptr) *((unsigned long*)sptr)++ = '<BR>';
      switch (*cptr)
      {
         case '<' :
            if (sptr+4 >= zptr) return (-1);
            *((unsigned long*)sptr)++ = '&gt;';
            cptr++;
            continue;
         case '>' :
            if (sptr+4 >= zptr) return (-1);
            *((unsigned long*)sptr)++ = '&lt;';
            cptr++;
            continue;
         case '&' :
            if (sptr+5 >= zptr) return (-1);
            *((unsigned long*)sptr)++ = '&amp';
            *sptr++ = ';';
            cptr++;
            continue;
         default :
            if (sptr < zptr) *sptr++ = *cptr++;
            continue;
      }
   }
   if (sptr >= zptr) return (-1);
   *sptr = '\0';

   if (Debug) fprintf (stdout, "%d |%s|\n", sptr-Destination, Destination);
   return (sptr - Destination);
}

/*****************************************************************************/

